@@ -21,10 +21,6 @@
 #include "advection.h"
 #include "source.h"
 
-
-#include "solid.h"
-
-#if 1
 static gdouble transverse_term (FttCell * cell,
 				gdouble size,
 				FttComponent c,
@@ -32,7 +28,7 @@ static gdouble transverse_term (FttCell * cell,
 {
   GfsStateVector * s = GFS_STATE (cell);
   gdouble vtan = par->use_centered_velocity ? 
-    GFS_VARIABLE (cell, GFS_VELOCITY_INDEX (c)) :
+    GFS_VARIABLE (cell, par->u[c]->i) :
     (s->f[2*c].un + s->f[2*c + 1].un)/2.;
   FttCellFace f;
   GfsGradient gf;
@@ -70,13 +66,11 @@ void gfs_cell_advected_face_values (FttCell * cell,
   size = ftt_cell_size (cell);
   for (c = 0; c < FTT_DIMENSION; c++) {
     gdouble unorm = par->use_centered_velocity ?
-      par->dt*GFS_VARIABLE (cell, GFS_VELOCITY_INDEX (c))/size :
+      par->dt*GFS_VARIABLE (cell, par->u[c]->i)/size :
       par->dt*(s->f[2*c].un + s->f[2*c + 1].un)/(2.*size);
     gdouble g = (* par->gradient) (cell, c, par->v->i);
-    gdouble vl = GFS_VARIABLE (cell, par->v->i) + 
-      MIN ((1. - unorm)/2., 0.5)*g;
-    gdouble vr = GFS_VARIABLE (cell, par->v->i) + 
-      MAX ((- 1. - unorm)/2., -0.5)*g;
+    gdouble vl = GFS_VARIABLE (cell, par->v->i) + MIN ((1. - unorm)/2., 0.5)*g;
+    gdouble vr = GFS_VARIABLE (cell, par->v->i) + MAX ((- 1. - unorm)/2., -0.5)*g;
     gdouble src = par->dt*gfs_variable_mac_source (par->v, cell)/2.;
     gdouble dv;
 
@@ -95,56 +89,6 @@ void gfs_cell_advected_face_values (FttCell * cell,
     s->f[2*c + 1].v = vr + src - dv;
   }
 }
-#else
-/**
- * gfs_cell_advected_face_values:
- * @cell: a #FttCell.
- * @par: the advection parameters.
- *
- * Fills the face variable (@v field of #GfsFaceStateVector) of all the
- * faces of @cell with the advected value of variable @par->v at time
- * t + dt/2.
- */
-void gfs_cell_advected_face_values (FttCell * cell,
-				    const GfsAdvectionParams * par)
-{
-  FttComponent c;
-  gdouble size;
-  gdouble g[FTT_DIMENSION];
-
-  g_return_if_fail (cell != NULL);
-  g_return_if_fail (par != NULL);
-
-  size = ftt_cell_size (cell);
-  for (c = 0; c < FTT_DIMENSION; c++)
-    g[c] = (* par->gradient) (cell, c, par->v);
-
-  for (c = 0; c < FTT_DIMENSION; c++) {
-    gdouble u = par->dt*GFS_VARIABLE (cell, GFS_VELOCITY (c))/size;
-    gdouble vl = GFS_VARIABLE (cell, par->v) + MIN ((1. - u)/2., 0.5)*g[c];
-    gdouble vr = GFS_VARIABLE (cell, par->v) + MAX ((-1. - u)/2., -0.5)*g[c];
-    gdouble dv;
-    FttComponent cp;
-#if FTT_2D
-
-    cp = FTT_ORTHOGONAL_COMPONENT (c);
-    dv = par->dt*GFS_VARIABLE (cell, GFS_VELOCITY (cp))*g[cp]/(2.*size);
-#else  /* FTT_3D */
-    static FttComponent orthogonal[FTT_DIMENSION][2] = {
-      {FTT_Y, FTT_Z}, {FTT_X, FTT_Z}, {FTT_X, FTT_Y}
-    };
-
-    cp = orthogonal[c][0];
-    dv = par->dt*GFS_VARIABLE (cell, GFS_VELOCITY (cp))*g[cp]/(2.*size);
-    cp = orthogonal[c][1];
-    dv += par->dt*GFS_VARIABLE (cell, GFS_VELOCITY (cp))*g[cp]/(2.*size);
-#endif /* FTT_3D */
-
-    cell->s.f[2*c].v     = vl - dv;
-    cell->s.f[2*c + 1].v = vr - dv;
-  }
-}
-#endif
 
 /**
  * gfs_cell_non_advected_face_values:
@@ -236,7 +180,8 @@ static gdouble interpolate_2D1 (const FttCell * cell,
   FttDirection dleft;
 
   g_return_val_if_fail (cell != NULL, 0.);
-  g_return_val_if_fail (!GFS_IS_MIXED (cell), 0.);
+  /* fixme: this routine does not take into account mixed cells
+     fractions (in contrast to interpolate_1D1 above) */
 
   dleft = FTT_OPPOSITE_DIRECTION (dright);
   v0 = GFS_STATE (cell)->f[dleft].v;
@@ -252,13 +197,16 @@ static gdouble interpolate_2D1 (const FttCell * cell,
       d[0] = FTT_OPPOSITE_DIRECTION (dright);
       d[1] = FTT_OPPOSITE_DIRECTION (d1);
       d[2] = d2;
-      n1 = ftt_cell_child_corner (n1, d);
-      /* check for mixed cell refinement violation */
-      g_assert (n1);
-      x1 = 1./4.;
-      y1 = 3./4.;
+      if ((n1 = ftt_cell_child_corner (n1, d))) {
+	v1 = GFS_STATE (n1)->f[dleft].v;
+	x1 = 1./4.;
+	y1 = 3./4.;
+      }
+      else
+	v1 = v0;
     }
-    v1 = GFS_STATE (n1)->f[dleft].v;
+    else
+      v1 = GFS_STATE (n1)->f[dleft].v;
   }
   else
     v1 = v0;
@@ -274,13 +222,16 @@ static gdouble interpolate_2D1 (const FttCell * cell,
       d[0] = FTT_OPPOSITE_DIRECTION (dright);
       d[1] = FTT_OPPOSITE_DIRECTION (d2);
       d[2] = d1;
-      n2 = ftt_cell_child_corner (n2, d);
-      /* check for mixed cell refinement violation */
-      g_assert (n2);
-      x2 = 3./4.;
-      y2 = 1./4.;
+      if ((n2 = ftt_cell_child_corner (n2, d))) {
+	v2 = GFS_STATE (n2)->f[dleft].v;
+	x2 = 3./4.;
+	y2 = 1./4.;
+      }
+      else
+	v2 = v0;
     }
-    v2 = GFS_STATE (n2)->f[dleft].v;
+    else
+      v2 = GFS_STATE (n2)->f[dleft].v;
   }
   else
     v2 = v0;
@@ -295,6 +246,7 @@ static gdouble interpolate_2D1 (const FttCell * cell,
  * gfs_face_upwinded_value:
  * @face: a #FttCellFace.
  * @upwinding: type of upwinding.
+ * @u: the cell-centered velocity.
  *
  * This function assumes that the face variable has been previously
  * defined using gfs_cell_advected_face_values().
@@ -302,7 +254,8 @@ static gdouble interpolate_2D1 (const FttCell * cell,
  * Returns: the upwinded value of the face variable.  
  */
 gdouble gfs_face_upwinded_value (const FttCellFace * face,
-				 GfsUpwinding upwinding)
+				 GfsUpwinding upwinding,
+				 GfsVariable ** u)
 {
   gdouble un = 0.;
 
@@ -313,11 +266,16 @@ gdouble gfs_face_upwinded_value (const FttCellFace * face,
 
   switch (upwinding) {
   case GFS_CENTERED_UPWINDING:
-    un = gfs_face_interpolated_value (face, GFS_VELOCITY_INDEX (face->d/2)); break;
+    g_return_val_if_fail (u != NULL, 0.);
+    un = gfs_face_interpolated_value (face, u[face->d/2]->i); 
+    break;
   case GFS_FACE_UPWINDING:
-    un = GFS_FACE_NORMAL_VELOCITY (face); break;
+    un = GFS_FACE_NORMAL_VELOCITY (face); 
+    break;
   case GFS_NO_UPWINDING:
     break;
+  default:
+    g_assert_not_reached ();
   }
   if (!FTT_FACE_DIRECT (face))
     un = - un;
@@ -397,7 +355,7 @@ void gfs_face_advection_flux (const FttCellFace * face,
   g_return_if_fail (par != NULL);
 
   flux = GFS_FACE_FRACTION (face)*GFS_FACE_NORMAL_VELOCITY (face)*par->dt*
-    gfs_face_upwinded_value (face, FALSE)/ftt_cell_size (face->cell);
+    gfs_face_upwinded_value (face, GFS_FACE_UPWINDING, NULL)/ftt_cell_size (face->cell);
   if (!FTT_FACE_DIRECT (face))
     flux = - flux;
   GFS_VARIABLE (face->cell, par->fv->i) -= flux;
@@ -438,7 +396,7 @@ void gfs_face_velocity_advection_flux (const FttCellFace * face,
   g_return_if_fail (face != NULL);
   g_return_if_fail (par != NULL);
 
-  c = GFS_VELOCITY_COMPONENT (par->v->i);
+  c = par->v->component;
   g_return_if_fail (c >= 0 && c < FTT_DIMENSION);
 
   flux = GFS_FACE_FRACTION (face)*GFS_FACE_NORMAL_VELOCITY (face)*par->dt
@@ -448,9 +406,9 @@ void gfs_face_velocity_advection_flux (const FttCellFace * face,
     flux *= GFS_FACE_NORMAL_VELOCITY (face);
   else /* tangential component */
 #else
-    flux *= gfs_face_upwinded_value (face, par->upwinding)
+    flux *= gfs_face_upwinded_value (face, par->upwinding, par->u)
       /* pressure correction */
-      - gfs_face_interpolated_value (face, GFS_GRADIENT_INDEX (c))*par->dt/2.;
+      - gfs_face_interpolated_value (face, par->g[c]->i)*par->dt/2.;
 #endif
   if (!FTT_FACE_DIRECT (face))
     flux = - flux;
@@ -493,7 +451,7 @@ void gfs_face_velocity_convective_flux (const FttCellFace * face,
   g_return_if_fail (par != NULL);
   g_return_if_fail (GFS_FACE_FRACTION (face) == 1.);
 
-  c = GFS_VELOCITY_COMPONENT (par->v->i);
+  c = par->v->component;
   g_return_if_fail (c >= 0 && c < FTT_DIMENSION);
 
 #if 0
@@ -504,9 +462,9 @@ void gfs_face_velocity_convective_flux (const FttCellFace * face,
       /* pressure correction */
       - gfs_face_interpolated_value (face, GFS_GRADIENT_INDEX (c))*par->dt/2.;
 #else
-  u = gfs_face_upwinded_value (face, par->upwinding)
+  u = gfs_face_upwinded_value (face, par->upwinding, par->u)
     /* pressure correction */
-    - gfs_face_interpolated_value (face, GFS_GRADIENT_INDEX (c))*par->dt/2.;
+    - gfs_face_interpolated_value (face, par->g[c]->i)*par->dt/2.;
 #endif
   u *= par->dt/(2.*ftt_cell_size (face->cell));
   if (!FTT_FACE_DIRECT (face))
@@ -535,7 +493,7 @@ void gfs_face_velocity_convective_flux (const FttCellFace * face,
 /**
  * gfs_face_advected_normal_velocity:
  * @face: a #FttCellFace.
- * @upwinding: the type of upwinding.
+ * @par: the #GfsAdvectionParams.
  *
  * Fills the normal component of the velocity at @face with the value
  * advected (to time t + dt/2) from the centered velocities.
@@ -545,16 +503,20 @@ void gfs_face_velocity_convective_flux (const FttCellFace * face,
  * gfs_cell_advected_face_values().  
  */
 void gfs_face_advected_normal_velocity (const FttCellFace * face,
-					GfsUpwinding * upwinding)
+					const GfsAdvectionParams * par)
 {
   gdouble u;
 
   g_return_if_fail (face != NULL);
+  g_return_if_fail (par != NULL);
 
   if (GFS_FACE_FRACTION (face) == 0.)
     return;
 
-  GFS_FACE_NORMAL_VELOCITY_LEFT (face) = u = gfs_face_upwinded_value (face, *upwinding);
+  GFS_FACE_NORMAL_VELOCITY_LEFT (face) = u =
+    GFS_CELL_IS_BOUNDARY (face->neighbor) ?
+    GFS_STATE (face->neighbor)->f[FTT_OPPOSITE_DIRECTION (face->d)].v :
+    gfs_face_upwinded_value (face, par->upwinding, par->u);
 
   switch (ftt_face_type (face)) {
   case FTT_FINE_FINE:
@@ -573,21 +535,22 @@ void gfs_face_advected_normal_velocity (const FttCellFace * face,
 /**
  * gfs_face_interpolated_normal_velocity:
  * @face: a #FttCellFace.
+ * @v: the velocity.
  *
  * Fills the normal component of the velocity at @face with the value
  * interpolated from the centered velocities.
  */
-void gfs_face_interpolated_normal_velocity (const FttCellFace * face)
+void gfs_face_interpolated_normal_velocity (const FttCellFace * face, GfsVariable ** v)
 {
   gdouble u;
 
   g_return_if_fail (face != NULL);
+  g_return_if_fail (v != NULL);
 
   if (GFS_FACE_FRACTION (face) == 0.)
     return;
 
-  GFS_FACE_NORMAL_VELOCITY_LEFT (face) = u = 
-    gfs_face_interpolated_value (face, GFS_VELOCITY_INDEX (face->d/2));
+  GFS_FACE_NORMAL_VELOCITY_LEFT (face) = u = gfs_face_interpolated_value (face, v[face->d/2]->i);
 
   switch (ftt_face_type (face)) {
   case FTT_FINE_FINE:
@@ -617,11 +580,27 @@ void gfs_face_reset_normal_velocity (const FttCellFace * face)
     GFS_FACE_NORMAL_VELOCITY_LEFT (face) = 0.;
 }
 
+static gboolean is_small (FttCell * cell)
+{
+  GfsSolidVector * solid = GFS_STATE (cell)->solid;
+
+  if (solid) {
+    FttDirection d;
+    FttCellNeighbors n;
+
+    ftt_cell_neighbors (cell, &n);
+    for (d = 0; d < FTT_NEIGHBORS; d++)
+      if (n.c[d] && solid->s[d] > 0. && solid->a/solid->s[d] < 0.5)
+	return TRUE;
+  }
+  return FALSE;
+}
+
 static void set_merged (FttCell * cell)
 {
   GfsSolidVector * solid = GFS_STATE (cell)->solid;
 
-  if (!GFS_IS_SMALL (cell))
+  if (!is_small (cell))
     solid->merged = NULL;
   else {
     FttCellNeighbors neighbor;
@@ -631,11 +610,39 @@ static void set_merged (FttCell * cell)
     ftt_cell_neighbors (cell, &neighbor);
     for (i = 0; i < FTT_NEIGHBORS && abest < 1.; i++)
       if (neighbor.c[i] && !GFS_CELL_IS_BOUNDARY (neighbor.c[i]) && solid->s[i] > 0.) {
-	gdouble a = GFS_IS_MIXED (neighbor.c[i]) ? GFS_STATE (neighbor.c[i])->solid->a : 1.;
-	
-	if (a > abest) {
-	  abest = a;
-	  solid->merged = neighbor.c[i];
+	if (FTT_CELL_IS_LEAF (neighbor.c[i])) {
+	  if (GFS_IS_MIXED (neighbor.c[i])) {
+	    gdouble a = GFS_STATE (neighbor.c[i])->solid->a;
+	    
+	    if (a > abest) {
+	      abest = a;
+	      solid->merged = neighbor.c[i];
+	    }
+	  }
+	  else {
+	    solid->merged = neighbor.c[i];
+	    return;
+	  }
+	}
+	else {
+	  FttCellChildren child;
+	  guint j, n = ftt_cell_children_direction (neighbor.c[i], FTT_OPPOSITE_DIRECTION (i), &child);
+
+	  for (j = 0; j < n; j++)
+	    if (child.c[j]) {
+	      if (GFS_IS_MIXED (child.c[j])) {
+		gdouble a = GFS_STATE (child.c[j])->solid->a;
+	    
+		if (a > abest) {
+		  abest = a;
+		  solid->merged = child.c[j];
+		}
+	      }
+	      else {
+		solid->merged = child.c[j];
+		return;
+	      }
+	    }
 	}
       }
     if (abest == 0.)
@@ -670,7 +677,7 @@ static void add_merged (GSList ** merged, FttCell * cell)
     *merged = g_slist_prepend (*merged, cell);
     cell->flags |= GFS_FLAG_USED;
 
-    if (GFS_IS_MIXED (cell) && solid->merged)
+    if (solid && solid->merged)
       add_merged (merged, solid->merged);
 
     ftt_cell_neighbors (cell, &neighbor);
@@ -687,7 +694,7 @@ static void add_merged (GSList ** merged, FttCell * cell)
 	    if (GFS_IS_MIXED (child.c[j]) &&
 		GFS_STATE (child.c[j])->solid->merged == cell)
 	      add_merged (merged, child.c[j]);
-	}	
+	}
 	else if (GFS_IS_MIXED (neighbor.c[i]) && 
 		 GFS_STATE (neighbor.c[i])->solid->merged == cell)
 	  add_merged (merged, neighbor.c[i]);
@@ -771,6 +778,7 @@ void gfs_advection_update (GSList * merged, const GfsAdvectionParams * par)
 
     if (GFS_IS_MIXED (cell)) {
 #if 1
+      g_assert (!is_small (cell));
       GFS_VARIABLE (cell, par->v->i) += 
 	GFS_VARIABLE (cell, par->fv->i)/GFS_STATE (cell)->solid->a;
 #else /* D. Calhoun approach */
@@ -878,14 +886,15 @@ void gfs_advection_params_init (GfsAdvectionParams * par)
 {
   g_return_if_fail (par != NULL);
 
+  par->fv = NULL;
+  par->u = NULL;
+  par->g = NULL;
   par->cfl = 0.8;
   par->dt = 1.;
   par->gradient = gfs_center_gradient;
   par->upwinding = GFS_FACE_UPWINDING;
   par->use_centered_velocity = TRUE;
   par->scheme = GFS_GODUNOV;
-  par->rho = 1.;
-  par->c = NULL;
 }
 
 void gfs_advection_params_read (GfsAdvectionParams * par, GtsFile * fp)
@@ -943,13 +952,8 @@ void gfs_advection_params_read (GfsAdvectionParams * par, GtsFile * fp)
       par->scheme = GFS_GODUNOV;
     else if (!strcmp (scheme, "none"))
       par->scheme = GFS_NONE;
-    else if (!strcmp (scheme, "vof")) {
+    else if (!strcmp (scheme, "vof"))
       par->scheme = GFS_VOF;
-      if (fp->type != GTS_ERROR && (par->cfl <= 0. || par->cfl > 0.5))
-	gts_file_variable_error (fp, var, "cfl", 
-				 "cfl `%g' is out of range `]0,0.5]'", 
-				 par->cfl);
-    }
     else if (fp->type != GTS_ERROR)
       gts_file_variable_error (fp, var, "scheme",
 			       "unknown scheme parameter `%s'", scheme);
