@@ -37,8 +37,7 @@ static FttVector rpos[FTT_NEIGHBORS] = {
 
 static void symmetry (FttCellFace * f, GfsBc * b)
 {
-  if (b->v->i == GFS_VELOCITY_INDEX (f->d/2) ||
-      b->v->i == GFS_GRADIENT_INDEX (f->d/2))
+  if (b->v->component == f->d/2)
     GFS_VARIABLE (f->cell, b->v->i) = - GFS_VARIABLE (f->neighbor, b->v->i);
   else
     GFS_VARIABLE (f->cell, b->v->i) =   GFS_VARIABLE (f->neighbor, b->v->i);
@@ -46,8 +45,7 @@ static void symmetry (FttCellFace * f, GfsBc * b)
 
 static void face_symmetry (FttCellFace * f, GfsBc * b)
 {
-  if (b->v->i == GFS_VELOCITY_INDEX (f->d/2) ||
-      b->v->i == GFS_GRADIENT_INDEX (f->d/2))
+  if (b->v->component == f->d/2)
     GFS_STATE (f->cell)->f[f->d].v = 
       GFS_STATE (f->neighbor)->f[FTT_OPPOSITE_DIRECTION (f->d)].v = 0.;
   else
@@ -154,7 +152,7 @@ static void bc_value_read (GtsObject ** o, GtsFile * fp)
     return;
   if (bc->val == NULL)
     bc->val = gfs_function_new (gfs_function_class (), 0.);
-  gfs_function_read (GFS_BC_VALUE (*o)->val, fp);
+  gfs_function_read (GFS_BC_VALUE (*o)->val, gfs_box_domain (GFS_BC (bc)->b->box), fp);
 }
 
 static void bc_value_destroy (GtsObject * o)
@@ -213,8 +211,7 @@ GfsBc * gfs_bc_value_new (GfsBcClass * k,
 static void dirichlet (FttCellFace * f, GfsBc * b)
 {
   GFS_VARIABLE (f->cell, b->v->i) = 
-    2.*gfs_function_face_value (GFS_BC_VALUE (b)->val, f, 
-		     GFS_SIMULATION (gfs_box_domain (b->b->box))->time.t)
+    2.*gfs_function_face_value (GFS_BC_VALUE (b)->val, f)
     - GFS_VARIABLE (f->neighbor, b->v->i);
 }
 
@@ -225,11 +222,8 @@ static void homogeneous_dirichlet (FttCellFace * f, GfsBc * b)
 
 static void face_dirichlet (FttCellFace * f, GfsBc * b)
 {
-  gdouble v = gfs_function_face_value (GFS_BC_VALUE (b)->val, f,
-		   GFS_SIMULATION (gfs_box_domain (b->b->box))->time.t);
-  
   GFS_STATE (f->cell)->f[f->d].v = 
-    GFS_STATE (f->neighbor)->f[FTT_OPPOSITE_DIRECTION (f->d)].v = v;
+    gfs_function_face_value (GFS_BC_VALUE (b)->val, f);
 }
 
 static void gfs_bc_dirichlet_init (GfsBc * object)
@@ -266,8 +260,7 @@ static void neumann (FttCellFace * f, GfsBc * b)
 {
   GFS_VARIABLE (f->cell, b->v->i) = 
     GFS_VARIABLE (f->neighbor, b->v->i) +
-    gfs_function_face_value (GFS_BC_VALUE (b)->val, f, 
-		  GFS_SIMULATION (gfs_box_domain (b->b->box))->time.t)
+    gfs_function_face_value (GFS_BC_VALUE (b)->val, f)
     *ftt_cell_size (f->cell);
 }
 
@@ -278,14 +271,10 @@ static void homogeneous_neumann (FttCellFace * f, GfsBc * b)
 
 static void face_neumann (FttCellFace * f, GfsBc * b)
 {
-  gdouble v = 
-    GFS_VARIABLE (f->neighbor, b->v->i) +
-    gfs_function_face_value (GFS_BC_VALUE (b)->val, f, 
-	         GFS_SIMULATION (gfs_box_domain (b->b->box))->time.t)
-    *ftt_cell_size (f->cell)/2.;
-
   GFS_STATE (f->cell)->f[f->d].v = 
-    GFS_STATE (f->neighbor)->f[FTT_OPPOSITE_DIRECTION (f->d)].v = v;
+    GFS_VARIABLE (f->neighbor, b->v->i) +
+    gfs_function_face_value (GFS_BC_VALUE (b)->val, f)
+    *ftt_cell_size (f->cell)/2.;
 }
 
 static void gfs_bc_neumann_init (GfsBc * object)
@@ -344,11 +333,17 @@ static void gfs_boundary_destroy (GtsObject * object)
 static void match (FttCell * cell, GfsBoundary * boundary)
 {
   FttCell * neighbor = ftt_cell_neighbor (cell, boundary->d);
+  FttCell * parent = ftt_cell_parent (cell);
   guint level = ftt_cell_level (cell);
 
   cell->flags |= GFS_FLAG_BOUNDARY;
+  if (parent && GFS_CELL_IS_GRADIENT_BOUNDARY (parent))
+    cell->flags |= GFS_FLAG_GRADIENT_BOUNDARY;
   if (neighbor == NULL || ftt_cell_level (neighbor) < level) {
-    g_assert (!FTT_CELL_IS_ROOT (cell));
+    if (FTT_CELL_IS_ROOT (cell))
+      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_ERROR,
+	     "root cell is entirely outside of the fluid domain\n"
+	     "the solid surface orientation may be incorrect");
     ftt_cell_destroy (cell, (FttCellCleanupFunc) gfs_cell_cleanup, NULL);
     boundary->changed = TRUE;
     return;
@@ -357,7 +352,10 @@ static void match (FttCell * cell, GfsBoundary * boundary)
     GfsSolidVector * s = GFS_STATE (neighbor)->solid;
 
     if (s && s->s[FTT_OPPOSITE_DIRECTION (boundary->d)] == 0.) {
-      g_assert (!FTT_CELL_IS_ROOT (cell));
+      if (FTT_CELL_IS_ROOT (cell))
+	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_ERROR,
+	       "root cell is entirely outside of the fluid domain\n"
+	       "the solid surface orientation may be incorrect");
       ftt_cell_destroy (cell, (FttCellCleanupFunc) gfs_cell_cleanup, NULL);
       boundary->changed = TRUE;
       return;
@@ -380,7 +378,7 @@ static void match (FttCell * cell, GfsBoundary * boundary)
 	if (c == boundary->d/2) {
 	  FttVector p1, p2;
 	  ftt_cell_pos (cell, &p1);
-	  ftt_cell_pos (cell, &p2);
+	  ftt_cell_pos (neighbor, &p2);
 	  (&t->cm.x)[c] = (&p1.x)[c] + (&p2.x)[c] - (&s->cm.x)[c];
 	  (&t->ca.x)[c] = (&p1.x)[c] + (&p2.x)[c] - (&s->ca.x)[c];
 	}
@@ -742,8 +740,8 @@ static void inflow_constant_read (GtsObject ** o, GtsFile * fp)
 {
   GfsBoundary * b = GFS_BOUNDARY (*o);  
   FttComponent c;
-  GfsVariable * v;
   GfsFunction * un = GFS_BOUNDARY_INFLOW_CONSTANT (*o)->un;
+  GfsVariable ** v;
 
   if (GTS_OBJECT_CLASS (gfs_boundary_inflow_constant_class ())->parent_class->read)
     (* GTS_OBJECT_CLASS (gfs_boundary_inflow_constant_class ())->parent_class->read) 
@@ -751,16 +749,16 @@ static void inflow_constant_read (GtsObject ** o, GtsFile * fp)
   if (fp->type == GTS_ERROR)
     return;
 
-  gfs_function_read (un, fp);
+  gfs_function_read (un, gfs_box_domain (b->box), fp);
 
-  v = gfs_variable_from_name (gfs_box_domain (b->box)->variables, "U");
-  for (c = 0; c < FTT_DIMENSION; c++, v = v->next)
+  v = gfs_domain_velocity (gfs_box_domain (b->box));
+  for (c = 0; c < FTT_DIMENSION; c++)
     if (c == b->d/2)
       gfs_boundary_add_bc (b, gfs_bc_value_new (gfs_bc_dirichlet_class (),
-						v, un, FALSE));
+						v[c], un, FALSE));
     else
       gfs_boundary_add_bc (b, gfs_bc_value_new (gfs_bc_dirichlet_class (),
-						v, NULL, FALSE));
+						v[c], NULL, FALSE));
 }
 
 static void gfs_boundary_inflow_constant_class_init (GtsObjectClass * klass)
@@ -808,9 +806,9 @@ static GtsColor outflow_color (GtsObject * o)
 
 static void outflow_read (GtsObject ** o, GtsFile * fp)
 {
-  GfsBoundary * b = GFS_BOUNDARY (*o);  
-  FttComponent c;
-  GfsVariable * v;
+  GfsBoundary * b = GFS_BOUNDARY (*o);
+  GfsDomain * domain;
+  GfsVariable ** v;
 
   if (GTS_OBJECT_CLASS (gfs_boundary_outflow_class ())->parent_class->read)
     (* GTS_OBJECT_CLASS (gfs_boundary_outflow_class ())->parent_class->read) 
@@ -818,14 +816,14 @@ static void outflow_read (GtsObject ** o, GtsFile * fp)
   if (fp->type == GTS_ERROR)
     return;
 
-  v = gfs_variable_from_name (gfs_box_domain (b->box)->variables, "U");
-  for (c = 0; c < b->d/2; c++, v = v->next) 
-    ;
-
-  gfs_boundary_add_bc (b, gfs_bc_value_new (gfs_bc_neumann_class (), v, NULL,
-					    FALSE));
+  domain = gfs_box_domain (b->box);
+  v = gfs_domain_velocity (domain);
+  gfs_boundary_add_bc (b, gfs_bc_value_new (gfs_bc_neumann_class (),
+					    v[b->d/2],
+					    NULL, FALSE));
   gfs_boundary_add_bc (b, gfs_bc_value_new (gfs_bc_dirichlet_class (),
-					    gfs_p, NULL, FALSE));
+					    gfs_variable_from_name (domain->variables, "P"),
+					    NULL, FALSE));
 }
 
 static void gfs_boundary_outflow_class_init (GfsBoundaryClass * klass)
