@@ -22,6 +22,7 @@
 
 #include <stdio.h>
 #include <string.h>
+#include <math.h>
 #include <glib.h>
 #include <gts.h>
 
@@ -50,6 +51,12 @@ struct _FttVector {
   gdouble x, y, z;
 };
 
+#if FTT_2D
+# define ftt_vector_norm(v) (sqrt((v)->x*(v)->x + (v)->y*(v)->y))
+#else  /* 3D */
+# define ftt_vector_norm(v) (sqrt((v)->x*(v)->x + (v)->y*(v)->y + (v)->z*(v)->z))
+#endif /* 3D */
+
 typedef enum
 {
   FTT_TRAVERSE_LEAFS          = 1 << 0,
@@ -382,17 +389,17 @@ FttCell * ftt_cell_child_corner (const FttCell * cell,
   g_return_val_if_fail (cell != NULL, NULL);
   g_return_val_if_fail (!FTT_CELL_IS_LEAF (cell), NULL);
 
-  g_assert (d[0] < FTT_NEIGHBORS);
-  g_assert (d[1] < FTT_NEIGHBORS);
-
   g_return_val_if_fail (d[0] < FTT_NEIGHBORS, NULL);
   g_return_val_if_fail (d[1] < FTT_NEIGHBORS, NULL);
-#  if FTT_2D3
-  if (d[0] >= FTT_NEIGHBORS_2D || d[1] >= FTT_NEIGHBORS_2D)
-    g_assert_not_implemented ();
-#  endif
 
-  i = index[d[0]][d[1]];
+#  if FTT_2D3
+  if (d[0] >= FTT_NEIGHBORS_2D)
+    i = index[d[1]][d[2]];
+  else if (d[1] >= FTT_NEIGHBORS_2D)
+    i = index[d[0]][d[2]];
+  else
+#  endif
+    i = index[d[0]][d[1]];
 #else  /* FTT_3D */
   static gint index[FTT_NEIGHBORS][FTT_NEIGHBORS][FTT_NEIGHBORS] = {
     {{-1,-1,-1,-1,-1,-1},{-1,-1,-1,-1,-1,-1},
@@ -571,7 +578,7 @@ FttCellFace ftt_cell_face (FttCell * cell,
 {
   FttCellFace f;
 
-  g_return_val_if_fail (cell != NULL, f);
+  g_assert (cell != NULL);
 
   f.cell = cell;
   f.neighbor = ftt_cell_neighbor (cell, d);
@@ -664,6 +671,15 @@ void                 ftt_cell_traverse               (FttCell * root,
 						      gint max_depth,
 						      FttCellTraverseFunc func,
 						      gpointer data);
+void                 ftt_cell_traverse_condition     (FttCell * root,
+						      FttTraverseType order,
+						      FttTraverseFlags flags,
+						      gint max_depth,
+						      FttCellTraverseFunc func,
+						      gpointer data,
+						      gboolean (* condition) (FttCell *, 
+									      gpointer),
+						      gpointer cdata);
 void                 ftt_cell_traverse_box           (FttCell * root,
 						      GtsBBox * box,
 						      FttTraverseType order,
@@ -697,6 +713,26 @@ void                 ftt_face_traverse_boundary      (FttCell * root,
 FttCell *            ftt_cell_locate                 (FttCell * root,
 						      FttVector target,
 						      gint max_depth);
+gdouble              ftt_cell_point_distance2_min    (FttCell * cell, 
+						      GtsPoint * p);
+void                 ftt_cell_point_distance2_internal (FttCell * root,
+							GtsPoint * p,
+							gdouble d,
+							gdouble (* distance2) (FttCell *, 
+									       GtsPoint *, 
+									       gpointer),
+							gpointer data,
+							FttCell ** closest,
+							gdouble * dmin);
+gdouble              ftt_cell_point_distance2        (FttCell * root,
+						      GtsPoint * p,
+						      gdouble (* distance2) (FttCell *, 
+									     GtsPoint *, 
+									     gpointer),
+						      gpointer data,
+						      FttCell ** closest);
+void                 ftt_cell_bbox                   (const FttCell * cell, 
+						      GtsBBox * bb);
 typedef void      (* FttCellCopyFunc)                (const FttCell * from,
 						      FttCell * to,
 						      gpointer data);
