@@ -6,7 +6,7 @@
 #ifndef GFSCONFIG_H
 #define GFSCONFIG_H
 
-#define GFS_MODULES_DIR "/home/popinet/adds//lib/gerris"
+#define GFS_MODULES_DIR "/home/popinet/adds/lib/gerris"
 #define GFS_HAS_MODULES 1
 
 
