@@ -24,6 +24,7 @@
 #include "config.h"
 #include "graphic.h"
 #include "solid.h"
+#include "variable.h"
 
 typedef struct {
   GPtrArray * colors;
@@ -334,172 +335,6 @@ void gfs_write_gts (GfsDomain * domain,
   gts_matrix_destroy (transform);
 }
 
-static void write_vertex (GtsPoint * p, gpointer * data)
-{
-  guint * index = data[0];
-  FILE * fp = data[1];
-
-  fprintf (fp, "%g %g %g\n", p->x, p->y, p->z);
-  VERTEX_CELL_FACE (p)->index = (*index)++;
-}
-
-static void write_face (GtsTriangle * t, FILE * fp)
-{
-  GtsVertex * v1, * v2, * v3;
-
-  gts_triangle_vertices (t, &v1, &v2, &v3);
-  fprintf (fp, "3 %u %u %u\n",
-	   VERTEX_CELL_FACE (v1)->index, 
-	   VERTEX_CELL_FACE (v2)->index,
-	   VERTEX_CELL_FACE (v3)->index);
-}
-
-static gdouble variable_value (VertexCellFace * p, GfsVariable * v)
-{
-  if (p->cell)
-    return GFS_VARIABLE (p->cell, v->i);
-  else if (p->face.neighbor)
-    return gfs_face_interpolated_value (&p->face, v->i);
-  else
-    return GFS_VARIABLE (p->face.cell, v->i);
-}
-
-static void write_variable (VertexCellFace * p, gpointer * data)
-{
-  fprintf (data[1], "%g\n", variable_value (p, data[0]));
-}
-
-static void write_vector (VertexCellFace * p, gpointer * data)
-{
-  GfsVariable * v = data[0];
-
-#if FTT_2D
-  fprintf (data[1], "%g %g 0\n", 
-	   variable_value (p, v), 
-	   variable_value (p, v->next));
-#else  /* 3D */
-  fprintf (data[1], "%g %g %g\n",
-	   variable_value (p, v), 
-	   variable_value (p, v->next),
-	   variable_value (p, v->next->next));
-#endif /* 3D */
-}
-
-void gfs_write_vtk (GfsDomain * domain,
-		    FttTraverseFlags flags,
-		    gint level,
-		    FILE * fp)
-{
-  GtsSurface * s;
-  GtsVertex * v1, * v2, * v3;
-  GtsEdge * e1, * e2, * e3;
-  gpointer data[6] = { NULL, NULL, NULL, NULL, NULL };
-  guint nvertices, nfaces, index = 0, i;
-  GfsVariable * v;
-  GSList * long_segments = NULL;
-  FttDirection d;
-
-  g_return_if_fail (domain != NULL);
-  g_return_if_fail (fp != NULL);
-
-  v1 = gts_vertex_new (gts_vertex_class (), -100., -100., 0.);
-  v2 = gts_vertex_new (gts_vertex_class (), 100., -100., 0.);
-  v3 = gts_vertex_new (gts_vertex_class (), 0., 100., 0.);
-  e1 = gts_edge_new (gts_edge_class (), v1, v2);
-  e2 = gts_edge_new (gts_edge_class (), v2, v3);
-  e3 = gts_edge_new (gts_edge_class (), v3, v1);
-  s = gts_surface_new (gts_surface_class (), 
-		       gts_face_class (), 
-		       gts_edge_class (), 
-		       vertex_cell_face_class ());
-  gts_surface_add_face (s, gts_face_new (gts_face_class (), e1, e2, e3));
-
-  data[0] = s;
-  gfs_domain_cell_traverse (domain, FTT_PRE_ORDER, flags, level, 
-			    (FttCellTraverseFunc) triangulate, data);
-  data[5] = &d;
-  for (d = 0; d < FTT_NEIGHBORS; d++)
-    gfs_domain_cell_traverse_boundary (domain, d, FTT_PRE_ORDER, flags, level, 
-				       (FttCellTraverseFunc) triangulate_face,
-				       data);
-
-  gts_allow_floating_vertices = TRUE;
-  gts_object_destroy (GTS_OBJECT (v1));
-  gts_object_destroy (GTS_OBJECT (v2));
-  gts_object_destroy (GTS_OBJECT (v3));
-  gts_allow_floating_vertices = FALSE;
-
-  gts_surface_foreach_edge (s, (GtsFunc) add_long_segment, &long_segments);
-  gts_allow_floating_edges = TRUE;
-  g_slist_foreach (long_segments, (GFunc) gts_object_destroy, NULL);
-  gts_allow_floating_edges = FALSE;
-  g_slist_free (long_segments);
-
-  nvertices = gts_surface_vertex_number (s);
-  fprintf (fp, 
-	   "# vtk DataFile Version 2.0\n"
-	   "Generated by Gerris\n"
-           "ASCII\n"
-	   "DATASET UNSTRUCTURED_GRID\n"
-	   "POINTS %u float\n",
-	   nvertices);
-  data[0] = &index;
-  data[1] = fp;
-  gts_surface_foreach_vertex (s, (GtsFunc) write_vertex, data);
-  nfaces = gts_surface_face_number (s);
-  fprintf (fp, "CELLS %u %u\n", nfaces, 4*nfaces);
-  gts_surface_foreach_face (s, (GtsFunc) write_face, fp);
-  fprintf (fp, "CELL_TYPES %u\n", nfaces);
-  for (i = 0; i < nfaces; i++)
-    fputs ("5\n", fp);
-  fprintf (fp, "POINT_DATA %u\n", nvertices);
-  v = domain->variables;
-  while (v) {
-    guint d;
-
-    if (!strcmp (v->name, "U") && v->next &&
-	!strcmp (v->next->name, "V")
-#if (!FTT_2D)
-	&& v->next->next && !strcmp (v->next->next->name, "W")
-#endif /* not FTT_2D */
-	) {
-      fputs ("VECTORS U float\n", fp);
-      data[0] = v;
-      data[1] = fp;
-      gts_surface_foreach_vertex (s, (GtsFunc) write_vector, data);
-      for (d = 0; d < FTT_DIMENSION; d++) v = v->next;
-    }
-    else if (v->name[strlen (v->name) - 1] == 'x' && v->next &&
-	     v->next->name[strlen (v->name) - 1] == 'y'
-#if (!FTT_2D)
-	     && v->next->next &&
-	     v->next->next->name[strlen (v->name) - 1] == 'z'
-#endif /* not FTT_2D */
-	     ) {
-      gchar * name = g_strndup (v->name, strlen (v->name) - 1);
-
-      fprintf (fp, "VECTORS %s float\n", name);
-      g_free (name);
-      data[0] = v;
-      data[1] = fp;
-      gts_surface_foreach_vertex (s, (GtsFunc) write_vector, data);
-      for (d = 0; d < FTT_DIMENSION; d++) v = v->next;
-    }
-    else {
-      fprintf (fp, 
-	       "SCALARS %s float 1\n"
-	       "LOOKUP_TABLE default\n",
-	       v->name);
-      data[0] = v;
-      data[1] = fp;
-      gts_surface_foreach_vertex (s, (GtsFunc) write_variable, data);
-      v = v->next;
-    }
-  }
-	 
-  gts_object_destroy (GTS_OBJECT (s));
-}
-
 static void extent (FttCell * cell, gpointer * data)
 {
   FttVector * min = data[0];
@@ -1618,12 +1453,13 @@ static void edge_list (GtsMatrix * b, GtsPoint * o,
 {
   guint i;
   GtsVertex * vold = NULL;
+  GtsVertex * vfirst = NULL;
   gboolean colored = FALSE;
   GtsMatrix * c;
 
   if (GTS_IS_COLORED_VERTEX (o) && 
       gts_object_class_is_from_class (GTS_OBJECT_CLASS (s->vertex_class),
-			       GTS_OBJECT_CLASS (gts_colored_vertex_class ())))
+				      GTS_OBJECT_CLASS (gts_colored_vertex_class ())))
     colored = TRUE;
   if (GFS_IS_TWISTED_VERTEX (o)) {
     gdouble t = GFS_TWISTED_VERTEX (o)->theta;
@@ -1642,7 +1478,7 @@ static void edge_list (GtsMatrix * b, GtsPoint * o,
 			b[2][0], b[2][1], b[2][2], 0.,
 			0., 0., 0., 0.);
 
-  for (i = 0; i <= ne; i++, profile = profile->next) {
+  for (i = 0; i <= ne && profile; i++, profile = profile->next) {
     GtsPoint * p = profile->data;
     GtsVertex * v = gts_vertex_new (s->vertex_class, p->x, p->y, 0.);
 
@@ -1653,11 +1489,14 @@ static void edge_list (GtsMatrix * b, GtsPoint * o,
     GTS_POINT (v)->x += o->x;
     GTS_POINT (v)->y += o->y;
     GTS_POINT (v)->z += o->z;
+
     if (vold != NULL)
       e[i-1] = gts_edge_new (s->edge_class, vold, v);
     vold = v;
+    if (!vfirst) vfirst = v;
   }
-
+  if (i <= ne && vold)
+    e[i-1] = gts_edge_new (s->edge_class, vold, vfirst);
   gts_matrix_destroy (c);
 }
 
@@ -1685,10 +1524,10 @@ static void add_face (GtsSurface * s, GtsEdge ** e1, GtsEdge ** e2,
   }
 }
 
-static GSList * next_far_enough (GSList * p, gdouble size)
+static GList * next_far_enough (GList * p, gdouble size)
 {
   GtsPoint * ps;
-  GSList * pf = NULL;
+  GList * pf = NULL;
 
   if (p == NULL)
     return NULL;
@@ -1703,9 +1542,10 @@ static GSList * next_far_enough (GSList * p, gdouble size)
   return pf;
 }
 
-static void gts_extrude_profile (GtsSurface * s,
-				 GSList * profile,
-				 GSList * path)
+void gfs_extrude_profile (GtsSurface * s,
+			  GSList * profile,
+			  gboolean closed,
+			  GList * path)
 {
   GtsMatrix * r;
   GtsPoint * p0, * p1, * p2;
@@ -1724,7 +1564,7 @@ static void gts_extrude_profile (GtsSurface * s,
     size = bb->y2 - bb->y1;
   gts_object_destroy (GTS_OBJECT (bb));
 
-size /= 4.;
+  size /= 4.;
 
   p0 = path->data;
   path = next_far_enough (path, size);
@@ -1732,7 +1572,7 @@ size /= 4.;
     return;
   p1 = path->data;
   r = gts_matrix_identity (NULL);
-  ne = g_slist_length (profile) - 1;
+  ne = closed ? g_slist_length (profile) : g_slist_length (profile) - 1;
   e1 = g_malloc (sizeof (GtsEdge *)*ne);
   e2 = g_malloc (sizeof (GtsEdge *)*ne);
 
@@ -1756,9 +1596,10 @@ size /= 4.;
 
   g_free (e1);
   g_free (e2);
+  gts_matrix_destroy (r);
 }
 
-static gdouble curve_cost (FttVector p1, FttVector p2, FttVector p3)
+static gdouble triangle_area (FttVector p1, FttVector p2, FttVector p3)
 {
   GtsVector v1, v2, a;
 
@@ -1768,19 +1609,31 @@ static gdouble curve_cost (FttVector p1, FttVector p2, FttVector p3)
   return gts_vector_norm (a)/2.;
 }
 
+static gdouble circumcircle_radius (FttVector p1, FttVector p2, FttVector p3)
+{
+  gdouble area = triangle_area (p1, p2, p3);
+
+  if (area == 0.)
+    return G_MAXDOUBLE;
+  else {
+    GtsVector a, b, c;
+    gts_vector_init (a, &p1, &p2);
+    gts_vector_init (b, &p2, &p3);
+    gts_vector_init (c, &p3, &p1);
+    return gts_vector_norm (a)*gts_vector_norm (b)*gts_vector_norm (c)/area;
+  }
+}
+
 static GSList * circle_profile (GtsPointClass * klass, 
 				gdouble radius, guint np)
 {
   GSList * lp = NULL;
   guint i;
 
-  for (i = 0; i <= np; i++) {
+  for (i = 0; i < np; i++) {
     gdouble a = 2.*M_PI*i/(gdouble) np;
 
-    lp = g_slist_prepend (lp, gts_point_new (klass, 
-					     radius*cos (a), 
-					     radius*sin(a), 
-					     0.));
+    lp = g_slist_prepend (lp, gts_point_new (klass, radius*cos (a), radius*sin(a), 0.));
   }
   return lp;
 }
@@ -1796,69 +1649,80 @@ static GSList * ribbon_profile (GtsPointClass * klass,
 }
 
 #if (!FTT_2D)
-static void vorticity_vector (FttCell * cell)
+static void vorticity_vector (FttCell * cell, gpointer * data)
 {
   gdouble size = ftt_cell_size (cell);
+  GfsVariable ** g = data[0];
+  GfsVariable ** v = data[1];
 
-  GFS_STATE (cell)->g[0] = (gfs_center_gradient (cell, FTT_Y, GFS_W) -
-			    gfs_center_gradient (cell, FTT_Z, GFS_V))/size;
-  GFS_STATE (cell)->g[1] = (gfs_center_gradient (cell, FTT_Z, GFS_U) -
-			    gfs_center_gradient (cell, FTT_X, GFS_W))/size;
-  GFS_STATE (cell)->g[2] = (gfs_center_gradient (cell, FTT_X, GFS_V) -
-			    gfs_center_gradient (cell, FTT_Y, GFS_U))/size;
+  GFS_VARIABLE (cell, g[0]->i) = (gfs_center_gradient (cell, FTT_Y, v[2]->i) -
+				  gfs_center_gradient (cell, FTT_Z, v[1]->i))/size;
+  GFS_VARIABLE (cell, g[1]->i) = (gfs_center_gradient (cell, FTT_Z, v[0]->i) -
+				  gfs_center_gradient (cell, FTT_X, v[2]->i))/size;
+  GFS_VARIABLE (cell, g[2]->i) = (gfs_center_gradient (cell, FTT_X, v[1]->i) -
+				  gfs_center_gradient (cell, FTT_Y, v[0]->i))/size;
 }
 #endif /* 3D */
 
-static GSList * grow_curve (GfsDomain * domain,
-			    FttVector p,
-			    GfsVariable * var,
-			    gdouble min, 
-			    gdouble max,
-			    gboolean twist,
-			    GSList * path,
-			    gdouble direction)
+static GList * grow_curve (GfsDomain * domain,
+			   GfsVariable ** U,
+			   FttVector p,
+			   GfsVariable * var,
+			   gdouble min, 
+			   gdouble max,
+			   gboolean twist,
+			   GList * path,
+			   gdouble direction,
+			   gboolean (* stop) (FttCell *, GList *, gpointer),
+			   gpointer data)
 {
   FttCell * cell;
-  gdouble delta = 0.1;
+  gdouble delta = 0.2;
   GtsPoint * oldp = NULL;
   FttVector p1, p2;
   gdouble cost = 0., theta = 0.;
-  gdouble maxcost = 2e-9;
+  gdouble maxcost = 4e-9;
   guint nstep = 0, nmax = 10000;
   GtsPointClass * path_class = gfs_vertex_class ();
   Colormap * colormap = NULL;
-  GfsVariable * U;
+#if (!FTT_2D)
+  GfsVariable * vort[FTT_DIMENSION];
+#endif /* 3D */
 
   if (min < max)
     colormap = colormap_jet ();
 
 #if (!FTT_2D)
   if (twist) {
-    path_class = GTS_POINT_CLASS (gfs_twisted_vertex_class ());
-    {
-      FttComponent c;
-      GfsVariable * g;
+    FttComponent c;
+    gpointer data[2];
 
-      gfs_domain_cell_traverse (domain, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
-			     (FttCellTraverseFunc) vorticity_vector, NULL);
-      for (c = 0, g = gfs_gx; c < FTT_DIMENSION; c++, g = g->next)
-	gfs_domain_cell_traverse (domain,
-				  FTT_POST_ORDER, FTT_TRAVERSE_NON_LEAFS, -1,
-				  (FttCellTraverseFunc) g->fine_coarse, g);
+    path_class = GTS_POINT_CLASS (gfs_twisted_vertex_class ());
+    for (c = 0; c < FTT_DIMENSION; c++) {
+      vort[c] = gfs_temporary_variable (domain);
+      gfs_variable_set_vector (vort[c], c);
     }
+    data[0] = vort;
+    data[1] = U;
+    gfs_domain_cell_traverse (domain, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
+			      (FttCellTraverseFunc) vorticity_vector, data);
+    for (c = 0; c < FTT_DIMENSION; c++)
+      gfs_domain_cell_traverse (domain,
+				FTT_POST_ORDER, FTT_TRAVERSE_NON_LEAFS, -1,
+				(FttCellTraverseFunc) vort[c]->fine_coarse, vort[c]);
   }
 #endif /* 3D */  
 
-  U = gfs_variable_from_name (domain->variables, "U");
   p1 = p2 = p;
-  while ((cell = gfs_domain_locate (domain, p, -1)) != NULL && nmax--) {
+  while ((cell = gfs_domain_locate (domain, p, -1)) != NULL &&
+	 circumcircle_radius (p1, p2, p) > ftt_cell_size (cell) &&
+	 nmax--) {
     gdouble h = delta*ftt_cell_size (cell);
     FttVector u;
     FttComponent c;
-    GfsVariable * v;
     gdouble nu = 0.;
 
-    cost += curve_cost (p1, p2, p);
+    cost += triangle_area (p1, p2, p);
     p1 = p2;
     p2 = p;
     if (oldp == NULL || cost > maxcost) {
@@ -1870,15 +1734,35 @@ static GSList * grow_curve (GfsDomain * domain,
 	  colormap_color (colormap, (GFS_VERTEX (oldp)->v - min)/(max - min));
       if (twist)
 	GFS_TWISTED_VERTEX (oldp)->theta = theta;
-      path = g_slist_prepend (path, oldp);
+      path = g_list_prepend (path, oldp);
+      if (stop != NULL && (* stop) (cell, path, data))
+	break;
       cost = 0.;
       nstep = 0;
     }
 
-    for (c = 0, v = U; c < FTT_DIMENSION; c++, v = v->next) {
-      ((gdouble *) &u)[c] = direction*gfs_interpolate (cell, p, v);
-      nu += ((gdouble *) &u)[c]*((gdouble *) &u)[c];
+    for (c = 0; c < FTT_DIMENSION; c++) {
+      (&u.x)[c] = direction*gfs_interpolate (cell, p, U[c]);
+      nu += (&u.x)[c]*(&u.x)[c];
     }
+    if (nu > 0) {
+      FttVector p1 = p;
+      FttCell * cell1;
+
+      nu = 2.*sqrt (nu);
+      for (c = 0; c < FTT_DIMENSION; c++)
+	(&p1.x)[c] += h*(&u.x)[c]/nu;
+      nu = 0.;
+      cell1 = gfs_domain_locate (domain, p1, -1);
+      if (!cell1)
+	break;
+      for (c = 0; c < FTT_DIMENSION; c++) {
+	(&u.x)[c] = direction*gfs_interpolate (cell1, p1, U[c]);
+	nu += (&u.x)[c]*(&u.x)[c];
+      }
+    }
+    else
+      break;
     if (nu > 0. && nstep++ < nmax) {
       FttVector p1;
 
@@ -1892,8 +1776,8 @@ static GSList * grow_curve (GfsDomain * domain,
 	GtsVector dx;
 
 	dx[0] = p1.x - p.x; dx[1] = p1.y - p.y; dx[2] = p1.z - p.z;
-	for (c = 0, v = gfs_gx; c < FTT_DIMENSION; c++, v = v->next)
-	  rot[c] = gfs_interpolate (cell, p1, v);
+	for (c = 0; c < FTT_DIMENSION; c++)
+	  rot[c] = gfs_interpolate (cell, p1, vort[c]);
 	theta += gts_vector_scalar (rot, dx)/nu;
       }
 #endif /* 3D */
@@ -1909,35 +1793,52 @@ static GSList * grow_curve (GfsDomain * domain,
 	GFS_VERTEX (oldp)->v = gfs_interpolate (cell, p2, var);
       if (twist)
 	GFS_TWISTED_VERTEX (oldp)->theta = theta;
-      path = g_slist_prepend (path, oldp);
+      path = g_list_prepend (path, oldp);
     }
   }
 
   if (colormap)
     colormap_destroy (colormap);
 
-  return direction > 0. ? g_slist_reverse (path) : path;
+#if (!FTT_2D)
+  if (twist) {
+    FttComponent c;
+    for (c = 0; c < FTT_DIMENSION; c++)
+      gts_object_destroy (GTS_OBJECT (vort[c]));
+  }
+#endif /* 3D */
+  return direction > 0. ? g_list_reverse (path) : path;
 }
 
-GSList * gfs_streamline_new (GfsDomain * domain,
-			     FttVector p,
-			     GfsVariable * var,
-			     gdouble min, 
-			     gdouble max,
-			     gboolean twist)
+GList * gfs_streamline_new (GfsDomain * domain,
+			    GfsVariable ** U,
+			    FttVector p,
+			    GfsVariable * var,
+			    gdouble min,
+			    gdouble max,
+			    gboolean twist,
+			    gboolean (* stop) (FttCell *, 
+					       GList *,
+					       gpointer),
+			    gpointer data)
 {
-  GSList * path;
+  GList * i, * path;
 
-  path = grow_curve (domain, p, var, min, max, twist, NULL, 1.);
-  path = grow_curve (domain, p, var, min, max, twist, path, -1.);
+  g_return_val_if_fail (domain != NULL, NULL);
+  g_return_val_if_fail (U != NULL, NULL);
+
+  i = grow_curve (domain, U, p, var, min, max, twist, NULL, 1., stop, data);
+  path = g_list_remove_link (i, i);
+  g_list_free (i);
+  path = grow_curve (domain, U, p, var, min, max, twist, path, -1., stop, data);
   return path;
 }
 
-void gfs_streamline_write (GSList * stream, FILE * fp)
+void gfs_streamline_write (GList * stream, FILE * fp)
 {
   g_return_if_fail (fp != NULL);
 
-  fprintf (fp, "GfsStreamline %u\n", g_slist_length (stream));
+  fprintf (fp, "GfsStreamline %u\n", g_list_length (stream));
   while (stream) {
     (* GTS_OBJECT (stream->data)->klass->write) (stream->data, fp);
     fputc ('\n', fp);
@@ -1945,9 +1846,9 @@ void gfs_streamline_write (GSList * stream, FILE * fp)
   }
 }
 
-GSList * gfs_streamline_read (GtsFile * fp)
+GList * gfs_streamline_read (GtsFile * fp)
 {
-  GSList * stream = NULL;
+  GList * stream = NULL;
   guint n = 0, nv;
 
   g_return_val_if_fail (fp != NULL, NULL);
@@ -1970,21 +1871,21 @@ GSList * gfs_streamline_read (GtsFile * fp)
 
     (*o->klass->read) (&o, fp);
     gts_file_first_token_after (fp, '\n');
-    stream = g_slist_prepend (stream, o);
+    stream = g_list_prepend (stream, o);
     n++;
   }
 
   if (fp->type == GTS_ERROR) {
-    g_slist_free (stream);
+    g_list_free (stream);
     return NULL;
   }
 
   return stream;
 }
 
-void gfs_streamline_draw (GSList * stream, FILE * fp)
+void gfs_streamline_draw (GList * stream, FILE * fp)
 {
-  guint n = g_slist_length (stream);
+  guint n = g_list_length (stream);
 
   g_return_if_fail (fp != NULL);
 
@@ -1998,10 +1899,10 @@ void gfs_streamline_draw (GSList * stream, FILE * fp)
   }
 }
 
-void gfs_streamline_destroy (GSList * stream)
+void gfs_streamline_destroy (GList * stream)
 {
-  g_slist_foreach (stream, (GFunc) gts_object_destroy, NULL);
-  g_slist_free (stream);
+  g_list_foreach (stream, (GFunc) gts_object_destroy, NULL);
+  g_list_free (stream);
 }
 
 void gfs_draw_stream_cylinder (GfsDomain * domain,
@@ -2012,7 +1913,7 @@ void gfs_draw_stream_cylinder (GfsDomain * domain,
 			       FILE * fp)
 {
   GSList * profile;
-  GSList * path;
+  GList * path;
   GtsSurface * s;
 
   g_return_if_fail (domain != NULL);
@@ -2023,13 +1924,15 @@ void gfs_draw_stream_cylinder (GfsDomain * domain,
 		       gts_edge_class (),
 		       min < max ? gts_colored_vertex_class () :
 		       gts_vertex_class ());
-  path = gfs_streamline_new (domain, p, var, min, max, FALSE);
+  path = gfs_streamline_new (domain, gfs_domain_velocity (domain), p, var, min, max, FALSE, 
+			     NULL, NULL);
   profile = circle_profile (gts_point_class (), radius, 10);
-  gts_extrude_profile (s, profile, path);
+  gfs_extrude_profile (s, profile, TRUE, path);
   gts_surface_write_oogl (s, fp);
   gts_object_destroy (GTS_OBJECT (s));
   gfs_streamline_destroy (path);
-  gfs_streamline_destroy (profile);
+  g_slist_foreach (profile, (GFunc) gts_object_destroy, NULL);
+  g_slist_free (profile);
 }
 
 void gfs_draw_stream_ribbon (GfsDomain * domain,
@@ -2039,7 +1942,8 @@ void gfs_draw_stream_ribbon (GfsDomain * domain,
 			     gdouble min, gdouble max,
 			     FILE * fp)
 {
-  GSList * path, * profile;
+  GList * path;
+  GSList * profile;
   GtsSurface * s;
 
   g_return_if_fail (domain != NULL);
@@ -2050,25 +1954,28 @@ void gfs_draw_stream_ribbon (GfsDomain * domain,
 		       gts_edge_class (),
 		       min < max ? gts_colored_vertex_class () :
 		       gts_vertex_class ());
-  path = gfs_streamline_new (domain, p, var, min, max, TRUE);
+  path = gfs_streamline_new (domain, gfs_domain_velocity (domain), p, var, min, max, TRUE, 
+			     NULL, NULL);
   profile = ribbon_profile (gts_point_class (), half_width);
-  gts_extrude_profile (s, profile, path);
+  gfs_extrude_profile (s, profile, FALSE, path);
   gts_surface_write_oogl (s, fp);
   gts_object_destroy (GTS_OBJECT (s));
   gfs_streamline_destroy (path);
-  gfs_streamline_destroy (profile);
+  g_slist_foreach (profile, (GFunc) gts_object_destroy, NULL);
+  g_slist_free (profile);
 }
 
 void gfs_draw_streamline (GfsDomain * domain,
 			  FttVector p,
 			  FILE * fp)
 {
-  GSList * path;
+  GList * path;
 
   g_return_if_fail (domain != NULL);
   g_return_if_fail (fp != NULL);
 
-  path = gfs_streamline_new (domain, p, NULL, 0., 0., FALSE);
+  path = gfs_streamline_new (domain, gfs_domain_velocity (domain), p, NULL, 0., 0., FALSE, 
+			     NULL, NULL);
   gfs_streamline_draw (path, fp);
   gfs_streamline_destroy (path);
 }
