@@ -19,15 +19,15 @@
 
 #include "config.h"
 
-#ifdef HAVE_FPU_CONTROL_H
-#  include <fpu_control.h>
-#  ifdef _FPU_IEEE
-     static fpu_control_t fpu_trap_exceptions = 
-     _FPU_IEEE & ~(_FPU_MASK_ZM /*| _FPU_MASK_IM | _FPU_MASK_OM*/);
-#  endif /* _FPU_IEEE */
-#endif /* HAVE_FPU_CONTROL_H */
+#ifdef HAVE_FPU_SETCW
+# include <fpu_control.h>
+  static fpu_control_t fpu_trap_exceptions = 
+       _FPU_IEEE & ~(_FPU_MASK_ZM /*| _FPU_MASK_IM | _FPU_MASK_OM*/);
+#endif /* HAVE_FPU_SETCW */
 
 #include <stdlib.h>
+#include <locale.h>
+
 #include "boundary.h"
 #include "init.h"
 #include "refine.h"
@@ -80,28 +80,6 @@ static void gfs_log (const gchar * log_domain,
 	   log_domain, stype[type], pe, message); 
 }
 
-static void cell_level (FttCell * cell, GfsVariable * v)
-{
-  GFS_VARIABLE (cell, v->i) = ftt_cell_level (cell);
-}
-
-static void cell_fraction (FttCell * cell, GfsVariable * v)
-{
-  GFS_VARIABLE (cell, v->i) = GFS_IS_MIXED (cell) ? GFS_STATE (cell)->solid->a : 1.;
-}
-
-static void cell_lambda2 (FttCell * cell, GfsVariable * v)
-{
-  gdouble size = ftt_cell_size (cell);
-  gfs_velocity_lambda2 (cell, v);
-  GFS_VARIABLE (cell, v->i) /= size*size;
-}
-
-static void cell_curvature (FttCell * cell, GfsVariable * v)
-{
-  GFS_VARIABLE (cell, v->i) = gfs_streamline_curvature (cell)/ftt_cell_size (cell);
-}
-
 /**
  * gfs_init:
  * @argc: a pointer on the number of command line arguments passed to
@@ -114,13 +92,14 @@ static void cell_curvature (FttCell * cell, GfsVariable * v)
  */
 void gfs_init (int * argc, char *** argv)
 {
-  guint i = 0;
-  GfsVariable * v;
   static gboolean initialized = FALSE;
 
   if (initialized)
     return;
 
+  if (!setlocale (LC_ALL, "POSIX"))
+    g_warning ("cannot set locale to POSIX");
+
 #ifdef HAVE_MPI
   MPI_Initialized (&initialized);
   if (!initialized) {
@@ -140,11 +119,9 @@ void gfs_init (int * argc, char *** argv)
 #endif /* HAVE_MPI */
   initialized = TRUE;
 
-#ifdef HAVE_FPU_CONTROL_H
-#  ifdef _FPU_IEEE
-     _FPU_SETCW (fpu_trap_exceptions);
-#  endif /* _FPU_IEEE */
-#endif /* HAVE_FPU_CONTROL_H */
+#ifdef HAVE_FPU_SETCW
+   _FPU_SETCW (fpu_trap_exceptions);
+#endif /* HAVE_FPU_SETCW */
 
   g_log_set_handler (G_LOG_DOMAIN,
 		     G_LOG_LEVEL_ERROR |
@@ -157,66 +134,17 @@ void gfs_init (int * argc, char *** argv)
 		     G_LOG_FLAG_RECURSION,
 		     (GLogFunc) gfs_log, NULL);
 
-  /* Initialize permanent variables */
-  gfs_div = v = gfs_variable_new (gfs_variable_class (), NULL, NULL, FALSE, i++);
-  v->temporary = TRUE;
-  g_assert (v->i == GFS_DIV);
-  gfs_dp = v = v->next = gfs_variable_new (gfs_variable_class (), NULL, NULL, TRUE, i++);
-  v->temporary = TRUE;
-  g_assert (v->i  == GFS_DP);
-  gfs_res = v = v->next = gfs_variable_new (gfs_variable_class (), NULL, NULL, FALSE, i++);
-  v->temporary = TRUE;
-  g_assert (v->i == GFS_RES);
-  gfs_gx = v = v->next = gfs_variable_new (gfs_variable_class (), NULL, NULL, FALSE, i++);
-  v->temporary = TRUE;
-  g_assert (v->i  == GFS_GX);
-  gfs_gy = v = v->next = gfs_variable_new (gfs_variable_class (), NULL, NULL, FALSE, i++);
-  v->temporary = TRUE;
-  g_assert (v->i  == GFS_GY);
-#if (!FTT_2D)
-  gfs_gz = v = v->next = gfs_variable_new (gfs_variable_class (), NULL, NULL, FALSE, i++);
-  v->temporary = TRUE;
-  g_assert (v->i  == GFS_GZ);
-#endif /* FTT_3D */
-  gfs_centered_variables = gfs_p = v = v->next = 
-    gfs_variable_new (gfs_variable_class (), NULL, "P", TRUE, i++);
-  g_assert (v->i   == GFS_P);
-  v = v->next = gfs_variable_new (gfs_variable_class (), NULL, "U", FALSE, i++);
-  g_assert (v->i == GFS_U);
-  v = v->next = gfs_variable_new (gfs_variable_class (), NULL, "V", FALSE, i++);
-  g_assert (v->i == GFS_V);
-#if (!FTT_2D)
-  v = v->next = gfs_variable_new (gfs_variable_class (), NULL, "W", FALSE, i++);
-  g_assert (v->i == GFS_W);
-#endif /* FTT_3D */
-
-  /* Initializes derived variables */
-  gfs_derived_first = v = 
-    gfs_variable_new (gfs_variable_class (), NULL, "Vorticity", FALSE, GFS_DIV);
-  v->derived = gfs_vorticity;
-  v = v->next = gfs_variable_new (gfs_variable_class (), NULL, "Divergence", FALSE, GFS_DIV);
-  v->derived = (GfsVariableDerivedFunc) gfs_divergence;
-  v = v->next = gfs_variable_new (gfs_variable_class (), NULL, "Velocity", FALSE, GFS_DIV);
-  v->derived = gfs_velocity_norm;
-  v = v->next = gfs_variable_new (gfs_variable_class (), NULL, "Velocity2", FALSE, GFS_DIV);
-  v->derived = gfs_velocity_norm2;
-  v = v->next = gfs_variable_new (gfs_variable_class (), NULL, "Level", FALSE, GFS_DIV);
-  v->derived = cell_level;
-  v = v->next = gfs_variable_new (gfs_variable_class (), NULL, "A", FALSE, GFS_DIV);
-  v->derived = cell_fraction;
-  v = v->next = gfs_variable_new (gfs_variable_class (), NULL, "Lambda2", FALSE, GFS_DIV);
-  v->derived = cell_lambda2;
-  v = v->next = gfs_variable_new (gfs_variable_class (), NULL, "Curvature", FALSE, GFS_DIV);
-  v->derived = cell_curvature;
-  gfs_derived_last = v;
-  v->next = gfs_centered_variables;
-
   /* Instantiates classes before reading any domain or simulation file */
   gfs_simulation_class ();
     gfs_ocean_class ();
+    gfs_advection_class ();
+    gfs_poisson_class ();
 
   gfs_variable_class ();
     gfs_variable_tracer_class ();
+    gfs_variable_residual_class ();
+    gfs_variable_filtered_class ();
+    gfs_variable_curvature_class ();
 
   gfs_surface_bc_class ();
 
@@ -226,6 +154,7 @@ void gfs_init (int * argc, char *** argv)
 
   gfs_bc_dirichlet_class ();
   gfs_bc_neumann_class ();
+  gfs_bc_flather_class ();
 
   gfs_boundary_class ();
     gfs_boundary_inflow_constant_class ();
@@ -236,7 +165,9 @@ void gfs_init (int * argc, char *** argv)
 
   gfs_refine_class ();
     gfs_refine_solid_class ();
-    gfs_refine_distance_class ();
+    gfs_refine_surface_class ();
+      gfs_refine_distance_class ();
+      gfs_refine_height_class ();
 
   gfs_event_class ();
     gfs_init_class ();
@@ -248,23 +179,31 @@ void gfs_init (int * argc, char *** argv)
     gfs_adapt_class ();
       gfs_adapt_vorticity_class ();
       gfs_adapt_streamline_curvature_class ();
+      gfs_adapt_function_class ();
       gfs_adapt_gradient_class ();
         gfs_adapt_curvature_class ();
       gfs_adapt_not_box_class ();
     gfs_event_sum_class ();
-      gfs_event_sum2_class ();
+    gfs_event_harmonic_class ();
     gfs_event_stop_class ();
     gfs_event_script_class ();
     gfs_source_generic_class ();
-      gfs_source_class ();
-      gfs_source_coriolis_class ();
-      gfs_source_hydrostatic_class ();
-      gfs_source_diffusion_class ();
-        gfs_source_diffusion_explicit_class ();
+      gfs_source_scalar_class ();
+        gfs_source_class ();
+          gfs_source_control_class ();
+        gfs_source_diffusion_class ();
+          gfs_source_diffusion_explicit_class ();
+      gfs_source_velocity_class ();
         gfs_source_viscosity_class ();
-            gfs_source_vector_class ();
+        gfs_source_friction_class ();
+        gfs_source_coriolis_class ();
         gfs_source_tension_class ();
+#if !FTT_2D
+        gfs_source_hydrostatic_class ();
+#endif /* 2D3 or 3D */
     gfs_remove_droplets_class ();
+    gfs_remove_ponds_class ();
+    gfs_event_filter_class ();
    
     gfs_output_class ();
       gfs_output_time_class ();
@@ -279,17 +218,18 @@ void gfs_init (int * argc, char *** argv)
       gfs_output_location_class ();
       gfs_output_simulation_class ();
       gfs_output_boundaries_class ();
-      gfs_output_energy_class ();
+      gfs_output_particle_class ();
 
       gfs_output_scalar_class ();
         gfs_output_scalar_norm_class ();
         gfs_output_scalar_stats_class ();
         gfs_output_scalar_sum_class ();
+        gfs_output_scalar_maxima_class ();
+        gfs_output_scalar_histogram_class ();
         gfs_output_error_norm_class ();
           gfs_output_correlation_class ();
 	gfs_output_squares_class ();
 	gfs_output_streamline_class ();
-	gfs_output_streakline_class ();
         gfs_output_ppm_class ();
 
   /* If modules are not supported, calls modules init functions */
