@@ -69,6 +69,22 @@ void            gfs_output_file_close   (GfsOutputFile * file);
 
 /* GfsOutputTime: Header */
 
+typedef struct _GfsOutputTime         GfsOutputTime;
+
+struct _GfsOutputTime {
+  /*< private >*/
+  GfsOutput parent;
+
+  /*< public >*/
+  GfsClock * clock;
+};
+
+#define GFS_OUTPUT_TIME(obj)            GTS_OBJECT_CAST (obj,\
+					         GfsOutputTime,\
+					         gfs_output_time_class ())
+#define GFS_IS_OUTPUT_TIME(obj)         (gts_object_is_from_class (obj,\
+						 gfs_output_time_class ()))
+
 GfsOutputClass * gfs_output_time_class  (void);
 
 /* GfsOutputProgress: Header */
@@ -112,7 +128,7 @@ struct _GfsOutputLocation {
   GfsOutput parent;
 
   /*< public >*/
-  FttVector p;
+  GArray * p;
 };
 
 #define GFS_OUTPUT_LOCATION(obj)            GTS_OBJECT_CAST (obj,\
@@ -126,13 +142,15 @@ GfsOutputClass * gfs_output_location_class  (void);
 /* GfsOutputSimulation: Header */
 
 typedef struct _GfsOutputSimulation         GfsOutputSimulation;
+typedef enum   { GFS, GFS_TEXT }            GfsOutputSimulationFormat;
 
 struct _GfsOutputSimulation {
   GfsOutput parent;
 
   gint max_depth;
-  GfsVariable * var;
-  gboolean binary;
+  GSList * var;
+  gboolean binary, surface;
+  GfsOutputSimulationFormat format;
 };
 
 #define GFS_OUTPUT_SIMULATION(obj)            GTS_OBJECT_CAST (obj,\
@@ -157,11 +175,12 @@ typedef struct _GfsOutputScalar         GfsOutputScalar;
 struct _GfsOutputScalar {
   /*< private >*/
   GfsOutput parent;
-
   gboolean autoscale;
+  GfsFunction * f;
   
   /*< public >*/
   GfsVariable * v;
+  gchar * name;
   gdouble min, max;
   gint maxlevel;
   GtsBBox * box;
@@ -187,9 +206,49 @@ GfsOutputClass * gfs_output_scalar_stats_class  (void);
 
 GfsOutputClass * gfs_output_scalar_sum_class  (void);
 
-/* GfsOutputEnergy: Header */
+/* GfsOutputScalarMaxima: Header */
 
-GfsOutputClass * gfs_output_energy_class  (void);
+typedef struct _GfsOutputScalarMaxima         GfsOutputScalarMaxima;
+
+struct _GfsOutputScalarMaxima {
+  /*< private >*/
+  GfsOutputScalar parent;
+  
+  /*< public >*/
+  guint N;
+  gdouble * m[4];
+};
+
+#define GFS_OUTPUT_SCALAR_MAXIMA(obj)            GTS_OBJECT_CAST (obj,\
+					         GfsOutputScalarMaxima,\
+					         gfs_output_scalar_maxima_class ())
+#define GFS_IS_OUTPUT_SCALAR_MAXIMA(obj)         (gts_object_is_from_class (obj,\
+						 gfs_output_scalar_maxima_class ()))
+
+GfsOutputClass * gfs_output_scalar_maxima_class  (void);
+
+/* GfsOutputScalarHistogram: Header */
+
+typedef struct _GfsOutputScalarHistogram         GfsOutputScalarHistogram;
+
+struct _GfsOutputScalarHistogram {
+  /*< private >*/
+  GfsOutputScalar parent;
+  
+  /*< public >*/
+  gdouble * x, * y, * w, W;
+  GfsFunction * wf, * yf;
+  gdouble last, dt;
+  guint n;
+};
+
+#define GFS_OUTPUT_SCALAR_HISTOGRAM(obj)            GTS_OBJECT_CAST (obj,\
+					         GfsOutputScalarHistogram,\
+					         gfs_output_scalar_histogram_class ())
+#define GFS_IS_OUTPUT_SCALAR_HISTOGRAM(obj)         (gts_object_is_from_class (obj,\
+						 gfs_output_scalar_histogram_class ()))
+
+GfsOutputClass * gfs_output_scalar_histogram_class  (void);
 
 /* GfsOutputErrorNorm: Header */
 
@@ -244,29 +303,25 @@ struct _GfsOutputStreamline {
 
 GfsOutputClass * gfs_output_streamline_class  (void);
 
-/* GfsOutputStreakline: Header */
+/* GfsOutputParticle: Header */
 
-typedef struct _GfsOutputStreakline         GfsOutputStreakline;
+typedef struct _GfsOutputParticle         GfsOutputParticle;
 
-struct _GfsOutputStreakline {
+struct _GfsOutputParticle {
   /*< private >*/
-  GfsOutputScalar parent;
-
-  gboolean started;
-  gdouble ds;
+  GfsOutput parent;
 
   /*< public >*/
   GtsPoint * p;
-  GSList * streak;
 };
 
-#define GFS_OUTPUT_STREAKLINE(obj)         GTS_OBJECT_CAST (obj,\
-					       GfsOutputStreakline,\
-					       gfs_output_streakline_class ())
-#define GFS_IS_OUTPUT_STREAKLINE(obj)     (gts_object_is_from_class (obj,\
-					       gfs_output_streakline_class ()))
+#define GFS_OUTPUT_PARTICLE(obj)         GTS_OBJECT_CAST (obj,\
+					       GfsOutputParticle,\
+					       gfs_output_particle_class ())
+#define GFS_IS_OUTPUT_PARTICLE(obj)     (gts_object_is_from_class (obj,\
+					       gfs_output_particle_class ()))
 
-GfsOutputClass * gfs_output_streakline_class  (void);
+GfsOutputClass * gfs_output_particle_class  (void);
 
 /* GfsOutputPPM: Header */
 
