@@ -22,170 +22,154 @@
 #include "solid.h"
 #include "source.h"
 
-#define FACE_GRADIENT(x, y, z, u) (gfs_face_weighted_gradient (x, y, z, u))
-//#define FACE_GRADIENT(x, y, z, u) (gfs_face_gradient_flux_centered (x, y, z, u))
-
-void gfs_face_gradient_flux_centered (const FttCellFace * face,
-				      GfsGradient * g,
-				      guint v,
-				      gint max_level)
+/**
+ * gfs_multilevel_params_write:
+ * @par: the multilevel parameters.
+ * @fp: a file pointer.
+ *
+ * Writes in @fp a text representation of the multilevel parameters
+ * @par.  
+ */
+void gfs_multilevel_params_write (GfsMultilevelParams * par, FILE * fp)
 {
-  guint level;
+  g_return_if_fail (par != NULL);
+  g_return_if_fail (fp != NULL);
 
-  g_return_if_fail (face != NULL);
-
-  g->a = g->b = 0.;
-  if (face->neighbor == NULL)
-    return;
-
-  level = ftt_cell_level (face->cell);
-  if (ftt_cell_level (face->neighbor) < level) {
-    g_assert_not_implemented ();
-#if 0
-    /* neighbor is at a shallower level */
-    Gradient gcf;
-    gdouble w = GFS_STATE (face->cell)->f[face->d].v;
-
-    gcf = gradient_flux_fine_coarse (face, v, max_level);
-    g->a = w*gcf.a;
-    g->b = w*(gcf.b*GFS_VARIABLE (face->neighbor, v) + gcf.c);
-#endif
-  }
-  else {
-    if (level == max_level || FTT_CELL_IS_LEAF (face->neighbor)) {
-      /* neighbor is at the same level */
-      gdouble w = GFS_STATE (face->cell)->f[face->d].v;
-
-      if (!GFS_IS_MIXED (face->cell) || !GFS_IS_MIXED (face->neighbor)) {
-	g->a = w;
-	g->b = w*GFS_VARIABLE (face->neighbor, v);
-      }
-      else {
-	FttComponent c = face->d/2;
-	FttComponent cp = FTT_ORTHOGONAL_COMPONENT (c);
-	FttCell * n[4];
-	gdouble ye = (1. - GFS_FACE_FRACTION (face))/2.;
-	GfsSolidVector * s = GFS_STATE (face->cell)->solid;
-
-	n[0] = face->cell; n[1] = face->neighbor;
-	if ((s->s[2*cp] == 1. && s->s[2*cp + 1] < 1.) ||
-	    (s->s[2*cp + 1] == 0. && 
-	     s->s[2*cp] < 1. && s->s[2*cp] > 0.)) {
-	  n[2] = ftt_cell_neighbor (n[0], 2*cp);
-	  n[3] = ftt_cell_neighbor (n[1], 2*cp);
-	}
-	else if ((s->s[2*cp + 1] == 1. && s->s[2*cp] < 1.) ||
-		 (s->s[2*cp] == 0. && 
-		  s->s[2*cp + 1] < 1. && s->s[2*cp + 1] > 0.)) {
-	  n[2] = ftt_cell_neighbor (n[0], 2*cp + 1);
-	  n[3] = ftt_cell_neighbor (n[1], 2*cp + 1);
-	}
-	else {
-	  g->a = w;
-	  g->b = w*GFS_VARIABLE (n[1], v);
-	  return;
-	}
-
-	//	g_assert (n[2] && n[3]);
-	if (n[2] && n[3]) {
-	  g->a = w*(1. - ye);
-	  g->b = w*(1. - ye)*GFS_VARIABLE (n[1], v) + 
-	    w*ye*(GFS_VARIABLE (n[3], v) - GFS_VARIABLE (n[2], v));
-	}
-	else {
-	  g->a = w;
-	  g->b = w*GFS_VARIABLE (n[1], v);
-	}
-      }
-    }
-    else {
-    g_assert_not_implemented ();
-#if 0
-      /* neighbor is at a deeper level */
-      FttCellChildren children;
-      FttCellFace f;
-      guint i, n;
-      
-      f.d = FTT_OPPOSITE_DIRECTION (face->d);
-      n = ftt_cell_children_direction (face->neighbor, f.d, &children);
-      f.neighbor = face->cell;
-      for (i = 0; i < n; i++) {
-	Gradient gcf;
-	gdouble w;
-
-	f.cell = children.c[i];
-	w = GFS_STATE (f.cell)->f[f.d].v;
-	
-	/* check for mixed cell refinement violation (topology.fig) */
-	g_assert (f.cell);
-	
-	gcf = gradient_fine_coarse (&f, v, max_level);
-	g->a += w*gcf.b;
-	g->b += w*(gcf.a*GFS_VARIABLE (f.cell, v) - gcf.c);
-      }
-#endif
-#if (!FTT_2D && !FTT_2D3)
-      g->a /= 2.;
-      g->b /= 2.;
-#endif /* not 2D and not 2D3 */
-    }
-  }
+  fprintf (fp,
+           "{\n"
+	   "  tolerance = %g\n"
+	   "  nrelax    = %u\n"
+           "  erelax    = %u\n"
+	   "  minlevel  = %u\n"
+	   "  nitermax  = %u\n"
+	   "  nitermin  = %u\n"
+	   "  weighted  = %d\n"
+	   "  beta      = %g\n"
+	   "}",
+	   par->tolerance,
+	   par->nrelax,
+	   par->erelax,
+	   par->minlevel,
+	   par->nitermax,
+	   par->nitermin,
+	   par->weighted,
+	   par->beta);
 }
 
-static void relax (FttCell * cell, gpointer * data)
+void gfs_multilevel_params_init (GfsMultilevelParams * par)
+{
+  g_return_if_fail (par != NULL);
+
+  par->tolerance = 1e-3;
+  par->nrelax    = 4;
+  par->erelax    = 1;
+  par->minlevel  = 0;
+  par->nitermax  = 100;
+  par->nitermin  = 0;
+
+  par->dimension = FTT_DIMENSION;
+  par->weighted = FALSE;
+  par->beta = 0.5;
+}
+
+void gfs_multilevel_params_read (GfsMultilevelParams * par, GtsFile * fp)
+{
+  GtsFileVariable var[] = {
+    {GTS_DOUBLE, "tolerance", TRUE},
+    {GTS_UINT,   "nrelax",    TRUE},
+    {GTS_UINT,   "erelax",    TRUE},
+    {GTS_UINT,   "minlevel",  TRUE},
+    {GTS_UINT,   "nitermax",  TRUE},
+    {GTS_UINT,   "nitermin",  TRUE},
+    {GTS_INT,    "weighted",  TRUE},
+    {GTS_DOUBLE, "beta",      TRUE},
+    {GTS_NONE}
+  };
+
+  g_return_if_fail (par != NULL);
+  g_return_if_fail (fp != NULL);
+
+  var[0].data = &par->tolerance;
+  var[1].data = &par->nrelax;
+  var[2].data = &par->erelax;
+  var[3].data = &par->minlevel;
+  var[4].data = &par->nitermax;
+  var[5].data = &par->nitermin;
+  var[6].data = &par->weighted;
+  var[7].data = &par->beta;
+
+  gts_file_assign_variables (fp, var);
+  if (fp->type == GTS_ERROR)
+    return;
+
+  if (par->tolerance <= 0.) {
+    gts_file_variable_error (fp, var, "tolerance",
+			     "tolerance `%g' must be strictly positive",
+			     par->tolerance);
+    return;
+  }
+  if (par->nrelax == 0)
+    gts_file_variable_error (fp, var, "nrelax", "nrelax must be non zero");
+  if (par->erelax == 0)
+    gts_file_variable_error (fp, var, "erelax", "erelax must be non zero");
+  if (par->beta < 0.5 || par->beta > 1.)
+    gts_file_variable_error (fp, var, "beta", "beta must be in [0.5,1]");
+}
+
+typedef struct {
+  guint u, rhs, dia, res;
+  gint maxlevel;
+  gdouble beta;
+} RelaxParams;
+
+static void relax (FttCell * cell, RelaxParams * p)
 {
-  guint u = ((GfsVariable *) data[0])->i;
-  guint rhs = ((GfsVariable *) data[1])->i;
-  gint max_level = *((gint *) data[2]);
   GfsGradient g;
   FttCellNeighbors neighbor;
   FttCellFace f;
   GfsGradient ng;
 
-  g.a = GFS_STATE (cell)->g[0];
+  g.a = GFS_VARIABLE (cell, p->dia);
   g.b = 0.;
   f.cell = cell;
   ftt_cell_neighbors (cell, &neighbor);
   for (f.d = 0; f.d < FTT_NEIGHBORS; f.d++) {
     f.neighbor = neighbor.c[f.d];
     if (f.neighbor) {
-      FACE_GRADIENT (&f, &ng, u, max_level);
+      gfs_face_weighted_gradient (&f, &ng, p->u, p->maxlevel);
       g.a += ng.a;
       g.b += ng.b;
     }
   }
   if (g.a > 0.)
-    GFS_VARIABLE (cell, u) = (g.b - GFS_VARIABLE (cell, rhs))/g.a;
+    GFS_VARIABLE (cell, p->u) = (g.b - GFS_VARIABLE (cell, p->rhs))/g.a;
   else
-    GFS_VARIABLE (cell, u) = 0.;
+    GFS_VARIABLE (cell, p->u) = 0.;
 }
 
-static void relax2D (FttCell * cell, gpointer * data)
+static void relax2D (FttCell * cell, RelaxParams * p)
 {
-  guint u = ((GfsVariable *) data[0])->i;
-  guint rhs = ((GfsVariable *) data[1])->i;
-  gint max_level = *((gint *) data[2]);
   GfsGradient g;
   FttCellNeighbors neighbor;
   FttCellFace f;
   GfsGradient ng;
 
-  g.a = GFS_STATE (cell)->g[0];
+  g.a = GFS_VARIABLE (cell, p->dia);
   g.b = 0.;
   f.cell = cell;
   ftt_cell_neighbors (cell, &neighbor);
   for (f.d = 0; f.d < FTT_NEIGHBORS_2D; f.d++) {
     f.neighbor = neighbor.c[f.d];
     if (f.neighbor) {
-      FACE_GRADIENT (&f, &ng, u, max_level);
+      gfs_face_weighted_gradient (&f, &ng, p->u, p->maxlevel);
       g.a += ng.a;
       g.b += ng.b;
     }
   }
   if (g.a > 0.)
-    GFS_VARIABLE (cell, u) = (g.b - GFS_VARIABLE (cell, rhs))/g.a;
+    GFS_VARIABLE (cell, p->u) = (g.b - GFS_VARIABLE (cell, p->rhs))/g.a;
   else
-    GFS_VARIABLE (cell, u) = 0.;
+    GFS_VARIABLE (cell, p->u) = 0.;
 }
 
 /**
@@ -195,6 +179,7 @@ static void relax2D (FttCell * cell, gpointer * data)
  * @max_depth: the maximum depth of the domain to relax.
  * @u: the variable to use as left-hand side.
  * @rhs: the variable to use as right-hand side.
+ * @dia: the diagonal weight.
  *
  * Apply one pass of a Jacobi relaxation to all the leaf cells of
  * @domain with a level inferior or equal to @max_depth and to all the
@@ -206,75 +191,71 @@ void gfs_relax (GfsDomain * domain,
 		guint d,
 		gint max_depth,
 		GfsVariable * u,
-		GfsVariable * rhs)
+		GfsVariable * rhs,
+		GfsVariable * dia)
 {
-  gpointer data[3];
+  RelaxParams p;
 
   g_return_if_fail (domain != NULL);
   g_return_if_fail (d > 1 && d <= 3);
   g_return_if_fail (u != NULL);
   g_return_if_fail (rhs != NULL);
+  g_return_if_fail (dia != NULL);
 
-  data[0] = u;
-  data[1] = rhs;
-  data[2] = &max_depth;
-
+  p.u = u->i;
+  p.rhs = rhs->i;
+  p.dia = dia->i;
+  p.maxlevel = max_depth;
   gfs_domain_cell_traverse (domain, FTT_PRE_ORDER, 
 			    FTT_TRAVERSE_LEVEL | FTT_TRAVERSE_LEAFS,
 			    max_depth,
-			    (FttCellTraverseFunc) (d == 2 ? relax2D : relax), data);
+			    (FttCellTraverseFunc) (d == 2 ? relax2D : relax), &p);
 }
 
-static void residual_set (FttCell * cell, gpointer * data)
+static void residual_set (FttCell * cell, RelaxParams * p)
 {
-  guint u = ((GfsVariable *) data[0])->i;
-  guint rhs = ((GfsVariable *) data[1])->i;
-  guint res = ((GfsVariable *) data[2])->i;
   GfsGradient g;
   FttCellNeighbors neighbor;
   FttCellFace f;
   GfsGradient ng;
 
-  g.a = GFS_STATE (cell)->g[0];
+  g.a = GFS_VARIABLE (cell, p->dia);
   g.b = 0.;
   f.cell = cell;
   ftt_cell_neighbors (cell, &neighbor);
   for (f.d = 0; f.d < FTT_NEIGHBORS; f.d++) {
     f.neighbor = neighbor.c[f.d];
     if (f.neighbor) {
-      FACE_GRADIENT (&f, &ng, u, -1);
+      gfs_face_weighted_gradient (&f, &ng, p->u, p->maxlevel);
       g.a += ng.a;
       g.b += ng.b;
     }
   }
-  GFS_VARIABLE (cell, res) = GFS_VARIABLE (cell, rhs) - 
-    (g.b - GFS_VARIABLE (cell, u)*g.a);
+  GFS_VARIABLE (cell, p->res) = GFS_VARIABLE (cell, p->rhs) - 
+    (g.b - GFS_VARIABLE (cell, p->u)*g.a);
 }
 
-static void residual_set2D (FttCell * cell, gpointer * data)
+static void residual_set2D (FttCell * cell, RelaxParams * p)
 {
-  guint u = ((GfsVariable *) data[0])->i;
-  guint rhs = ((GfsVariable *) data[1])->i;
-  guint res = ((GfsVariable *) data[2])->i;
   GfsGradient g;
   FttCellNeighbors neighbor;
   FttCellFace f;
   GfsGradient ng;
 
-  g.a = GFS_STATE (cell)->g[0];
+  g.a = GFS_VARIABLE (cell, p->dia);
   g.b = 0.;
   f.cell = cell;
   ftt_cell_neighbors (cell, &neighbor);
   for (f.d = 0; f.d < FTT_NEIGHBORS_2D; f.d++) {
     f.neighbor = neighbor.c[f.d];
     if (f.neighbor) {
-      FACE_GRADIENT (&f, &ng, u, -1);
+      gfs_face_weighted_gradient (&f, &ng, p->u, p->maxlevel);
       g.a += ng.a;
       g.b += ng.b;
     }
   }
-  GFS_VARIABLE (cell, res) = GFS_VARIABLE (cell, rhs) - 
-    (g.b - GFS_VARIABLE (cell, u)*g.a);
+  GFS_VARIABLE (cell, p->res) = GFS_VARIABLE (cell, p->rhs) - 
+    (g.b - GFS_VARIABLE (cell, p->u)*g.a);
 }
 
 /**
@@ -285,6 +266,7 @@ static void residual_set2D (FttCell * cell, gpointer * data)
  * @max_depth: maximum depth of the traversal.
  * @u: the variable to use as left-hand side.
  * @rhs: the variable to use as right-hand side.
+ * @dia: the diagonal weight.
  * @res: the variable to use to store the residual.
  *
  * For each cell of @domain, computes the sum of the residual over
@@ -296,22 +278,25 @@ void gfs_residual (GfsDomain * domain,
 		   guint d,
 		   FttTraverseFlags flags,
 		   gint max_depth,
-		   GfsVariable * u, GfsVariable * rhs, GfsVariable * res)
+		   GfsVariable * u, GfsVariable * rhs, GfsVariable * dia,
+		   GfsVariable * res)
 {
-  gpointer data[3];
+  RelaxParams p;
 
   g_return_if_fail (domain != NULL);
   g_return_if_fail (d > 1 && d <= 3);
   g_return_if_fail (u != NULL);
   g_return_if_fail (rhs != NULL);
+  g_return_if_fail (dia != NULL);
   g_return_if_fail (res != NULL);
 
-  data[0] = u;
-  data[1] = rhs;
-  data[2] = res;
-
+  p.u = u->i;
+  p.rhs = rhs->i;
+  p.dia = dia->i;
+  p.res = res->i;
+  p.maxlevel = max_depth;
   gfs_domain_cell_traverse (domain, FTT_PRE_ORDER, flags, max_depth,
-			    (FttCellTraverseFunc) (d == 2 ? residual_set2D : residual_set), data);
+			    (FttCellTraverseFunc) (d == 2 ? residual_set2D : residual_set), &p);
 }
 
 static void reset_coeff (FttCell * cell)
@@ -345,19 +330,29 @@ static void poisson_coeff (FttCellFace * face, gdouble * lambda2)
   }
 }
 
-static void poisson_density_coeff (FttCellFace * face,
-				   gpointer * data)
+static void reset_alpha_coeff (FttCell * cell, gpointer * data)
 {
-  GfsVariable * c = data[0];
-  gdouble * rho = data[1];
-  gdouble * lambda2 = data[2];
-  gdouble v = lambda2[face->d/2], cval;
+  FttDirection d;
+  GfsFaceStateVector * f = GFS_STATE (cell)->f;
+  GfsFunction * alpha = data[0];
+  GfsVariable * a = data[1];
+  
+  for (d = 0; d < FTT_NEIGHBORS; d++)
+    f[d].v = 0.;
+  GFS_VARIABLE (cell, a->i) = gfs_function_value (alpha, cell);
+}
+
+static void poisson_alpha_coeff (FttCellFace * face,
+				 gpointer * data)
+{
+  gdouble * lambda2 = data[0];
+  GfsVariable * alpha = data[1];
+  gdouble v = lambda2[face->d/2];
   GfsStateVector * s = GFS_STATE (face->cell);
 
   if (GFS_IS_MIXED (face->cell))
     v *= s->solid->s[face->d];
-  cval = gfs_face_interpolated_value (face, c->i);
-  v /= 1. + (cval > 1. ? 1. : cval < 0. ? 0. : cval)*(*rho - 1.);
+  v *= gfs_face_interpolated_value (face, alpha->i);
   s->f[face->d].v = v;
 
   switch (ftt_face_type (face)) {
@@ -394,14 +389,15 @@ static void face_coeff_from_below (FttCell * cell)
 /**
  * gfs_poisson_coefficients:
  * @domain: a #GfsDomain.
- * @c: the volume fraction.
- * @rho: the relative density.
+ * @alpha: the inverse of density or %NULL.
  *
- * Initializes the face coefficients for the Poisson equation.
+ * Initializes the face coefficients for the Poisson equation
+ * $\nabla\cdot\alpha\nabla p=\dots$.
+ *
+ * If @alpha is %NULL, it is taken to be unity.
  */
 void gfs_poisson_coefficients (GfsDomain * domain,
-			       GfsVariable * c,
-			       gdouble rho)
+			       GfsFunction * alpha)
 {
   gdouble lambda2[FTT_DIMENSION];
   FttComponent i;
@@ -413,117 +409,179 @@ void gfs_poisson_coefficients (GfsDomain * domain,
 
     lambda2[i] = lambda*lambda;
   }
-  gfs_domain_cell_traverse (domain,
-			    FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
-			    (FttCellTraverseFunc) reset_coeff, NULL);
-  if (c == NULL || rho == 1.)
+  if (alpha == NULL) {
+    gfs_domain_cell_traverse (domain,
+			      FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
+			      (FttCellTraverseFunc) reset_coeff, NULL);
     gfs_domain_face_traverse (domain, FTT_XYZ, 
 			      FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
 			      (FttFaceTraverseFunc) poisson_coeff, lambda2);
+  }
   else {
-    gpointer data[3];
+    gpointer data[2];
 
-    data[0] = c;
-    data[1] = &rho;
-    data[2] = lambda2;
+    data[0] = alpha;
+    data[1] = gfs_temporary_variable (domain);
+    gfs_domain_cell_traverse (domain,
+			      FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
+			      (FttCellTraverseFunc) reset_alpha_coeff, data);
+    data[0] = lambda2;
     gfs_domain_face_traverse (domain, FTT_XYZ, 
 			      FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
-			      (FttFaceTraverseFunc) poisson_density_coeff, 
+			      (FttFaceTraverseFunc) poisson_alpha_coeff, 
 			      data);
+    gts_object_destroy (data[1]);
   }
   gfs_domain_cell_traverse (domain,
 			    FTT_POST_ORDER, FTT_TRAVERSE_NON_LEAFS, -1,
-			    (FttCellTraverseFunc) face_coeff_from_below, 
-			    NULL);
+			    (FttCellTraverseFunc) face_coeff_from_below, NULL);
 }
 
-static void correct (FttCell * cell, GfsVariable * u) 
+static void correct (FttCell * cell, gpointer * data)
 {
-  GFS_VARIABLE (cell, u->i) += GFS_STATE (cell)->dp;
+  GfsVariable * u = data[0];
+  GfsVariable * dp = data[1];
+  GFS_VARIABLE (cell, u->i) += GFS_VARIABLE (cell, dp->i);
+}
+
+static void get_from_above (FttCell * parent, GfsVariable * v)
+{
+  guint level = ftt_cell_level (parent);
+  FttCellNeighbors n;
+  FttCellChildren child;
+  FttComponent c;
+  FttVector h;
+  guint i;
+
+  ftt_cell_neighbors (parent, &n);
+  for (c = 0; c < FTT_DIMENSION; c++) {
+    FttCellFace f;
+    GfsGradient g;
+    gdouble g1, g2;
+    
+    f.cell = parent;
+    f.d = 2*c;
+    f.neighbor = n.c[f.d];
+    gfs_face_gradient (&f, &g, v->i, level);
+    g1 = g.b - g.a*GFS_VARIABLE (parent, v->i);
+    f.d = 2*c + 1;
+    f.neighbor = n.c[f.d];
+    gfs_face_gradient (&f, &g, v->i, level);
+    g2 = g.b - g.a*GFS_VARIABLE (parent, v->i);
+    (&h.x)[c] = (g1 - g2)/2.;
+  }
+
+  ftt_cell_children (parent, &child);
+  for (i = 0; i < FTT_CELLS; i++) 
+    if (child.c[i]) {
+      FttVector p;
+      
+      GFS_VARIABLE (child.c[i], v->i) = GFS_VARIABLE (parent, v->i);
+      ftt_cell_relative_pos (child.c[i], &p);
+      for (c = 0; c < FTT_DIMENSION; c++)
+	GFS_VARIABLE (child.c[i], v->i) += (&p.x)[c]*(&h.x)[c];
+    }
 }
 
 /**
  * gfs_poisson_cycle:
  * @domain: the domain on which to solve the Poisson equation.
- * @d: number of dimensions (2 or 3).
- * @levelmin: the top level of the multigrid hierarchy.
- * @depth: the total depth of the domain.
- * @nrelax: the number of relaxations to apply at each level.
+ * @p: the #GfsMultilevelParams.
  * @u: the variable to use as left-hand side.
  * @rhs: the variable to use as right-hand side.
+ * @dia: the diagonal weight.
+ * @res: the residual.
  *
  * Apply one multigrid iteration to the Poisson equation defined by @u
  * and @rhs.
  *
- * The initial value of %GFS_RES on the leaves of @root must be set to
+ * The initial value of @res on the leaves of @root must be set to
  * the residual of the Poisson equation (using gfs_residual()).
  *
  * The face coefficients must be set using gfs_poisson_coefficients().
  *
  * The values of @u on the leaf cells are updated as well as the values
- * of %GFS_RES (i.e. the cell tree is ready for another iteration).
+ * of @res (i.e. the cell tree is ready for another iteration).
  */
 void gfs_poisson_cycle (GfsDomain * domain,
-			guint d,
-			guint levelmin,
-			guint depth,
-			guint nrelax,
+			GfsMultilevelParams * p,
 			GfsVariable * u,
-			GfsVariable * rhs)
+			GfsVariable * rhs,
+			GfsVariable * dia,
+			GfsVariable * res)
 {
-  guint n, l;
-
+  guint n, l, nrelax, minlevel;
+  GfsVariable * dp;
+  gpointer data[2];
+  
   g_return_if_fail (domain != NULL);
-  g_return_if_fail (d > 1 && d <= 3);
+  g_return_if_fail (p != NULL);
+  g_return_if_fail (p->dimension > 1 && p->dimension <= 3);
   g_return_if_fail (u != NULL);
   g_return_if_fail (rhs != NULL);
+  g_return_if_fail (dia != NULL);
+  g_return_if_fail (res != NULL);
+
+  dp = gfs_temporary_variable (domain);
+  minlevel = MAX (domain->rootlevel, p->minlevel);
 
   /* compute residual on non-leafs cells */
   gfs_domain_cell_traverse (domain, 
 			    FTT_POST_ORDER, FTT_TRAVERSE_NON_LEAFS, -1,
-			    (FttCellTraverseFunc) gfs_get_from_below_extensive, 
-			    gfs_res);
+			    (FttCellTraverseFunc) gfs_get_from_below_extensive, res);
 
   /* relax top level */
-  gfs_domain_cell_traverse (domain, 
-			    FTT_PRE_ORDER, FTT_TRAVERSE_LEVEL, levelmin,
-			    (FttCellTraverseFunc) gfs_cell_reset, gfs_dp);
-  for (n = 0; n < 10*nrelax; n++) {
+  gfs_domain_cell_traverse (domain,
+			    FTT_PRE_ORDER, FTT_TRAVERSE_LEVEL | FTT_TRAVERSE_LEAFS, minlevel,
+			    (FttCellTraverseFunc) gfs_cell_reset, dp);
+  nrelax = p->nrelax;
+  for (l = minlevel; l < p->depth; l++)
+    nrelax *= p->erelax;
+  for (n = 0; n < nrelax; n++) {
     gfs_domain_homogeneous_bc (domain,
 			       FTT_TRAVERSE_LEVEL | FTT_TRAVERSE_LEAFS,
-			       levelmin, gfs_dp, u);
-    gfs_relax (domain, d, levelmin, gfs_dp, gfs_res);
+			       minlevel, dp, u);
+    gfs_relax (domain, p->dimension, minlevel, dp, res, dia);
   }
+  nrelax /= p->erelax;
 
   /* relax from top to bottom */
-  for (l = levelmin + 1; l <= depth; l++) {
+  for (l = minlevel + 1; l <= p->depth; l++, nrelax /= p->erelax) {
     /* get initial guess from coarser grid */ 
-    gfs_domain_cell_traverse (domain, 
-			      FTT_PRE_ORDER, FTT_TRAVERSE_LEVEL, l,
-			      (FttCellTraverseFunc) gfs_get_from_above, 
-			      gfs_dp);
+    gfs_domain_cell_traverse (domain,
+			      FTT_PRE_ORDER, FTT_TRAVERSE_LEVEL | FTT_TRAVERSE_NON_LEAFS, l - 1,
+			      (FttCellTraverseFunc) get_from_above, dp);
     for (n = 0; n < nrelax; n++) {
       gfs_domain_homogeneous_bc (domain, 
 				 FTT_TRAVERSE_LEVEL | FTT_TRAVERSE_LEAFS,
-				 l, gfs_dp, u);
-      gfs_relax (domain, d, l, gfs_dp, gfs_res);
+				 l, dp, u);
+      gfs_relax (domain, p->dimension, l, dp, res, dia);
     }
   }
   /* correct on leaf cells */
+  data[0] = u;
+  data[1] = dp;
   gfs_domain_cell_traverse (domain, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
-			    (FttCellTraverseFunc) correct, u);
+			    (FttCellTraverseFunc) correct, data);
   gfs_domain_bc (domain, FTT_TRAVERSE_LEAFS, -1, u);
   /* compute new residual on leaf cells */
-  gfs_residual (domain, d, FTT_TRAVERSE_LEAFS, -1, u, rhs, gfs_res);
+  gfs_residual (domain, p->dimension, FTT_TRAVERSE_LEAFS, -1, u, rhs, dia, res);
+
+  gts_object_destroy (GTS_OBJECT (dp));
 }
 
-static void diffusion_coef (FttCellFace * face, gpointer * data)
+typedef struct {
+  GfsSourceDiffusion * d;
+  gdouble lambda2[FTT_DIMENSION];
+  gdouble dt;
+  GfsVariable * dia;
+  GfsFunction * alpha;
+} DiffusionCoeff;
+
+static void diffusion_coef (FttCellFace * face, DiffusionCoeff * c)
 {
-  GfsSourceDiffusion * d = data[0];
-  gdouble * lambda2 = data[1];
-  gdouble * dt = data[2];
   GfsStateVector * s = GFS_STATE (face->cell);
-  gdouble v = lambda2[face->d/2]*(*dt)*gfs_source_diffusion_face (d, face);
+  gdouble v = c->lambda2[face->d/2]*c->dt*gfs_source_diffusion_face (c->d, face);
 
   if (GFS_IS_MIXED (face->cell))
     v *= s->solid->s[face->d];
@@ -542,16 +600,12 @@ static void diffusion_coef (FttCellFace * face, gpointer * data)
   }
 }
 
-static void diffusion_mixed_coef (FttCell * cell, gpointer * data)
+static void diffusion_mixed_coef (FttCell * cell, DiffusionCoeff * c)
 {
   reset_coeff (cell);
-  if (GFS_IS_MIXED (cell)) {
-    GfsSourceDiffusion * d = data[0];
-    gdouble * dt = data[2];
-
-    GFS_STATE (cell)->solid->v = *dt*gfs_source_diffusion_cell (d, cell);
-  }
-  GFS_STATE (cell)->g[0] = 1.;
+  if (GFS_IS_MIXED (cell))
+    GFS_STATE (cell)->solid->v = c->dt*gfs_source_diffusion_cell (c->d, cell);
+  GFS_VARIABLE (cell, c->dia->i) = c->alpha ? 1./gfs_function_value (c->alpha, cell) : 1.;
 }
 
 /**
@@ -559,133 +613,115 @@ static void diffusion_mixed_coef (FttCell * cell, gpointer * data)
  * @domain: a #GfsDomain.
  * @d: a #GfsSourceDiffusion.
  * @dt: the time-step.
+ * @dia: where to store the diagonal weight.
+ * @alpha: the inverse of density or %NULL.
+ * @beta: the implicitness parameter (0.5 Crank-Nicholson, 1. backward Euler).
  *
  * Initializes the face coefficients for the diffusion equation.
  */
 void gfs_diffusion_coefficients (GfsDomain * domain,
 				 GfsSourceDiffusion * d,
-				 gdouble dt)
+				 gdouble dt,
+				 GfsVariable * dia,
+				 GfsFunction * alpha,
+				 gdouble beta)
 {
-  gdouble lambda2[FTT_DIMENSION];
-  gpointer data[3];
+  DiffusionCoeff coef;
   FttComponent i;
 
   g_return_if_fail (domain != NULL);
   g_return_if_fail (d != NULL);
+  g_return_if_fail (dia != NULL);
+  g_return_if_fail (beta >= 0.5 && beta <= 1.);
 
   for (i = 0; i < FTT_DIMENSION; i++) {
     gdouble lambda = (&domain->lambda.x)[i];
 
-    lambda2[i] = lambda*lambda;
+    coef.lambda2[i] = lambda*lambda;
   }
-  data[0] = d;
-  data[1] = lambda2;
-  data[2] = &dt;
+  coef.d = d;
+  coef.dt = beta*dt;
+  coef.dia = dia;
+  coef.alpha = alpha;
   gfs_domain_cell_traverse (domain,
 			    FTT_PRE_ORDER, FTT_TRAVERSE_ALL, -1,
-			    (FttCellTraverseFunc) diffusion_mixed_coef, data);
+			    (FttCellTraverseFunc) diffusion_mixed_coef, &coef);
   gfs_domain_face_traverse (domain, FTT_XYZ, 
 			    FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
-			    (FttFaceTraverseFunc) diffusion_coef, data);
+			    (FttFaceTraverseFunc) diffusion_coef, &coef);
   gfs_domain_cell_traverse (domain,
 			    FTT_POST_ORDER, FTT_TRAVERSE_NON_LEAFS, -1,
 			    (FttCellTraverseFunc) face_coeff_from_below, 
 			    NULL);
 }
 
-static void density (FttCell * cell, gpointer * data)
-{
-  GfsVariable * v = data[0];
-  gdouble c = GFS_VARIABLE (cell, v->i);
-  gdouble * rho = data[1];
 
-  GFS_STATE (cell)->g[0] = 1. + *rho*(c > 1. ? 1. : c < 0. ? 0. : c);
-}
-
-/**
- * gfs_viscosity_coefficients:
- * @domain: a #GfsDomain.
- * @d: a #GfsSourceDiffusion.
- * @dt: the time-step.
- * @c: the volume fraction (at t+dt/2).
- * @rho: the relative density.
- *
- * Initializes the face coefficients for the diffusion equation for
- * the velocity.
- */
-void gfs_viscosity_coefficients (GfsDomain * domain,
-				 GfsSourceDiffusion * d,
-				 gdouble dt,
-				 GfsVariable * c,
-				 gdouble rho)
-{
-  g_return_if_fail (domain != NULL);
-  g_return_if_fail (d != NULL);
-  g_return_if_fail (c != NULL);
-
-  gfs_diffusion_coefficients (domain, d, dt);
-  if (rho != 1.) {
-    gpointer data[2];
-
-    rho -= 1.;
-    data[0] = c;
-    data[1] = &rho;
-    gfs_domain_cell_traverse (domain,
-			      FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
-			      (FttCellTraverseFunc) density, data);
-    gfs_domain_cell_traverse (domain,
-			      FTT_POST_ORDER, FTT_TRAVERSE_NON_LEAFS, -1,
-			      (FttCellTraverseFunc) gfs_get_from_below_intensive, 
-			      gfs_gx);
-  }
-}
-
-/**
- * gfs_diffusion_rhs:
- * @cell: a #FttCell.
- * @v: a #GfsVariable.
- *
- * Adds to the @div variable of @cell the right-hand side of the
- * diffusion equation for variable @v.
- *
- * The diffusion coefficients must have been already set using
- * gfs_diffusion_coefficients().
- */
-void gfs_diffusion_rhs (FttCell * cell, GfsVariable * v)
+static void diffusion_rhs (FttCell * cell, RelaxParams * p)
 {
   gdouble a, f, h, val;
   FttCellNeighbors neighbor;
   FttCellFace face;
   
-  g_return_if_fail (cell != NULL);
-  g_return_if_fail (v != NULL);
-
   if (GFS_IS_MIXED (cell)) {
-    a = GFS_STATE (cell)->solid->a*GFS_STATE (cell)->g[0];
+    a = GFS_STATE (cell)->solid->a*GFS_VARIABLE (cell, p->dia);
     if (((cell)->flags & GFS_FLAG_DIRICHLET) != 0)
-      f = gfs_cell_dirichlet_gradient_flux (cell, v->i, -1, GFS_STATE (cell)->solid->fv);
+      f = gfs_cell_dirichlet_gradient_flux (cell, p->u, -1, GFS_STATE (cell)->solid->fv);
     else
       f = GFS_STATE (cell)->solid->fv;
   }
   else {
-    a = GFS_STATE (cell)->g[0];
+    a = GFS_VARIABLE (cell, p->dia);
     f = 0.; /* Neumann condition by default */
   }
   h = ftt_cell_size (cell);
-  val = GFS_VARIABLE (cell, v->i);
+  val = GFS_VARIABLE (cell, p->u);
   face.cell = cell;
   ftt_cell_neighbors (cell, &neighbor);
   for (face.d = 0; face.d < FTT_NEIGHBORS; face.d++) {
     GfsGradient g;
 
     face.neighbor = neighbor.c[face.d];
-    gfs_face_gradient_flux (&face, &g, v->i, -1);
+    gfs_face_gradient_flux (&face, &g, p->u, -1);
     f += g.b - g.a*val;
   }
-  GFS_STATE (cell)->div += val + f/(2.*h*h*a);
+  GFS_VARIABLE (cell, p->rhs) += val + p->beta*f/(h*h*a);
 }
 
-static void diffusion_relax (FttCell * cell, gint * maxlevel)
+/**
+ * gfs_diffusion_rhs:
+ * @domain: a #GfsDomain.
+ * @v: a #GfsVariable.
+ * @rhs: a #GfsVariable.
+ * @dia: the diagonal weight.
+ * @beta: the implicitness parameter (0.5 Crank-Nicholson, 1. backward Euler).
+ *
+ * Adds to the @rhs variable of @cell the right-hand side of the
+ * diffusion equation for variable @v.
+ *
+ * The diffusion coefficients must have been already set using
+ * gfs_diffusion_coefficients().
+ */
+void gfs_diffusion_rhs (GfsDomain * domain, 
+			GfsVariable * v, GfsVariable * rhs, GfsVariable * dia,
+			gdouble beta)
+{
+  RelaxParams p;
+
+  g_return_if_fail (domain != NULL);
+  g_return_if_fail (v != NULL);
+  g_return_if_fail (rhs != NULL);
+  g_return_if_fail (dia != NULL);
+  g_return_if_fail (beta >= 0.5 && beta <= 1.);
+
+  p.u = v->i;
+  p.rhs = rhs->i;
+  p.dia = dia->i;
+  p.beta = (1. - beta)/beta;
+  gfs_domain_cell_traverse (domain, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
+			    (FttCellTraverseFunc) diffusion_rhs, &p);
+}
+
+static void diffusion_relax (FttCell * cell, RelaxParams * p)
 {
   gdouble a;
   GfsGradient g = { 0., 0. };
@@ -694,12 +730,12 @@ static void diffusion_relax (FttCell * cell, gint * maxlevel)
   FttCellFace face;
 
   if (GFS_IS_MIXED (cell)) {
-    a = GFS_STATE (cell)->solid->a*GFS_STATE (cell)->g[0];
+    a = GFS_STATE (cell)->solid->a*GFS_VARIABLE (cell, p->dia);
     if (((cell)->flags & GFS_FLAG_DIRICHLET) != 0)
-      g.b = gfs_cell_dirichlet_gradient_flux (cell, gfs_dp->i, *maxlevel, 0.);
+      g.b = gfs_cell_dirichlet_gradient_flux (cell, p->u, p->maxlevel, 0.);
   }
   else
-    a = GFS_STATE (cell)->g[0];
+    a = GFS_VARIABLE (cell, p->dia);
 
   face.cell = cell;
   ftt_cell_neighbors (cell, &neighbor);
@@ -707,31 +743,19 @@ static void diffusion_relax (FttCell * cell, gint * maxlevel)
     GfsGradient ng;
 
     face.neighbor = neighbor.c[face.d];
-    gfs_face_gradient_flux (&face, &ng, gfs_dp->i, *maxlevel);
+    gfs_face_gradient_flux (&face, &ng, p->u, p->maxlevel);
     g.a += ng.a;
     g.b += ng.b;
   }
-  a *= 2.*h*h;
+  a *= h*h;
   g_assert (a > 0.);
   g.a = 1. + g.a/a;
-  g.b = GFS_STATE (cell)->res + g.b/a;
+  g.b = GFS_VARIABLE (cell, p->res) + g.b/a;
   g_assert (g.a > 0.);
-  GFS_STATE (cell)->dp = g.b/g.a;
+  GFS_VARIABLE (cell, p->u) = g.b/g.a;
 }
 
-/**
- * gfs_diffusion_residual:
- * @cell: a #FttCell.
- * @v: a #GfsVariable.
- *
- * Sets the @res variable of @cell to the residual of the diffusion
- * equation for @v.
- *
- * The diffusion coefficients must have been set using
- * gfs_diffusion_coefficients() and the right-hand side using
- * gfs_diffusion_rhs().
- */
-void gfs_diffusion_residual (FttCell * cell, GfsVariable * v)
+static void diffusion_residual (FttCell * cell, RelaxParams * p)
 {
   gdouble a;
   GfsGradient g = { 0., 0. };
@@ -739,20 +763,16 @@ void gfs_diffusion_residual (FttCell * cell, GfsVariable * v)
   FttCellNeighbors neighbor;
   FttCellFace face;
 
-  g_return_if_fail (cell != NULL);
-  g_return_if_fail (v != NULL);
-
   h = ftt_cell_size (cell);
   if (GFS_IS_MIXED (cell)) {
-    a = GFS_STATE (cell)->solid->a*GFS_STATE (cell)->g[0];
+    a = GFS_STATE (cell)->solid->a*GFS_VARIABLE (cell, p->dia);
     if (((cell)->flags & GFS_FLAG_DIRICHLET) != 0)
-      g.b = gfs_cell_dirichlet_gradient_flux (cell, v->i, -1, 
-					      GFS_STATE (cell)->solid->fv);
+      g.b = gfs_cell_dirichlet_gradient_flux (cell, p->u, -1, GFS_STATE (cell)->solid->fv);
     else
       g.b = GFS_STATE (cell)->solid->fv;
   }
   else
-    a = GFS_STATE (cell)->g[0];
+    a = GFS_VARIABLE (cell, p->dia);
 
   face.cell = cell;
   ftt_cell_neighbors (cell, &neighbor);
@@ -760,15 +780,52 @@ void gfs_diffusion_residual (FttCell * cell, GfsVariable * v)
     GfsGradient ng;
 
     face.neighbor = neighbor.c[face.d];
-    gfs_face_gradient_flux (&face, &ng, v->i, -1);
+    gfs_face_gradient_flux (&face, &ng, p->u, -1);
     g.a += ng.a;
     g.b += ng.b;
   }
-  a *= 2.*h*h;
+  a *= h*h;
   g_assert (a > 0.);
   g.a = 1. + g.a/a;
-  g.b = GFS_STATE (cell)->div + g.b/a;
-  GFS_STATE (cell)->res = g.b - g.a*GFS_VARIABLE (cell, v->i);
+  g.b = GFS_VARIABLE (cell, p->rhs) + g.b/a;
+  GFS_VARIABLE (cell, p->res) = g.b - g.a*GFS_VARIABLE (cell, p->u);
+}
+
+/**
+ * gfs_diffusion_residual:
+ * @domain: a #GfsDomain.
+ * @u: the variable to use as left-hand side.
+ * @rhs: the right-hand side.
+ * @dia: the diagonal weight.
+ * @res: the residual.
+ *
+ * Sets the @res variable of each leaf cell of @domain to the residual
+ * of the diffusion equation for @v.
+ *
+ * The diffusion coefficients must have been set using
+ * gfs_diffusion_coefficients() and the right-hand side using
+ * gfs_diffusion_rhs().
+ */
+void gfs_diffusion_residual (GfsDomain * domain,
+			     GfsVariable * u,
+			     GfsVariable * rhs,
+			     GfsVariable * dia,
+			     GfsVariable * res)
+{
+  RelaxParams p;
+
+  g_return_if_fail (domain != NULL);
+  g_return_if_fail (u != NULL);
+  g_return_if_fail (rhs != NULL);
+  g_return_if_fail (dia != NULL);
+  g_return_if_fail (res != NULL);
+
+  p.u = u->i;
+  p.rhs = rhs->i;
+  p.dia = dia->i;
+  p.res = res->i;
+  gfs_domain_cell_traverse (domain, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
+			    (FttCellTraverseFunc) diffusion_residual, &p);
 }
 
 /**
@@ -778,71 +835,90 @@ void gfs_diffusion_residual (FttCell * cell, GfsVariable * v)
  * @depth: the total depth of the domain.
  * @nrelax: the number of relaxations to apply at each level.
  * @u: the variable to use as left-hand side.
+ * @rhs: the right-hand side.
+ * @dia: the diagonal weight.
+ * @res: the residual.
  *
  * Apply one multigrid iteration to the diffusion equation for @u.
  *
- * The initial value of %GFS_RES on the leaves of @root must be set to
+ * The initial value of @res on the leaves of @root must be set to
  * the residual of the diffusion equation using gfs_diffusion_residual().
  *
  * The diffusion coefficients must be set using gfs_diffusion_coefficients().
  *
  * The values of @u on the leaf cells are updated as well as the values
- * of %GFS_RES (i.e. the cell tree is ready for another iteration).
+ * of @res (i.e. the cell tree is ready for another iteration).
  */
 void gfs_diffusion_cycle (GfsDomain * domain,
 			  guint levelmin,
 			  guint depth,
 			  guint nrelax,
-			  GfsVariable * u)
+			  GfsVariable * u,
+			  GfsVariable * rhs,
+			  GfsVariable * dia,
+			  GfsVariable * res)
 {
-  gint n, l;
+  guint n;
+  GfsVariable * dp;
+  RelaxParams p;
+  gpointer data[2];
 
   g_return_if_fail (domain != NULL);
   g_return_if_fail (u != NULL);
+  g_return_if_fail (rhs != NULL);
+  g_return_if_fail (dia != NULL);
+  g_return_if_fail (res != NULL);
+
+  dp = gfs_temporary_variable (domain);
 
   /* compute residual on non-leafs cells */
   gfs_domain_cell_traverse (domain, 
 			    FTT_POST_ORDER, FTT_TRAVERSE_NON_LEAFS, -1,
-			    (FttCellTraverseFunc) gfs_get_from_below_intensive, 
-			    gfs_res);
+			    (FttCellTraverseFunc) gfs_get_from_below_intensive, res);
 
   /* relax top level */
   gfs_domain_cell_traverse (domain, 
 			    FTT_PRE_ORDER, FTT_TRAVERSE_LEVEL, levelmin,
-			    (FttCellTraverseFunc) gfs_cell_reset, gfs_dp);
-  l = levelmin;
+			    (FttCellTraverseFunc) gfs_cell_reset, dp);
+  p.maxlevel = levelmin;
+  p.u = dp->i;
+  p.res = res->i;
+  p.dia = dia->i;
   for (n = 0; n < 10*nrelax; n++) {
     gfs_domain_homogeneous_bc (domain, 
 			       FTT_TRAVERSE_LEVEL | FTT_TRAVERSE_LEAFS,
-			       levelmin, gfs_dp, u);
+			       levelmin, dp, u);
     gfs_domain_cell_traverse (domain, FTT_PRE_ORDER, 
 			      FTT_TRAVERSE_LEVEL | FTT_TRAVERSE_LEAFS, 
 			      levelmin,
-			      (FttCellTraverseFunc) diffusion_relax, &l);
+			      (FttCellTraverseFunc) diffusion_relax, &p);
   }
 
   /* relax from top to bottom */
-  for (l = levelmin + 1; l <= depth; l++) {
+  for (p.maxlevel = levelmin + 1; p.maxlevel <= depth; p.maxlevel++) {
     /* get initial guess from coarser grid */ 
-    gfs_domain_cell_traverse (domain, 
-			      FTT_PRE_ORDER, FTT_TRAVERSE_LEVEL, l,
-			      (FttCellTraverseFunc) gfs_get_from_above, 
-			      gfs_dp);
+    gfs_domain_cell_traverse (domain,
+			      FTT_PRE_ORDER, FTT_TRAVERSE_LEVEL | FTT_TRAVERSE_NON_LEAFS,
+			      p.maxlevel - 1,
+			      (FttCellTraverseFunc) get_from_above, dp);
     for (n = 0; n < nrelax; n++) {
       gfs_domain_homogeneous_bc (domain, 
 				 FTT_TRAVERSE_LEVEL | FTT_TRAVERSE_LEAFS,
-				 l, gfs_dp, u);
+				 p.maxlevel, dp, u);
       gfs_domain_cell_traverse (domain, FTT_PRE_ORDER, 
-				FTT_TRAVERSE_LEVEL | FTT_TRAVERSE_LEAFS, l,
-				(FttCellTraverseFunc) diffusion_relax, &l);
+				FTT_TRAVERSE_LEVEL | FTT_TRAVERSE_LEAFS, p.maxlevel,
+				(FttCellTraverseFunc) diffusion_relax, &p);
     }
   }
   /* correct on leaf cells */
+  data[0] = u;
+  data[1] = dp;
   gfs_domain_cell_traverse (domain, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
-			    (FttCellTraverseFunc) correct, u);
+			    (FttCellTraverseFunc) correct, data);
   gfs_domain_bc (domain, FTT_TRAVERSE_LEAFS, -1, u);
   /* compute new residual on leaf cells */
-  gfs_domain_cell_traverse (domain, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
-			    (FttCellTraverseFunc) gfs_diffusion_residual, u);
+  gfs_diffusion_residual (domain, u, rhs, dia, res);
+
+  gts_object_destroy (GTS_OBJECT (dp));
 }
 
