@@ -20,22 +20,20 @@
 #include <stdlib.h>
 #include "refine.h"
 #include "solid.h"
+#include "adaptive.h"
 
 /* GfsRefine: Object */
 
 static gboolean refine_maxlevel (FttCell * cell, GfsFunction * maxlevel)
 {
-  FttVector p;
-
-  ftt_cell_pos (cell, &p);
-  return (ftt_cell_level (cell) < gfs_function_value (maxlevel, &p, 0.));
+  return (ftt_cell_level (cell) < gfs_function_value (maxlevel, cell));
 }
 
 static void refine_box (GfsBox * box, GfsFunction * maxlevel)
 {
   ftt_cell_refine (box->root, 
 		   (FttCellRefineFunc) refine_maxlevel, maxlevel,
-		   (FttCellInitFunc) gfs_cell_init, gfs_box_domain (box));
+		   (FttCellInitFunc) gfs_cell_fine_init, gfs_box_domain (box));
 }
 
 static void gfs_refine_refine (GfsRefine * refine, GfsSimulation * sim)
@@ -83,7 +81,7 @@ static void gfs_refine_read (GtsObject ** o, GtsFile * fp)
   }
   gts_file_next_token (fp);
 
-  gfs_function_read (refine->maxlevel, fp);
+  gfs_function_read (refine->maxlevel, gfs_object_simulation (refine), fp);
   if (fp->type == GTS_ERROR)
     return;
 
@@ -138,41 +136,72 @@ GfsRefine * gfs_refine_new (GfsRefineClass * klass)
 
 /* GfsRefineSolid: Object */
 
-static gboolean refine_solid_maxlevel (FttCell * cell, GfsRefine * refine)
-{
-  if (GFS_IS_MIXED (cell)) {
-    FttVector p;
+typedef struct _GfsRefineSolid           GfsRefineSolid;
 
-    gfs_cell_cm (cell, &p);
-    return (ftt_cell_level (cell) < gfs_function_value (refine->maxlevel, &p, 0.));
+struct _GfsRefineSolid {
+  GfsRefine parent;
+
+  GfsDerivedVariable * v;
+};
+  
+#define GFS_REFINE_SOLID(obj)            GTS_OBJECT_CAST (obj,\
+					           GfsRefineSolid,\
+					           gfs_refine_solid_class ())
+
+static void refine_solid_destroy (GtsObject * object)
+{
+  gfs_domain_remove_derived_variable (GFS_DOMAIN (gfs_object_simulation (object)), 
+				      "SolidCurvature");
+
+  (* GTS_OBJECT_CLASS (gfs_refine_solid_class ())->parent_class->destroy) (object);
+}
+
+static void max_kappa (GtsVertex * v, gpointer * data)
+{
+  GtsSurface * s = data[0];
+  gdouble * max = data[1];
+  GtsVector Kh;
+
+  if (gts_vertex_mean_curvature_normal (v, s, Kh)) {
+    gdouble kappa = gts_vector_norm (Kh)/(FTT_DIMENSION - 1);
+    if (kappa > *max)
+      *max = kappa;
   }
-  return FALSE;
 }
 
-static void refine_solid_fractions (FttCell * cell, GfsSimulation * sim)
+static gdouble solid_curvature (FttCell * cell, FttCellFace * face, 
+				GfsDomain * domain, GtsSurface * s)
 {
-  gfs_cell_init (cell, GFS_DOMAIN (sim));
-  if (GFS_CELL_IS_BOUNDARY (ftt_cell_parent (cell)))
-    cell->flags |= GFS_FLAG_BOUNDARY;
-  else
-    gfs_cell_init_solid_fractions (cell, sim->surface, sim->stree, sim->is_open,
-				   TRUE, (FttCellCleanupFunc) gfs_cell_cleanup,
-				   NULL);
+  gdouble kappa = gfs_solid_is_thin (cell, s) ? 1./ftt_cell_size (cell) : 0.;
+  gpointer data[] = {GTS_OBJECT (s)->reserved, &kappa};
+  gts_surface_foreach_vertex (s, (GtsFunc) max_kappa, data);
+  return kappa;
 }
 
-static void refine_solid (GfsBox * box, gpointer * data)
+static void refine_solid_read (GtsObject ** o, GtsFile * fp)
+{
+  GfsDerivedVariable v = { "SolidCurvature", solid_curvature };
+  GfsRefineSolid * refine = GFS_REFINE_SOLID (*o);
+
+  refine->v = gfs_domain_add_derived_variable (GFS_DOMAIN (gfs_object_simulation (*o)), v);
+  if (!refine->v) {
+    gts_file_error (fp, "derived variable `SolidCurvature' already defined");
+    return;
+  }
+
+  (* GTS_OBJECT_CLASS (gfs_refine_solid_class ())->parent_class->read) (o, fp);
+}
+
+static void refine_cut_cell (FttCell * cell, GtsSurface * s, gpointer * data)
 {
   GfsRefine * refine = data[0];
-  GfsSimulation * sim = data[1];
+  GfsDomain * domain = data[1];
 
-  gfs_cell_init_solid_fractions (box->root, 
-  				 sim->surface, sim->stree, sim->is_open, 
-  				 TRUE, (FttCellCleanupFunc) gfs_cell_cleanup,
-  				 NULL);
-g_assert (!FTT_CELL_IS_DESTROYED (box->root));
-  ftt_cell_refine (box->root, 
-		   (FttCellRefineFunc) refine_solid_maxlevel, refine,
-		   (FttCellInitFunc) refine_solid_fractions, sim);
+  GTS_OBJECT (s)->reserved = GFS_SIMULATION (domain)->surface;
+  GFS_REFINE_SOLID (refine)->v->data = s;
+  if (ftt_cell_level (cell) < gfs_function_value (refine->maxlevel, cell))
+    ftt_cell_refine_single (cell, (FttCellInitFunc) gfs_cell_fine_init, domain);
+  GFS_REFINE_SOLID (refine)->v->data = NULL;
 }
 
 static void gfs_refine_solid_refine (GfsRefine * refine, GfsSimulation * sim)
@@ -182,24 +211,28 @@ static void gfs_refine_solid_refine (GfsRefine * refine, GfsSimulation * sim)
 
     data[0] = refine;
     data[1] = sim;
-    gts_container_foreach (GTS_CONTAINER (sim), (GtsFunc) refine_solid, data);
+    gfs_domain_traverse_cut (GFS_DOMAIN (sim), sim->surface, 
+			     FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS,
+			     (FttCellTraverseCutFunc) refine_cut_cell, data);
   }
 }
 
-static void gfs_refine_solid_class_init (GfsRefineSolidClass * klass)
+static void gfs_refine_solid_class_init (GfsRefineClass * klass)
 {
-  GFS_REFINE_CLASS (klass)->refine = gfs_refine_solid_refine;
+  klass->refine = gfs_refine_solid_refine;
+  GTS_OBJECT_CLASS (klass)->destroy = refine_solid_destroy;
+  GTS_OBJECT_CLASS (klass)->read = refine_solid_read;
 }
 
-GfsRefineSolidClass * gfs_refine_solid_class (void)
+GfsRefineClass * gfs_refine_solid_class (void)
 {
-  static GfsRefineSolidClass * klass = NULL;
+  static GfsRefineClass * klass = NULL;
 
   if (klass == NULL) {
     GtsObjectClassInfo gfs_refine_solid_info = {
       "GfsRefineSolid",
       sizeof (GfsRefineSolid),
-      sizeof (GfsRefineSolidClass),
+      sizeof (GfsRefineClass),
       (GtsObjectClassInitFunc) gfs_refine_solid_class_init,
       (GtsObjectInitFunc) NULL,
       (GtsArgSetFunc) NULL,
@@ -212,40 +245,37 @@ GfsRefineSolidClass * gfs_refine_solid_class (void)
   return klass;
 }
 
-/* GfsRefineDistance: Object */
+/* GfsRefineSurface: Object */
 
-static void refine_distance_destroy (GtsObject * object)
+static void refine_surface_destroy (GtsObject * object)
 {
-  GfsRefineDistance * d = GFS_REFINE_DISTANCE (object);
+  GfsRefineSurface * d = GFS_REFINE_SURFACE (object);
 
   if (d->surface)
     gts_object_destroy (GTS_OBJECT (d->surface));
-  if (d->stree)
-    gts_bb_tree_destroy (d->stree, TRUE);  
 
-  (* GTS_OBJECT_CLASS (gfs_refine_distance_class ())->parent_class->destroy) 
-    (object);
+  (* GTS_OBJECT_CLASS (gfs_refine_surface_class ())->parent_class->destroy) (object);
 }
 
-static void refine_distance_write (GtsObject * o, FILE * fp)
+static void refine_surface_write (GtsObject * o, FILE * fp)
 {
-  GfsRefineDistance * d = GFS_REFINE_DISTANCE (o);
+  GfsRefineSurface * d = GFS_REFINE_SURFACE (o);
   
-  (* GTS_OBJECT_CLASS (gfs_refine_distance_class ())->parent_class->write) (o, fp);
+  (* GTS_OBJECT_CLASS (gfs_refine_surface_class ())->parent_class->write) (o, fp);
   fprintf (fp, " { ");
   gts_surface_write (d->surface, fp);
   fputs ("}\n", fp);
 }
 
-static void refine_distance_read (GtsObject ** o, GtsFile * fp)
+static void refine_surface_read (GtsObject ** o, GtsFile * fp)
 {
-  GfsRefineDistance * refine;
+  GfsRefineSurface * refine;
 
-  (* GTS_OBJECT_CLASS (gfs_refine_distance_class ())->parent_class->read) (o, fp);
+  (* GTS_OBJECT_CLASS (gfs_refine_surface_class ())->parent_class->read) (o, fp);
   if (fp->type == GTS_ERROR)
     return;
 
-  refine = GFS_REFINE_DISTANCE (*o);
+  refine = GFS_REFINE_SURFACE (*o);
   if (fp->type != '{') {
     FILE * f;
     GtsFile * gf;
@@ -284,72 +314,206 @@ static void refine_distance_read (GtsObject ** o, GtsFile * fp)
     }
     fp->scope_max--;
   }
-  refine->stree = gts_bb_tree_surface (refine->surface);
 
   gts_file_next_token (fp);
 }
 
-static gboolean refine_distance_maxlevel (FttCell * cell, GfsRefine * refine)
+static void gfs_refine_surface_refine (GfsRefine * refine, GfsSimulation * sim)
 {
-  FttVector pos;
-  GtsPoint p;
-  gdouble d;
+  gpointer data[2];
 
-  ftt_cell_pos (cell, &pos);
-  p.x = pos.x; p.y = pos.y; p.z = pos.z;
-  d = gts_bb_tree_point_distance (GFS_REFINE_DISTANCE (refine)->stree, &p,
-				  (GtsBBoxDistFunc) gts_point_triangle_distance, NULL);
-  return (ftt_cell_level (cell) < gfs_function_value (refine->maxlevel, &pos, d));
+  data[0] = refine;
+  data[1] = sim;
+  gfs_domain_traverse_cut (GFS_DOMAIN (sim), GFS_REFINE_SURFACE (refine)->surface,
+			   FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS,
+			   (FttCellTraverseCutFunc) refine_cut_cell, data);
 }
 
-static void refine_distance (GfsBox * box, gpointer data)
+static void gfs_refine_surface_class_init (GfsRefineClass * klass)
 {
-  ftt_cell_refine (box->root, 
-		   (FttCellRefineFunc) refine_distance_maxlevel, data,
-		   (FttCellInitFunc) gfs_cell_init, gfs_box_domain (box));
+  klass->refine = gfs_refine_surface_refine;
+
+  GTS_OBJECT_CLASS (klass)->destroy = refine_surface_destroy;
+  GTS_OBJECT_CLASS (klass)->write = refine_surface_write;
+  GTS_OBJECT_CLASS (klass)->read = refine_surface_read;
 }
 
-static void refine_distance_refine (GfsRefine * refine, GfsSimulation * sim)
+static void refine_surface_init (GfsRefineSurface * r)
 {
-  if (GFS_REFINE_DISTANCE (refine)->surface)
-    gts_container_foreach (GTS_CONTAINER (sim), (GtsFunc) refine_distance, 
-			   refine);
-}
-
-static void gfs_refine_distance_class_init (GfsRefineDistanceClass * klass)
-{
-  GFS_REFINE_CLASS (klass)->refine = refine_distance_refine;
-  
-  GTS_OBJECT_CLASS (klass)->destroy = refine_distance_destroy;
-  GTS_OBJECT_CLASS (klass)->write = refine_distance_write;
-  GTS_OBJECT_CLASS (klass)->read = refine_distance_read;
-}
-
-static void gfs_refine_distance_init (GfsRefineDistance * d)
-{
-  d->surface = gts_surface_new (gts_surface_class (),
-				gts_face_class (),
-				gts_edge_class (),
+  r->surface = gts_surface_new (gts_surface_class (), 
+				gts_face_class (), 
+				gts_edge_class (), 
 				gts_vertex_class ());
 }
 
-GfsRefineDistanceClass * gfs_refine_distance_class (void)
+GfsRefineClass * gfs_refine_surface_class (void)
 {
-  static GfsRefineDistanceClass * klass = NULL;
+  static GfsRefineClass * klass = NULL;
+
+  if (klass == NULL) {
+    GtsObjectClassInfo gfs_refine_surface_info = {
+      "GfsRefineSurface",
+      sizeof (GfsRefineSurface),
+      sizeof (GfsRefineClass),
+      (GtsObjectClassInitFunc) gfs_refine_surface_class_init,
+      (GtsObjectInitFunc) refine_surface_init,
+      (GtsArgSetFunc) NULL,
+      (GtsArgGetFunc) NULL
+    };
+    klass = gts_object_class_new (GTS_OBJECT_CLASS (gfs_refine_class ()),
+				  &gfs_refine_surface_info);
+  }
+
+  return klass;
+}
+
+/* GfsRefineDistance: Object */
+
+static void refine_distance_destroy (GtsObject * object)
+{
+  GfsRefineDistance * d = GFS_REFINE_DISTANCE (object);
+
+  if (d->stree)
+    gts_bb_tree_destroy (d->stree, TRUE);
+  gfs_domain_remove_derived_variable (GFS_DOMAIN (gfs_object_simulation (object)), "Distance");
+
+  (* GTS_OBJECT_CLASS (gfs_refine_distance_class ())->parent_class->destroy) (object);
+}
+
+static gdouble cell_distance (FttCell * cell, 
+			      FttCellFace * face, 
+			      GfsSimulation * sim,
+			      GfsRefineDistance * refine)
+{
+  FttVector pos;
+  gdouble h = GFS_DIAGONAL*ftt_cell_size (cell), d;
+  GtsPoint p;
+
+  ftt_cell_pos (cell, &pos);
+  p.x = pos.x; p.y = pos.y; p.z = pos.z;
+  d = gts_bb_tree_point_distance (refine->stree, &p,
+				  (GtsBBoxDistFunc) gts_point_triangle_distance, NULL);
+  return d > h ? d - h : 0.;
+}
+
+static void refine_distance_read (GtsObject ** o, GtsFile * fp)
+{
+  GfsDerivedVariable v = { "Distance", cell_distance };
+
+  v.data = *o;
+  if (!gfs_domain_add_derived_variable (GFS_DOMAIN (gfs_object_simulation (*o)), v)) {
+    gts_file_error (fp, "derived variable `Distance' already defined");
+    return;
+  }
+
+  (* GTS_OBJECT_CLASS (gfs_refine_distance_class ())->parent_class->read) (o, fp);
+  if (fp->type == GTS_ERROR)
+    return;
+
+  GFS_REFINE_DISTANCE (*o)->stree = gts_bb_tree_surface (GFS_REFINE_SURFACE (*o)->surface);
+}
+
+static void gfs_refine_distance_class_init (GfsRefineClass * klass)
+{
+  klass->refine = gfs_refine_refine;
+
+  GTS_OBJECT_CLASS (klass)->destroy = refine_distance_destroy;
+  GTS_OBJECT_CLASS (klass)->read = refine_distance_read;
+}
+
+GfsRefineClass * gfs_refine_distance_class (void)
+{
+  static GfsRefineClass * klass = NULL;
 
   if (klass == NULL) {
     GtsObjectClassInfo gfs_refine_distance_info = {
       "GfsRefineDistance",
       sizeof (GfsRefineDistance),
-      sizeof (GfsRefineDistanceClass),
+      sizeof (GfsRefineClass),
       (GtsObjectClassInitFunc) gfs_refine_distance_class_init,
-      (GtsObjectInitFunc) gfs_refine_distance_init,
+      (GtsObjectInitFunc) NULL,
       (GtsArgSetFunc) NULL,
       (GtsArgGetFunc) NULL
     };
-    klass = gts_object_class_new (GTS_OBJECT_CLASS (gfs_refine_class ()),
+    klass = gts_object_class_new (GTS_OBJECT_CLASS (gfs_refine_surface_class ()),
 				  &gfs_refine_distance_info);
   }
 
   return klass;
 }
+
+/* GfsRefineHeight: Object */
+
+static void refine_height_destroy (GtsObject * object)
+{
+  gfs_domain_remove_derived_variable (GFS_DOMAIN (gfs_object_simulation (object)), "Height");
+
+  (* GTS_OBJECT_CLASS (gfs_refine_height_class ())->parent_class->destroy) (object);
+}
+
+static gdouble interpolated_value (GtsSurface * s, FttVector * p)
+{
+  GtsPoint q;
+  GtsFace * t;
+
+  q.x = p->x; q.y = p->y;
+  t = gts_point_locate (&q, s, NULL);
+  if (t == NULL) {
+    g_warning ("cannot locate point (%g,%g)", p->x, p->y);
+    return 0.;
+  }
+  gts_triangle_interpolate_height (GTS_TRIANGLE (t), &q);
+  return q.z;
+}
+
+static gdouble cell_height (FttCell * cell, 
+			    FttCellFace * face, 
+			    GfsSimulation * sim,
+			    GfsRefineSurface * refine)
+{
+  FttVector pos;
+  ftt_cell_pos (cell, &pos);
+  return interpolated_value (refine->surface, &pos);
+}
+
+static void refine_height_read (GtsObject ** o, GtsFile * fp)
+{
+  GfsDerivedVariable v = { "Height", cell_height };
+
+  v.data = *o;
+  if (!gfs_domain_add_derived_variable (GFS_DOMAIN (gfs_object_simulation (*o)), v)) {
+    gts_file_error (fp, "derived variable `Height' already defined");
+    return;
+  }
+
+  (* GTS_OBJECT_CLASS (gfs_refine_height_class ())->parent_class->read) (o, fp);
+}
+
+static void gfs_refine_height_class_init (GfsRefineClass * klass)
+{
+  klass->refine = gfs_refine_refine;
+
+  GTS_OBJECT_CLASS (klass)->destroy = refine_height_destroy;
+  GTS_OBJECT_CLASS (klass)->read = refine_height_read;
+}
+
+GfsRefineClass * gfs_refine_height_class (void)
+{
+  static GfsRefineClass * klass = NULL;
+
+  if (klass == NULL) {
+    GtsObjectClassInfo gfs_refine_height_info = {
+      "GfsRefineHeight",
+      sizeof (GfsRefineSurface),
+      sizeof (GfsRefineClass),
+      (GtsObjectClassInitFunc) gfs_refine_height_class_init,
+      (GtsObjectInitFunc) NULL,
+      (GtsArgSetFunc) NULL,
+      (GtsArgGetFunc) NULL
+    };
+    klass = gts_object_class_new (GTS_OBJECT_CLASS (gfs_refine_surface_class ()),
+				  &gfs_refine_height_info);
+  }
+
+  return klass;
+}
