@@ -22,35 +22,34 @@
 
 #include <gts.h>
 
-#include "fluid.h"
+#include "domain.h"
 
 #ifdef __cplusplus
 extern "C" {
 #endif /* __cplusplus */
 
 void         gfs_cell_fluid                              (FttCell * cell);
-void         gfs_cell_init_solid_fractions        (FttCell * root, 
-						   GtsSurface * s,
-						   GNode * stree,
-						   gboolean is_open,
-						   gboolean destroy_solid,
-						   FttCellCleanupFunc cleanup,
-						   gpointer data);
+gboolean     gfs_solid_is_thin                           (FttCell * cell, 
+							  GtsSurface * s);
+gboolean     gfs_set_2D_solid_fractions_from_surface     (FttCell * cell,
+							  GtsSurface * s);
+guint        gfs_domain_init_solid_fractions             (GfsDomain * domain,
+							  GtsSurface * s,
+							  gboolean destroy_solid,
+							  FttCellCleanupFunc cleanup,
+							  gpointer data,
+							  GfsVariable * status);
 void         gfs_cell_init_solid_fractions_from_children (FttCell * cell);
-gboolean     gfs_cell_check_solid_fractions              (FttCell * root,
-							  GtsSurface * solid,
-							  GNode * stree,
-							  gboolean is_open);
-gboolean     gfs_refine_mixed                       (const FttCell * cell);
-void         gfs_cell_init_fraction                 (FttCell * root, 
-						     GtsSurface * s,
-						     GNode * stree,
-						     gboolean is_open,
-						     GfsVariable * c);
-void         gfs_cell_cm                            (const FttCell * cell, 
-						     FttVector * cm);
-void         gfs_face_ca                            (const FttCellFace * face, 
-						     FttVector * ca);
+gboolean     gfs_cell_check_solid_fractions              (FttCell * root);
+void         gfs_domain_init_fraction                    (GfsDomain * domain,
+							  GtsSurface * s,
+							  GfsVariable * c);
+void         gfs_cell_cm                                 (const FttCell * cell, 
+							  FttVector * cm);
+void         gfs_solid_normal                            (const FttCell * cell,
+							  FttVector * n);
+void         gfs_face_ca                                 (const FttCellFace * face, 
+							  FttVector * ca);
 
 #ifdef __cplusplus
 }
