@@ -106,28 +106,56 @@ void gfs_domain_variable_centered_sources (GfsDomain * domain,
 
 /* GfsSourceGeneric: Object */
 
-static void source_generic_write (GtsObject * o, FILE * fp)
+static void source_generic_init (GfsSourceGeneric * s)
 {
-  if (GTS_OBJECT_CLASS (gfs_source_generic_class ())->parent_class->write)
-    (* GTS_OBJECT_CLASS (gfs_source_generic_class ())->parent_class->write) 
-      (o, fp);
-
-  g_assert (GFS_SOURCE_GENERIC (o)->v);
-  fprintf (fp, " %s", GFS_SOURCE_GENERIC (o)->v->name);
+  GFS_EVENT (s)->istep = 1;
 }
 
-static void source_generic_read (GtsObject ** o, GtsFile * fp)
+GfsSourceGenericClass * gfs_source_generic_class (void)
 {
-  GfsSourceGeneric * source;
+  static GfsSourceGenericClass * klass = NULL;
+
+  if (klass == NULL) {
+    GtsObjectClassInfo gfs_source_generic_info = {
+      "GfsSourceGeneric",
+      sizeof (GfsSourceGeneric),
+      sizeof (GfsSourceGenericClass),
+      (GtsObjectClassInitFunc) NULL,
+      (GtsObjectInitFunc) source_generic_init,
+      (GtsArgSetFunc) NULL,
+      (GtsArgGetFunc) NULL
+    };
+    klass = gts_object_class_new (GTS_OBJECT_CLASS (gfs_event_class ()),
+				  &gfs_source_generic_info);
+  }
+
+  return klass;
+}
+
+/* GfsSourceScalar: Object */
+
+static void source_scalar_write (GtsObject * o, FILE * fp)
+{
+  if (GTS_OBJECT_CLASS (gfs_source_scalar_class ())->parent_class->write)
+    (* GTS_OBJECT_CLASS (gfs_source_scalar_class ())->parent_class->write) 
+      (o, fp);
+
+  g_assert (GFS_SOURCE_SCALAR (o)->v);
+  fprintf (fp, " %s", GFS_SOURCE_SCALAR (o)->v->name);
+}
+
+static void source_scalar_read (GtsObject ** o, GtsFile * fp)
+{
+  GfsSourceScalar * source;
   GfsDomain * domain;
 
-  if (GTS_OBJECT_CLASS (gfs_source_generic_class ())->parent_class->read)
-    (* GTS_OBJECT_CLASS (gfs_source_generic_class ())->parent_class->read) 
+  if (GTS_OBJECT_CLASS (gfs_source_scalar_class ())->parent_class->read)
+    (* GTS_OBJECT_CLASS (gfs_source_scalar_class ())->parent_class->read) 
       (o, fp);
   if (fp->type == GTS_ERROR)
     return;
 
-  source = GFS_SOURCE_GENERIC (*o);
+  source = GFS_SOURCE_SCALAR (*o);
   domain =  GFS_DOMAIN (gfs_object_simulation (source));
   if (fp->type != GTS_STRING) {
     gts_file_error (fp, "expecting a string (GfsVariable)");
@@ -147,104 +175,76 @@ static void source_generic_read (GtsObject ** o, GtsFile * fp)
   gts_file_next_token (fp);
 }
 
-static void source_generic_class_init (GfsSourceGenericClass * klass)
+static void source_scalar_class_init (GfsSourceGenericClass * klass)
 {
-  GTS_OBJECT_CLASS (klass)->read =  source_generic_read;
-  GTS_OBJECT_CLASS (klass)->write = source_generic_write;
+  GTS_OBJECT_CLASS (klass)->read =  source_scalar_read;
+  GTS_OBJECT_CLASS (klass)->write = source_scalar_write;
 }
 
-static void source_generic_init (GfsSourceGeneric * s)
-{
-  GFS_EVENT (s)->istep = 1;
-}
-
-GfsSourceGenericClass * gfs_source_generic_class (void)
+GfsSourceGenericClass * gfs_source_scalar_class (void)
 {
   static GfsSourceGenericClass * klass = NULL;
 
   if (klass == NULL) {
-    GtsObjectClassInfo gfs_source_generic_info = {
-      "GfsSourceGeneric",
-      sizeof (GfsSourceGeneric),
+    GtsObjectClassInfo gfs_source_scalar_info = {
+      "GfsSourceScalar",
+      sizeof (GfsSourceScalar),
       sizeof (GfsSourceGenericClass),
-      (GtsObjectClassInitFunc) source_generic_class_init,
-      (GtsObjectInitFunc) source_generic_init,
+      (GtsObjectClassInitFunc) source_scalar_class_init,
+      (GtsObjectInitFunc) NULL,
       (GtsArgSetFunc) NULL,
       (GtsArgGetFunc) NULL
     };
-    klass = gts_object_class_new (GTS_OBJECT_CLASS (gfs_event_class ()),
-				  &gfs_source_generic_info);
+    klass = gts_object_class_new (GTS_OBJECT_CLASS (gfs_source_generic_class ()),
+				  &gfs_source_scalar_info);
   }
 
   return klass;
 }
 
-/* GfsSourceVector: Object */
+/* GfsSourceVelocity: Object */
 
-static void source_vector_write (GtsObject * o, FILE * fp)
+static void source_velocity_read (GtsObject ** o, GtsFile * fp)
 {
-  FttComponent c;
-
-  if (GTS_OBJECT_CLASS (gfs_source_vector_class ())->parent_class->write)
-    (* GTS_OBJECT_CLASS (gfs_source_vector_class ())->parent_class->write) 
-      (o, fp);
-
-  for (c = 1; c < FTT_DIMENSION; c++) {
-    g_assert (GFS_SOURCE_VECTOR (o)->v[c]);
-    fprintf (fp, " %s", GFS_SOURCE_VECTOR (o)->v[c]->name);
-  }
-}
-
-static void source_vector_read (GtsObject ** o, GtsFile * fp)
-{
-  GfsSourceVector * source;
+  GfsSourceVelocity * source;
   GfsDomain * domain;
   FttComponent c;
 
-  if (GTS_OBJECT_CLASS (gfs_source_vector_class ())->parent_class->read)
-    (* GTS_OBJECT_CLASS (gfs_source_vector_class ())->parent_class->read) 
+  if (GTS_OBJECT_CLASS (gfs_source_velocity_class ())->parent_class->read)
+    (* GTS_OBJECT_CLASS (gfs_source_velocity_class ())->parent_class->read) 
       (o, fp);
   if (fp->type == GTS_ERROR)
     return;
 
-  source = GFS_SOURCE_VECTOR (*o);
+  source = GFS_SOURCE_VELOCITY (*o);
   domain =  GFS_DOMAIN (gfs_object_simulation (source));
-  for (c = 1; c < FTT_DIMENSION; c++) {
-    if (fp->type != GTS_STRING) {
-      gts_file_error (fp, "expecting a string (GfsVariable)");
-      return;
-    }
-    source->v[c] = gfs_variable_from_name (domain->variables, 
-					   fp->token->str);
-    if (source->v[c] == NULL) {
-      gts_file_error (fp, "unknown variable `%s'", fp->token->str);
-      return;
-    }
+  if (!(source->v = gfs_domain_velocity (domain))) {
+    gts_file_error (fp, "cannot find velocity components");
+    return;
+  }
+  for (c = 0; c < FTT_DIMENSION; c++) {
     if (source->v[c]->sources == NULL)
       source->v[c]->sources = 
 	gts_container_new (GTS_CONTAINER_CLASS (gts_slist_container_class ()));
     gts_container_add (source->v[c]->sources, GTS_CONTAINEE (source));
-  
-    gts_file_next_token (fp);
   }
 }
 
-static void source_vector_class_init (GfsSourceGenericClass * klass)
+static void source_velocity_class_init (GfsSourceGenericClass * klass)
 {
-  GTS_OBJECT_CLASS (klass)->read =  source_vector_read;
-  GTS_OBJECT_CLASS (klass)->write = source_vector_write;
+  GTS_OBJECT_CLASS (klass)->read =  source_velocity_read;
 }
 
-GfsSourceGenericClass * gfs_source_vector_class (void)
+GfsSourceGenericClass * gfs_source_velocity_class (void)
 {
   static GfsSourceGenericClass * klass = NULL;
 
   if (klass == NULL) {
     GtsObjectClassInfo source_info = {
       "GfsSourceGeneric",
-      sizeof (GfsSourceVector),
+      sizeof (GfsSourceVelocity),
       sizeof (GfsSourceGenericClass),
-      (GtsObjectClassInitFunc) source_vector_class_init,
+      (GtsObjectClassInitFunc) source_velocity_class_init,
       (GtsObjectInitFunc) NULL,
       (GtsArgSetFunc) NULL,
       (GtsArgGetFunc) NULL
@@ -275,8 +275,7 @@ static void source_read (GtsObject ** o, GtsFile * fp)
     return;
 
   GFS_SOURCE (*o)->intensity = gfs_function_new (gfs_function_class (), 0.);
-  gfs_object_simulation (GFS_SOURCE (*o)->intensity) = gfs_object_simulation (*o);
-  gfs_function_read (GFS_SOURCE (*o)->intensity, fp);
+  gfs_function_read (GFS_SOURCE (*o)->intensity, gfs_object_simulation (*o), fp);
 }
 
 static void source_write (GtsObject * o, FILE * fp)
@@ -291,14 +290,7 @@ static gdouble source_value (GfsSourceGeneric * s,
 			     FttCell * cell, 
 			     GfsVariable * v)
 {
-  FttVector p;
-
-  if (v->centered)
-    ftt_cell_pos (cell, &p);
-  else
-    gfs_cell_cm (cell, &p);
-  return gfs_function_value (GFS_SOURCE (s)->intensity, &p,
-			     gfs_object_simulation (s)->time.t);
+  return gfs_function_value (GFS_SOURCE (s)->intensity, cell);
 }
 
 static void source_class_init (GfsSourceGenericClass * klass)
@@ -324,49 +316,153 @@ GfsSourceGenericClass * gfs_source_class (void)
       (GtsArgSetFunc) NULL,
       (GtsArgGetFunc) NULL
     };
-    klass = gts_object_class_new (GTS_OBJECT_CLASS (gfs_source_generic_class ()),
+    klass = gts_object_class_new (GTS_OBJECT_CLASS (gfs_source_scalar_class ()),
 				  &source_info);
   }
 
   return klass;
 }
 
+/* GfsSourceControl: Object */
+
+static gboolean source_control_event (GfsEvent * event, GfsSimulation * sim)
+{
+  if ((* gfs_event_class ()->event) (event, sim)) {
+    GfsSourceControl * s = GFS_SOURCE_CONTROL (event);
+    GtsRange r = gfs_domain_stats_variable (GFS_DOMAIN (sim), GFS_SOURCE_SCALAR (event)->v,
+					    FTT_TRAVERSE_LEAFS, -1);
+    s->s = (gfs_function_value (GFS_SOURCE (s)->intensity, NULL) - r.mean)/
+      sim->advection_params.dt;
+    return TRUE;
+  }
+  return FALSE;
+}
+
+static gdouble source_control_value (GfsSourceGeneric * s, 
+				     FttCell * cell, 
+				     GfsVariable * v)
+{
+  return GFS_SOURCE_CONTROL (s)->s;
+}
+
+static void source_control_class_init (GfsSourceGenericClass * klass)
+{
+  GFS_EVENT_CLASS (klass)->event = source_control_event;
+  klass->mac_value = klass->centered_value = source_control_value;
+}
+
+GfsSourceGenericClass * gfs_source_control_class (void)
+{
+  static GfsSourceGenericClass * klass = NULL;
+
+  if (klass == NULL) {
+    GtsObjectClassInfo source_control_info = {
+      "GfsSourceControl",
+      sizeof (GfsSourceControl),
+      sizeof (GfsSourceGenericClass),
+      (GtsObjectClassInitFunc) source_control_class_init,
+      (GtsObjectInitFunc) NULL,
+      (GtsArgSetFunc) NULL,
+      (GtsArgGetFunc) NULL
+    };
+    klass = gts_object_class_new (GTS_OBJECT_CLASS (gfs_source_class ()),
+				  &source_control_info);
+  }
+
+  return klass;
+}
+
 /* GfsDiffusion: Object */
 
+static void diffusion_destroy (GtsObject * o)
+{
+  GfsDiffusion * d = GFS_DIFFUSION (o);
+
+  if (d->mu != gfs_function_get_variable (d->val))
+    gts_object_destroy (GTS_OBJECT (d->mu));
+  gts_object_destroy (GTS_OBJECT (d->val));
+
+  (* GTS_OBJECT_CLASS (gfs_diffusion_class ())->parent_class->destroy) (o);
+}
+
 static void diffusion_read (GtsObject ** o, GtsFile * fp)
 {
-  if (fp->type != GTS_INT && fp->type != GTS_FLOAT) {
-    gts_file_error (fp, "expecting a number (D)");
+  GfsDiffusion * d = GFS_DIFFUSION (*o);
+
+  gfs_function_read (d->val, gfs_object_simulation (*o), fp);
+  if (fp->type == GTS_ERROR)
+    return;
+
+  if (fp->type == '{') {
+    gfs_multilevel_params_read (&d->par, fp);
+    if (fp->type == GTS_ERROR)
     return;
   }
-  GFS_DIFFUSION (*o)->val = atof (fp->token->str);
-  gts_file_next_token (fp);
+
+  if (gfs_function_get_constant_value (d->val) == G_MAXDOUBLE &&
+      (d->mu = gfs_function_get_variable (d->val)) == NULL)
+    d->mu = gfs_temporary_variable (GFS_DOMAIN (gfs_object_simulation (*o)));
 }
 
 static void diffusion_write (GtsObject * o, FILE * fp)
 {
-  fprintf (fp, " %g", GFS_DIFFUSION (o)->val);
+  gfs_function_write (GFS_DIFFUSION (o)->val, fp);
+  fputc (' ', fp);
+  gfs_multilevel_params_write (&GFS_DIFFUSION (o)->par, fp);
+}
+
+static void update_mu (FttCell * cell, GfsDiffusion * d)
+{
+  GFS_VARIABLE (cell, d->mu->i) = gfs_function_value (d->val, cell);
+}
+
+static gboolean diffusion_event (GfsEvent * event, GfsSimulation * sim)
+{
+  GfsDiffusion * d = GFS_DIFFUSION (event);
+
+  if (d->mu) {
+    if (d->mu != gfs_function_get_variable (d->val))
+      gfs_domain_cell_traverse (GFS_DOMAIN (sim), FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
+				(FttCellTraverseFunc) update_mu, event);
+    gfs_domain_cell_traverse (GFS_DOMAIN (sim),
+			      FTT_POST_ORDER, FTT_TRAVERSE_NON_LEAFS, -1,
+			      (FttCellTraverseFunc) gfs_get_from_below_intensive, d->mu);
+    gfs_domain_bc (GFS_DOMAIN (sim), FTT_TRAVERSE_ALL, -1, d->mu);
+    return TRUE;
+  }
+  return FALSE;
 }
 
 static gdouble diffusion_face (GfsDiffusion * d, FttCellFace * f)
 {
-  return d->val;
+  return d->mu ? gfs_face_interpolated_value (f, d->mu->i) :
+    gfs_function_get_constant_value (d->val);
 }
 
 static gdouble diffusion_cell (GfsDiffusion * d, FttCell * cell)
 {
-  return d->val;
+  return d->mu ? GFS_VARIABLE (cell, d->mu->i) :
+    gfs_function_get_constant_value (d->val);
 }
 
 static void diffusion_class_init (GfsDiffusionClass * klass)
 {
+  GTS_OBJECT_CLASS (klass)->destroy = diffusion_destroy;
   GTS_OBJECT_CLASS (klass)->read = diffusion_read;
   GTS_OBJECT_CLASS (klass)->write = diffusion_write;
-  GFS_EVENT_CLASS (klass)->event = NULL;
+  GFS_EVENT_CLASS (klass)->event = diffusion_event;
   klass->face = diffusion_face;
   klass->cell = diffusion_cell;
 }
 
+static void diffusion_init (GfsDiffusion * d)
+{
+  gfs_multilevel_params_init (&d->par);
+  d->par.tolerance = 1e-6;
+  d->val = gfs_function_new (gfs_function_class (), 0.);
+  d->mu = NULL;
+}
+
 GfsDiffusionClass * gfs_diffusion_class (void)
 {
   static GfsDiffusionClass * klass = NULL;
@@ -377,7 +473,7 @@ GfsDiffusionClass * gfs_diffusion_class (void)
       sizeof (GfsDiffusion),
       sizeof (GfsDiffusionClass),
       (GtsObjectClassInitFunc) diffusion_class_init,
-      (GtsObjectInitFunc) NULL,
+      (GtsObjectInitFunc) diffusion_init,
       (GtsArgSetFunc) NULL,
       (GtsArgGetFunc) NULL
     };
@@ -398,136 +494,6 @@ gdouble gfs_diffusion_cell (GfsDiffusion * d, FttCell * cell)
   return (* GFS_DIFFUSION_CLASS (GTS_OBJECT (d)->klass)->cell) (d, cell);
 }
 
-/* GfsDiffusionMulti: Object */
-
-static void diffusion_multi_destroy (GtsObject * object)
-{
-  g_slist_foreach (GFS_DIFFUSION_MULTI (object)->d, (GFunc) gts_object_destroy, NULL);
-  g_slist_free (GFS_DIFFUSION_MULTI (object)->d);
-
-  (* GTS_OBJECT_CLASS (gfs_diffusion_multi_class ())->parent_class->destroy) 
-    (object);
-}
-
-static void diffusion_multi_read (GtsObject ** o, GtsFile * fp)
-{
-  GfsDiffusionMulti * m = GFS_DIFFUSION_MULTI (*o);
-  gboolean constant = TRUE;
-  
-  while (fp->type != GTS_ERROR && fp->type != '\n') {
-    GtsObject * d = gts_object_new (GTS_OBJECT_CLASS (gfs_diffusion_class ()));
-
-    (* d->klass->read) (&d, fp);
-    if (m->d || d->klass != GTS_OBJECT_CLASS (gfs_diffusion_class ()))
-      constant = FALSE;
-    m->d = g_slist_prepend (m->d, d);
-  }
-  m->d = g_slist_reverse (m->d);
-  if (fp->type != GTS_ERROR && m->d->next)
-    m->c = gfs_variable_from_name (GFS_DOMAIN (gfs_object_simulation (m))->variables, "C");
-  if (!constant) {
-    m->mu = gfs_domain_add_variable (GFS_DOMAIN (gfs_object_simulation (m)), 
-				     "_gfs_diffusion_multi");
-    g_assert (m->mu);
-  }
-}
-
-static void diffusion_multi_write (GtsObject * o, FILE * fp)
-{
-  GSList * i = GFS_DIFFUSION_MULTI (o)->d;
-
-  while (i) {
-    (* GTS_OBJECT (i->data)->klass->write) (i->data, fp);
-    i = i->next;
-  }
-}
-
-static gdouble diffusion_multi_face (GfsDiffusion * d, FttCellFace * f)
-{
-  gdouble mu1 = gfs_diffusion_face (GFS_DIFFUSION_MULTI (d)->d->data, f);
-
-  if (!GFS_DIFFUSION_MULTI (d)->d->next)
-    return mu1;
-  else {
-    /* fixme: c should be evaluated at t or t + dt/2 */
-    gdouble c = gfs_face_interpolated_value (f, GFS_DIFFUSION_MULTI (d)->c->i);
-    gdouble mu2 = gfs_diffusion_face (GFS_DIFFUSION_MULTI (d)->d->next->data, f);
-
-    return mu1 + c*(mu2 - mu1);
-  }
-}
-
-static gdouble diffusion_multi_cell (GfsDiffusion * d, FttCell * cell)
-{
-  gdouble mu1 = gfs_diffusion_cell (GFS_DIFFUSION_MULTI (d)->d->data, cell);
-
-  if (!GFS_DIFFUSION_MULTI (d)->d->next)
-    return mu1;
-  else {
-    /* fixme: c should be evaluated at t or t + dt/2 */
-    gdouble c = GFS_VARIABLE (cell, GFS_DIFFUSION_MULTI (d)->c->i);
-    gdouble mu2 = gfs_diffusion_cell (GFS_DIFFUSION_MULTI (d)->d->next->data, cell);
-    
-    return mu1 + c*(mu2 - mu1);
-  }
-}
-
-static void set_mu (FttCell * cell, GfsDiffusion * d)
-{
-  GFS_VARIABLE (cell, GFS_DIFFUSION_MULTI (d)->mu->i) = diffusion_multi_cell (d, cell);
-}
-
-static gboolean diffusion_multi_event (GfsEvent * event, GfsSimulation * sim)
-{
-  GfsDiffusionMulti * m = GFS_DIFFUSION_MULTI (event);
-  GSList * i = m->d;
-
-  while (i) {
-    if (GFS_EVENT_CLASS (GTS_OBJECT (i->data)->klass)->event)
-      (* GFS_EVENT_CLASS (GTS_OBJECT (i->data)->klass)->event) (event, sim);
-    i = i->next;
-  }
-
-  if (m->mu) {
-    gfs_domain_cell_traverse (GFS_DOMAIN (sim), FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
-			      (FttCellTraverseFunc) set_mu, m);
-    gfs_domain_bc (GFS_DOMAIN (sim), FTT_TRAVERSE_LEAFS, -1, m->mu);
-  }
-  
-  return TRUE;
-}
-
-static void diffusion_multi_class_init (GfsDiffusionClass * klass)
-{
-  GTS_OBJECT_CLASS (klass)->read = diffusion_multi_read;
-  GTS_OBJECT_CLASS (klass)->write = diffusion_multi_write;
-  GTS_OBJECT_CLASS (klass)->destroy = diffusion_multi_destroy;
-  GFS_EVENT_CLASS (klass)->event = diffusion_multi_event;
-  klass->face = diffusion_multi_face;
-  klass->cell = diffusion_multi_cell;
-}
-
-GfsDiffusionClass * gfs_diffusion_multi_class (void)
-{
-  static GfsDiffusionClass * klass = NULL;
-
-  if (klass == NULL) {
-    GtsObjectClassInfo diffusion_multi_info = {
-      "GfsDiffusionMulti",
-      sizeof (GfsDiffusionMulti),
-      sizeof (GfsDiffusionClass),
-      (GtsObjectClassInitFunc) diffusion_multi_class_init,
-      (GtsObjectInitFunc) NULL,
-      (GtsArgSetFunc) NULL,
-      (GtsArgGetFunc) NULL
-    };
-    klass = gts_object_class_new (GTS_OBJECT_CLASS (gfs_diffusion_class ()),
-				  &diffusion_multi_info);
-  }
-
-  return klass;
-}
-
 /* GfsSourceDiffusion: Object */
 
 static void source_diffusion_destroy (GtsObject * o)
@@ -562,12 +528,12 @@ static void source_diffusion_read (GtsObject ** o, GtsFile * fp)
     return;
 
   d = GFS_SOURCE_DIFFUSION (*o);
-  if (previous_diffusion_source (GFS_SOURCE_GENERIC (d)->v, d)) {
+  if (previous_diffusion_source (GFS_SOURCE_SCALAR (d)->v, d)) {
     gts_file_error (fp, "only one diffusion source can be specified");
     return;
   }
 
-  gfs_object_simulation (d->D) = gfs_object_simulation (d);
+  gfs_object_simulation_set (d->D, gfs_object_simulation (d));
   (* GTS_OBJECT (d->D)->klass->read) ((GtsObject **) &d->D, fp);
 }
 
@@ -575,9 +541,7 @@ static void source_diffusion_write (GtsObject * o, FILE * fp)
 {
   GfsSourceDiffusion * d = GFS_SOURCE_DIFFUSION (o);
 
-  if (GTS_OBJECT_CLASS (gfs_source_diffusion_class ())->parent_class->write)
-    (* GTS_OBJECT_CLASS (gfs_source_diffusion_class ())->parent_class->write) 
-      (o, fp);
+  (* GTS_OBJECT_CLASS (gfs_source_diffusion_class ())->parent_class->write) (o, fp);
   (* GTS_OBJECT (d->D)->klass->write) (GTS_OBJECT (d->D), fp);
 }
 
@@ -606,7 +570,7 @@ static gdouble source_diffusion_value (GfsSourceGeneric * s,
   if (GFS_IS_MIXED (cell)) /* this improves results for channel test */
     return 0.;
 
-  c = GFS_VELOCITY_COMPONENT (v->i);
+  c = v->component;
 
   v0 = GFS_VARIABLE (cell, v->i);
   f.cell = cell;
@@ -645,7 +609,7 @@ static void source_diffusion_class_init (GfsSourceGenericClass * klass)
 
 static void source_diffusion_init (GfsSourceDiffusion * d)
 {
-  d->D = GFS_DIFFUSION (gts_object_new (GTS_OBJECT_CLASS (gfs_diffusion_multi_class ())));
+  d->D = GFS_DIFFUSION (gts_object_new (GTS_OBJECT_CLASS (gfs_diffusion_class ())));
 }
 
 GfsSourceGenericClass * gfs_source_diffusion_class (void)
@@ -662,7 +626,7 @@ GfsSourceGenericClass * gfs_source_diffusion_class (void)
       (GtsArgSetFunc) NULL,
       (GtsArgGetFunc) NULL
     };
-    klass = gts_object_class_new (GTS_OBJECT_CLASS (gfs_source_generic_class ()),
+    klass = gts_object_class_new (GTS_OBJECT_CLASS (gfs_source_scalar_class ()),
 				  &source_diffusion_info);
   }
 
@@ -691,7 +655,7 @@ gdouble gfs_source_diffusion_cell (GfsSourceDiffusion * d, FttCell * cell)
 static void explicit_diffusion (FttCell * cell, GfsSourceGeneric * s)
 {
   GFS_VARIABLE (cell, GFS_SOURCE_DIFFUSION_EXPLICIT (s)->s->i) = 
-    source_diffusion_value (s, cell, s->v);
+    source_diffusion_value (s, cell, GFS_SOURCE_SCALAR (s)->v);
 }
 
 static gboolean gfs_source_diffusion_explicit_event (GfsEvent * event, GfsSimulation * sim)
@@ -713,7 +677,15 @@ static void gfs_source_diffusion_explicit_read (GtsObject ** o, GtsFile * fp)
     return;
 
   GFS_SOURCE_DIFFUSION_EXPLICIT (*o)->s = 
-    gfs_domain_add_variable (GFS_DOMAIN (gfs_object_simulation (*o)), NULL);
+    gfs_temporary_variable (GFS_DOMAIN (gfs_object_simulation (*o)));
+}
+
+static void gfs_source_diffusion_explicit_destroy (GtsObject * o)
+{
+  if (GFS_SOURCE_DIFFUSION_EXPLICIT (o)->s)
+    gts_object_destroy (GTS_OBJECT (GFS_SOURCE_DIFFUSION_EXPLICIT (o)->s));
+
+  (* GTS_OBJECT_CLASS (gfs_source_diffusion_explicit_class ())->parent_class->destroy) (o);
 }
 
 static gdouble source_diffusion_explicit_value (GfsSourceGeneric * s, 
@@ -727,6 +699,7 @@ static void gfs_source_diffusion_explicit_class_init (GfsSourceGenericClass * kl
 {
   GFS_EVENT_CLASS (klass)->event = gfs_source_diffusion_explicit_event;
   GTS_OBJECT_CLASS (klass)->read = gfs_source_diffusion_explicit_read;
+  GTS_OBJECT_CLASS (klass)->destroy = gfs_source_diffusion_explicit_destroy;
   klass->mac_value = klass->centered_value = source_diffusion_explicit_value;
 }
 
@@ -755,52 +728,60 @@ GfsSourceGenericClass * gfs_source_diffusion_explicit_class (void)
 
 static void source_viscosity_read (GtsObject ** o, GtsFile * fp)
 {
-  GfsVariable * v;
+  GfsSourceViscosity * source;
+  GfsSourceDiffusion * d;
+  GfsDomain * domain;
   FttComponent c;
 
-  if (GTS_OBJECT_CLASS (gfs_source_viscosity_class ())->parent_class->read)
-    (* GTS_OBJECT_CLASS (gfs_source_viscosity_class ())->parent_class->read)
-      (o, fp);
+  (* GTS_OBJECT_CLASS (gfs_source_velocity_class ())->parent_class->read) (o, fp);
   if (fp->type == GTS_ERROR)
     return;
-  
-  v = GFS_SOURCE_GENERIC (*o)->v->next;
-  for (c = 1; c < FTT_DIMENSION; c++, v = v->next)
-    if (!v) {
-      gts_file_error (fp, "not enough velocity components");
-      return;
-    }
-    else {
-      if (v->sources == NULL)
-	v->sources = 
-	  gts_container_new (GTS_CONTAINER_CLASS (gts_slist_container_class ()));
-      else if (previous_diffusion_source (v, NULL)) {
-	gts_file_error (fp, "only one diffusion source can be specified for a given variable");
-	return;
-      }
-      gts_container_add (v->sources, GTS_CONTAINEE (*o));
-    }
+
+  source = GFS_SOURCE_VISCOSITY (*o);
+  domain =  GFS_DOMAIN (gfs_object_simulation (source));
+  if (!(source->v = gfs_domain_velocity (domain))) {
+    gts_file_error (fp, "cannot find velocity components");
+    return;
+  }
+  for (c = 0; c < FTT_DIMENSION; c++) {
+    if (source->v[c]->sources == NULL)
+      source->v[c]->sources = 
+	gts_container_new (GTS_CONTAINER_CLASS (gts_slist_container_class ()));
+    gts_container_add (source->v[c]->sources, GTS_CONTAINEE (source));
+  }
+
+  d = GFS_SOURCE_DIFFUSION (*o);
+  gfs_object_simulation_set (d->D, gfs_object_simulation (d));
+  (* GTS_OBJECT (d->D)->klass->read) ((GtsObject **) &d->D, fp);
+}
+
+static void source_viscosity_write (GtsObject * o, FILE * fp)
+{
+  GfsSourceDiffusion * d = GFS_SOURCE_DIFFUSION (o);
+
+  (* GTS_OBJECT_CLASS (gfs_source_velocity_class ())->parent_class->write) (o, fp);
+  (* GTS_OBJECT (d->D)->klass->write) (GTS_OBJECT (d->D), fp);
 }
 
 static gdouble source_viscosity_non_diffusion_value (GfsSourceGeneric * s,
 						     FttCell * cell,
 						     GfsVariable * v)
 {
-  GfsVariable * mu = GFS_DIFFUSION_MULTI (GFS_SOURCE_DIFFUSION (s)->D)->mu;
+  GfsVariable * mu = GFS_SOURCE_DIFFUSION (s)->D->mu;
 
   if (mu == NULL)
     return 0.;
   else {
-    FttComponent c = GFS_VELOCITY_COMPONENT (v->i), i;
-    gdouble rho = 1.
-      /* fixme: + GFS_STATE (cell)->c*(gfs_object_simulation (s)->physical_params.rho - 1.)*/;
+    GfsVariable ** u = GFS_SOURCE_VISCOSITY (s)->v;
+    FttComponent c = v->component, j;
+    GfsFunction * alpha = gfs_object_simulation (s)->physical_params.alpha;
     gdouble h = ftt_cell_size (cell);
     gdouble a = 0.;
 
-    for (i = 0; i < FTT_DIMENSION; i++)
-      a += (gfs_center_gradient (cell, c, GFS_VELOCITY_INDEX (i))*
-	    gfs_center_gradient (cell, i, mu->i));
-    return a/(rho*h*h);
+    for (j = 0; j < FTT_DIMENSION; j++)
+      a += (gfs_center_gradient (cell, c, u[j]->i)*
+	    gfs_center_gradient (cell, j, mu->i));
+    return a*(alpha ? gfs_function_value (alpha, cell) : 1.)/(h*h);
   }
 }
 
@@ -808,16 +789,17 @@ static gdouble source_viscosity_value (GfsSourceGeneric * s,
 				       FttCell * cell,
 				       GfsVariable * v)
 {
-  gdouble rho = 1.
-    /* fixme: + GFS_STATE (cell)->c*(gfs_object_simulation (s)->physical_params.rho - 1.)*/;
+  GfsFunction * alpha = gfs_object_simulation (s)->physical_params.alpha;
 
-  return (source_diffusion_value (s, cell, v)/rho +
+  return (source_diffusion_value (s, cell, v)*(alpha ? gfs_function_value (alpha, cell) : 1.) +
 	  source_viscosity_non_diffusion_value (s, cell, v));
 }
 
 static void source_viscosity_class_init (GfsSourceGenericClass * klass)
 {
   GTS_OBJECT_CLASS (klass)->read = source_viscosity_read;
+  GTS_OBJECT_CLASS (klass)->write = source_viscosity_write;
+
   klass->mac_value = source_viscosity_value;
   klass->centered_value = source_viscosity_non_diffusion_value;
 }
@@ -829,7 +811,7 @@ GfsSourceGenericClass * gfs_source_viscosity_class (void)
   if (klass == NULL) {
     GtsObjectClassInfo source_viscosity_info = {
       "GfsSourceViscosity",
-      sizeof (GfsSourceDiffusion),
+      sizeof (GfsSourceViscosity),
       sizeof (GfsSourceGenericClass),
       (GtsObjectClassInitFunc) source_viscosity_class_init,
       (GtsObjectInitFunc) NULL,
@@ -847,8 +829,16 @@ GfsSourceGenericClass * gfs_source_viscosity_class (void)
 
 static void source_coriolis_destroy (GtsObject * o)
 {
+  FttComponent c;
+
   if (GFS_SOURCE_CORIOLIS (o)->omegaz)
     gts_object_destroy (GTS_OBJECT (GFS_SOURCE_CORIOLIS (o)->omegaz));
+  if (GFS_SOURCE_CORIOLIS (o)->drag)
+    gts_object_destroy (GTS_OBJECT (GFS_SOURCE_CORIOLIS (o)->drag));
+
+  for (c = 0; c <  2; c++)
+    if (GFS_SOURCE_CORIOLIS (o)->u[c])
+      gts_object_destroy (GTS_OBJECT (GFS_SOURCE_CORIOLIS (o)->u[c]));
 
   (* GTS_OBJECT_CLASS (gfs_source_class ())->parent_class->destroy) (o);
 }
@@ -856,55 +846,65 @@ static void source_coriolis_destroy (GtsObject * o)
 static void gfs_source_coriolis_read (GtsObject ** o, GtsFile * fp)
 {
   FttComponent c;
-  GfsVariable * v;
   GfsDomain * domain = GFS_DOMAIN (gfs_object_simulation (*o));
 
-  if (GTS_OBJECT_CLASS (gfs_source_coriolis_class ())->parent_class->read)
-    (* GTS_OBJECT_CLASS (gfs_source_coriolis_class ())->parent_class->read) (o, fp);
+  (* GTS_OBJECT_CLASS (gfs_source_coriolis_class ())->parent_class->read) (o, fp);
   if (fp->type == GTS_ERROR)
     return;
 
-  GFS_SOURCE_CORIOLIS (*o)->omegaz = gfs_function_new (gfs_function_class (), 0.);
-  gfs_object_simulation (GFS_SOURCE_CORIOLIS (*o)->omegaz) = gfs_object_simulation (*o);
-  gfs_function_read (GFS_SOURCE_CORIOLIS (*o)->omegaz, fp);
+  for (c = 0; c < FTT_DIMENSION; c++) {
+    GfsVariable * v = GFS_SOURCE_VELOCITY (*o)->v[c];
 
-  v = GFS_SOURCE_GENERIC (*o)->v->next;
-  for (c = 1; c < 2; c++, v = v->next) {
-    if (!v) {
-      gts_file_error (fp, "not enough velocity components");
-      return;
-    }
-    else {
-      if (v->sources == NULL)
-	v->sources = gts_container_new (GTS_CONTAINER_CLASS (gts_slist_container_class ()));
-      gts_container_add (v->sources, GTS_CONTAINEE (*o));
+    if (v->sources) {
+      GSList * i = GTS_SLIST_CONTAINER (v->sources)->items;
+      
+      while (i) {
+	if (i->data != *o && GFS_IS_SOURCE_CORIOLIS (i->data)) {
+	  gts_file_error (fp, "variable '%s' cannot have multiple Coriolis source terms", v->name);
+	  return;
+	}
+	i = i->next;
+      }
     }
   }
-  for (c = 0; c <  2; c++) {
-    GFS_SOURCE_CORIOLIS (*o)->u[c] = gfs_domain_add_variable (domain, NULL);
-    g_assert (GFS_SOURCE_CORIOLIS (*o)->u[c]);
+
+  GFS_SOURCE_CORIOLIS (*o)->omegaz = gfs_function_new (gfs_function_class (), 0.);
+  gfs_function_read (GFS_SOURCE_CORIOLIS (*o)->omegaz, gfs_object_simulation (*o), fp);
+
+  if (fp->type != '\n') {
+    GFS_SOURCE_CORIOLIS (*o)->drag = gfs_function_new (gfs_function_class (), 0.);
+    gfs_function_read (GFS_SOURCE_CORIOLIS (*o)->drag, gfs_object_simulation (*o), fp);
   }
+
+#if (!FTT_2D)
+  gts_container_remove (GFS_SOURCE_VELOCITY (*o)->v[FTT_Z]->sources, GTS_CONTAINEE (*o));
+#endif /* 3D */ 
+ 
+  for (c = 0; c <  2; c++)
+    GFS_SOURCE_CORIOLIS (*o)->u[c] = gfs_temporary_variable (domain);
 }
 
 static void gfs_source_coriolis_write (GtsObject * o, FILE * fp)
 {
-  if (GTS_OBJECT_CLASS (gfs_source_coriolis_class ())->parent_class->write)
-    (* GTS_OBJECT_CLASS (gfs_source_coriolis_class ())->parent_class->write) (o, fp);
-  gfs_function_write (GFS_SOURCE_CORIOLIS (o)->omegaz, fp);
+  GfsSourceCoriolis * s = GFS_SOURCE_CORIOLIS (o);
+
+  (* GTS_OBJECT_CLASS (gfs_source_coriolis_class ())->parent_class->write) (o, fp);
+  gfs_function_write (s->omegaz, fp);
+  if (s->drag)
+    gfs_function_write (s->drag, fp);
 }
 
 static gdouble gfs_source_coriolis_mac_value (GfsSourceGeneric * s,
 					      FttCell * cell,
 					      GfsVariable * v)
 {
-  FttVector p;
+  GfsSourceVelocity * sv = GFS_SOURCE_VELOCITY (s);
   gdouble f;
 
-  gfs_cell_cm (cell, &p);
-  f = gfs_function_value (GFS_SOURCE_CORIOLIS (s)->omegaz, &p, gfs_object_simulation (s)->time.t);
-  switch (GFS_VELOCITY_COMPONENT (v->i)) {
-  case FTT_X: return   f*GFS_STATE (cell)->v;
-  case FTT_Y: return - f*GFS_STATE (cell)->u;
+  f = gfs_function_value (GFS_SOURCE_CORIOLIS (s)->omegaz, cell);
+  switch (v->component) {
+  case FTT_X: return   f*GFS_VARIABLE (cell, sv->v[1]->i);
+  case FTT_Y: return - f*GFS_VARIABLE (cell, sv->v[0]->i);
   default: g_assert_not_reached ();
   }
   return 0.;
@@ -912,14 +912,15 @@ static gdouble gfs_source_coriolis_mac_value (GfsSourceGeneric * s,
 
 static void save_coriolis (FttCell * cell, GfsSourceCoriolis * s)
 {
+  GfsSourceVelocity * sv = GFS_SOURCE_VELOCITY (s);
   FttComponent c;
-  FttVector p;
   gdouble f;
 
-  gfs_cell_cm (cell, &p);
-  f = gfs_function_value (s->omegaz, &p, gfs_object_simulation (s)->time.t)/2.;
+  f = gfs_function_value (s->omegaz, cell)/2.;
   for (c = 0; c < 2; c++)
-    GFS_VARIABLE (cell, s->u[c]->i) = c == FTT_X ? f*GFS_STATE (cell)->v : -f*GFS_STATE (cell)->u;
+    GFS_VARIABLE (cell, s->u[c]->i) = c == FTT_X ?
+      f*GFS_VARIABLE (cell, sv->v[1]->i) :
+      - f*GFS_VARIABLE (cell, sv->v[0]->i);
 }
 
 static gboolean gfs_source_coriolis_event (GfsEvent * event, GfsSimulation * sim)
@@ -936,10 +937,7 @@ static gdouble gfs_source_coriolis_centered_value (GfsSourceGeneric * s,
 						   FttCell * cell,
 						   GfsVariable * v)
 {
-  FttComponent c = GFS_VELOCITY_COMPONENT (v->i);
-  GfsSourceCoriolis * b = GFS_SOURCE_CORIOLIS (s);
-
-  return GFS_VARIABLE (cell, b->u[c]->i);
+  return GFS_VARIABLE (cell, GFS_SOURCE_CORIOLIS (s)->u[v->component]->i);
 }
 
 static void gfs_source_coriolis_class_init (GfsSourceGenericClass * klass)
@@ -967,84 +965,75 @@ GfsSourceGenericClass * gfs_source_coriolis_class (void)
       (GtsArgSetFunc) NULL,
       (GtsArgGetFunc) NULL
     };
-    klass = gts_object_class_new (GTS_OBJECT_CLASS (gfs_source_generic_class ()),
+    klass = gts_object_class_new (GTS_OBJECT_CLASS (gfs_source_velocity_class ()),
 				  &gfs_source_coriolis_info);
   }
 
   return klass;
 }
 
+/**
+ * gfs_has_source_coriolis:
+ * @domain: a #GfsDomain.
+ *
+ * Returns: the #GfsSourceCoriolis associated with @domain or %NULL.
+ */
+GfsSourceCoriolis * gfs_has_source_coriolis (GfsDomain * domain)
+{
+  GfsVariable * v;
+
+  g_return_val_if_fail (domain != NULL, NULL);
+
+  v = gfs_variable_from_name (domain->variables, "U");
+  g_return_val_if_fail (v != NULL, NULL);
+
+  if (v->sources) {
+    GSList * i = GTS_SLIST_CONTAINER (v->sources)->items;
+
+    while (i) {
+      if (GFS_IS_SOURCE_CORIOLIS (i->data))
+	return i->data;
+      i = i->next;
+    }
+  }
+  return NULL;
+}
+
 static void implicit_coriolis (FttCell * cell, GfsSourceCoriolis * s)
 {
-  FttVector p;
+  GfsSourceVelocity * sv = GFS_SOURCE_VELOCITY (s);
   gdouble c, u, v;
   GfsSimulation * sim = gfs_object_simulation (s);
 
-  gfs_cell_cm (cell, &p);
-  c = sim->advection_params.dt*gfs_function_value (s->omegaz, &p, sim->time.t)/2.;
-  u = GFS_STATE (cell)->u;
-  v = GFS_STATE (cell)->v;
-  GFS_STATE (cell)->u = (u + c*v)/(1. + c*c);
-  GFS_STATE (cell)->v = (v - c*u)/(1. + c*c);
+  c = sim->advection_params.dt*gfs_function_value (s->omegaz, cell)/2.;
+  u = GFS_VARIABLE (cell, sv->v[0]->i);
+  v = GFS_VARIABLE (cell, sv->v[1]->i);
+  if (s->drag) {
+    gdouble e = sim->advection_params.dt*gfs_function_value (s->drag, cell)/2.;
+    GFS_VARIABLE (cell, sv->v[0]->i) = (u + c*v/(1. + e))/((1. + e) + c*c/(1. + e));
+    GFS_VARIABLE (cell, sv->v[1]->i) = (v - c*u/(1. + e))/((1. + e) + c*c/(1. + e));
+  }
+  else {
+    GFS_VARIABLE (cell, sv->v[0]->i) = (u + c*v)/(1. + c*c);
+    GFS_VARIABLE (cell, sv->v[1]->i) = (v - c*u)/(1. + c*c);
+  }
 }
 
 /**
  * gfs_source_coriolis_implicit:
- * @sim: a #GfsSimulation.
- * @apar: the advection parameters.
- * @p: the pressure at time n.
+ * @domain: a #GfsDomain.
+ * @dt: the timestep.
  *
- * Applies the implicit part of the Coriolis source term of @sim.
- *
- * Returns: %TRUE if the implicit part was applied, %FALSE otherwise.
+ * Applies the implicit part of the Coriolis source term of @domain.
  */
-gboolean gfs_source_coriolis_implicit (GfsSimulation * sim,
-				       GfsAdvectionParams * apar,
-				       GfsVariable * p)
+void gfs_source_coriolis_implicit (GfsDomain * domain,
+				   gdouble dt)
 {
-  GfsVariable * v;
+  GfsSourceCoriolis * s;
 
-  g_return_val_if_fail (sim != NULL, FALSE);
-  g_return_val_if_fail (p != NULL, FALSE);
+  g_return_if_fail (domain != NULL);
 
-  v = gfs_variable_from_name (GFS_DOMAIN (sim)->variables, "U");
-  g_assert (v);
-  if (v->sources) {
-    GSList * i = GTS_SLIST_CONTAINER (v->sources)->items;
-    guint implicit_sources = 0;
-    GfsSourceCoriolis * s = NULL;
-
-    while (i) {
-      if (GFS_IS_SOURCE_CORIOLIS (i->data)) {
-	if (s != NULL) {
-	  g_warning ("Multiple Coriolis source terms are not consistent");
-	  return FALSE;
-	}
-	else if (implicit_sources > 0) {
-	  g_warning ("Multiple implicit source terms are not consistent");
-	  return FALSE;
-	}
-	else
-	  s = i->data;
-      }
-      else if (!GFS_SOURCE_GENERIC_CLASS (GTS_OBJECT (i->data)->klass)->centered_value) {
-	implicit_sources++;
-	if (s != NULL || implicit_sources > 1) {
-	  g_warning ("Multiple implicit source terms are not consistent");
-	  return FALSE;
-	}
-      }
-      i = i->next;
-    }
-
-    if (s != NULL) {
-      gfs_poisson_coefficients (GFS_DOMAIN (sim), apar->c, apar->rho);
-      gfs_correct_normal_velocities (GFS_DOMAIN (sim), 2, p, apar->dt);
-      gfs_correct_centered_velocities (GFS_DOMAIN (sim), 2, apar->dt);
-      gfs_domain_cell_traverse (GFS_DOMAIN (sim), FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
-				(FttCellTraverseFunc) implicit_coriolis, s);
-      return TRUE;
-    }
-  }
-  return FALSE;
+  if ((s = gfs_has_source_coriolis (domain)))
+    gfs_domain_cell_traverse (domain, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
+			      (FttCellTraverseFunc) implicit_coriolis, s);
 }
