@@ -19,6 +19,7 @@
 
 #include <math.h>
 #include "vof.h"
+#include "variable.h"
 
 #define THRESHOLD(c) {if ((c) < 0.) c = 0.; else if ((c) > 1.) c = 1.;}
 
@@ -33,9 +34,11 @@
  */
 gdouble gfs_line_area (FttVector * m, gdouble alpha, gdouble c1)
 {
+  FttVector n;
   gdouble a, v;
 
   g_return_val_if_fail (m != NULL, 0.);
+  g_return_val_if_fail (m->x >= 0. && m->y >= 0., 0.);
 
   if (alpha <= 0.)
     return 0.;
@@ -43,38 +46,39 @@ gdouble gfs_line_area (FttVector * m, gdouble alpha, gdouble c1)
   if (alpha >= m->x*c1 + m->y || c1 == 0.)
     return c1;
 
-  g_assert (m->x >= 1e-9 && m->y >= 1e-9);
+  n = *m; n.x += 1e-4; n.y += 1e-4;
 
   v = alpha*alpha;
 
-  a = alpha - m->x*c1;
+  a = alpha - n.x*c1;
   if (a > 0.)
     v -= a*a;
 
-  a = alpha - m->y;
+  a = alpha - n.y;
   if (a > 0.)
     v -= a*a;
 
-  return v/(2.*m->x*m->y);
+  return v/(2.*n.x*n.y);
 }
 
 /**
  * gfs_line_center:
  * @m: normal to the line.
  * @alpha: line constant.
- * @c: area of cell fraction.
+ * @a: area of cell fraction.
  * @p: a #FttVector.
  *
  * Fills @p with the position of the center of mass of the fraction of
  * a square cell lying under the line (@m,@alpha).
  */
-void gfs_line_center (FttVector * m, gdouble alpha, gdouble c, FttVector * p)
+void gfs_line_center (FttVector * m, gdouble alpha, gdouble a, FttVector * p)
 {
-  gdouble a;
+  FttVector n;
+  gdouble b;
 
   g_return_if_fail (m != NULL);
   g_return_if_fail (p != NULL);
-  g_return_if_fail (c > 0. && c < 1.);
+  g_return_if_fail (m->x >= 0. && m->y >= 0.);
 
   if (alpha <= 0.) {
     p->x = p->y = 0.;
@@ -86,24 +90,26 @@ void gfs_line_center (FttVector * m, gdouble alpha, gdouble c, FttVector * p)
     return;
   }
 
-  g_assert (m->x >= 1e-9 && m->y >= 1e-9);
+  g_return_if_fail (a > 0. && a < 1.);
+
+  n = *m; n.x += 1e-4; n.y += 1e-4;
 
   p->x = p->y = alpha*alpha*alpha;
 
-  a = alpha - m->x;
-  if (a > 0.) {
-    p->x -= a*a*(alpha + 2.*m->x);
-    p->y -= a*a*a;
+  b = alpha - n.x;
+  if (b > 0.) {
+    p->x -= b*b*(alpha + 2.*n.x);
+    p->y -= b*b*b;
   }
 
-  a = alpha - m->y;
-  if (a > 0.) {
-    p->y -= a*a*(alpha + 2.*m->y);
-    p->x -= a*a*a;
+  b = alpha - n.y;
+  if (b > 0.) {
+    p->y -= b*b*(alpha + 2.*n.y);
+    p->x -= b*b*b;
   }
   
-  p->x /= 6.*m->x*m->x*m->y*c;
-  p->y /= 6.*m->x*m->y*m->y*c;
+  p->x /= 6.*n.x*n.x*n.y*a;
+  p->y /= 6.*n.x*n.y*n.y*a;
 }
 
 #if (!FTT_2D)
@@ -118,25 +124,27 @@ void gfs_line_center (FttVector * m, gdouble alpha, gdouble c, FttVector * p)
  */
 gdouble gfs_plane_volume (FttVector * m, gdouble alpha, gdouble c1)
 {
+  FttVector n;
   gdouble a, amax, v;
   gdouble * md;
   guint j;
 
   g_return_val_if_fail (m != NULL, 0.);
+  g_return_val_if_fail (m->x >= 0. && m->y >= 0. && m->z >= 0., 0.);
   
   if (alpha <= 0.)
     return 0.;
 
-  amax = m->x*c1 + m->y + m->z;
-  if (alpha >= amax || c1 == 0.)
+  if (alpha >= m->x*c1 + m->y + m->z || c1 == 0.)
     return c1;
 
-  g_assert (m->x >= 1e-9 && m->y >= 1e-9 && m->z >= 1e-9);
+  n = *m; n.x += 1e-4; n.y += 1e-4; n.z += 1e-4;
+  amax = n.x*c1 + n.y + n.z;
 
-  md = &m->x;
+  md = &n.x;
   v = alpha*alpha*alpha;
 
-  a = alpha - m->x*c1;
+  a = alpha - n.x*c1;
   if (a > 0.)
     v -= a*a*a;
   for (j = 1; j < 3; j++) {
@@ -146,7 +154,7 @@ gdouble gfs_plane_volume (FttVector * m, gdouble alpha, gdouble c1)
   }
 
   amax = alpha - amax;
-  a = amax + m->x*c1;
+  a = amax + n.x*c1;
   if (a > 0.)
     v += a*a*a;
   for (j = 1; j < 3; j++) {
@@ -155,7 +163,86 @@ gdouble gfs_plane_volume (FttVector * m, gdouble alpha, gdouble c1)
       v += a*a*a;
   }
 
-  return v/(6.*m->x*m->y*m->z);
+  return v/(6.*n.x*n.y*n.z);
+}
+
+/**
+ * gfs_plane_center:
+ * @m: normal to the plane.
+ * @alpha: plane constant.
+ * @a: volume of cell fraction.
+ * @p: a #FttVector.
+ *
+ * Fills @p with the position of the center of mass of the fraction of
+ * a cubic cell lying under the plane (@m,@alpha).
+ */
+void gfs_plane_center (FttVector * m, gdouble alpha, gdouble a, FttVector * p)
+{
+  FttVector n;
+  gdouble b, amax;
+
+  g_return_if_fail (m != NULL);
+  g_return_if_fail (p != NULL);
+  g_return_if_fail (m->x >= 0. && m->y >= 0. && m->z >= 0.);
+
+  if (alpha <= 0.) {
+    p->x = p->y = p->z = 0.;
+    return;
+  }
+
+  if (alpha >= m->x + m->y + m->z) {
+    p->x = p->y = p->z = 0.5;
+    return;
+  }
+
+  g_return_if_fail (a > 0. && a < 1.);
+
+  n = *m; n.x += 1e-4; n.y += 1e-4; n.z += 1e-4;
+
+  amax = n.x + n.y + n.z;
+  p->x = p->y = p->z = alpha*alpha*alpha*alpha;
+
+  b = alpha - n.x;
+  if (b > 0.) {
+    p->x -= b*b*b*(3.*n.x + alpha);
+    p->y -= b*b*b*b;
+    p->z -= b*b*b*b;
+  }
+  b = alpha - n.y;
+  if (b > 0.) {
+    p->y -= b*b*b*(3.*n.y + alpha);
+    p->x -= b*b*b*b;
+    p->z -= b*b*b*b;
+  }
+  b = alpha - n.z;
+  if (b > 0.) {
+    p->z -= b*b*b*(3.*n.z + alpha);
+    p->x -= b*b*b*b;
+    p->y -= b*b*b*b;
+  }
+
+  amax = alpha - amax;
+  b = amax + n.x;
+  if (b > 0.) {
+    p->y += b*b*b*(3.*n.y + alpha - n.z);
+    p->z += b*b*b*(3.*n.z + alpha - n.y);
+    p->x += b*b*b*b;
+  }
+  b = amax + n.y;
+  if (b > 0.) {
+    p->x += b*b*b*(3.*n.x + alpha - n.z);
+    p->z += b*b*b*(3.*n.z + alpha - n.x);
+    p->y += b*b*b*b;
+  }
+  b = amax + n.z;
+  if (b > 0.) {
+    p->x += b*b*b*(3.*n.x + alpha - n.y);
+    p->y += b*b*b*(3.*n.y + alpha - n.x);
+    p->z += b*b*b*b;
+  }
+
+  b = 24.*n.x*n.y*n.z*a;
+  p->x /= b*n.x; p->y /= b*n.y; p->z /= b*n.z;
 }
 #endif /* 3D */
 
@@ -275,6 +362,47 @@ gdouble gfs_plane_alpha (FttVector * m, gdouble c)
 }
 #endif /* 3D */
 
+/**
+ * gfs_youngs_normal:
+ * @cell: a #FttCell.
+ * @v: a #GfsVariable.
+ * @n: a #FttVector.
+ *
+ * Fills @n with the Youngs-averaged gradients of @v 
+ * normalised by the size of @cell.
+ */
+void gfs_youngs_normal (FttCell * cell, GfsVariable * v, FttVector * n)
+{
+  static FttDirection d[(FTT_DIMENSION - 1)*4][FTT_DIMENSION] = {
+#if FTT_2D
+    {FTT_RIGHT, FTT_TOP}, {FTT_LEFT, FTT_TOP}, {FTT_LEFT, FTT_BOTTOM}, {FTT_RIGHT, FTT_BOTTOM}
+#else  /* 3D */
+    {FTT_RIGHT, FTT_TOP, FTT_FRONT}, {FTT_LEFT, FTT_TOP, FTT_FRONT}, 
+    {FTT_LEFT, FTT_BOTTOM, FTT_FRONT}, {FTT_RIGHT, FTT_BOTTOM, FTT_FRONT},
+    {FTT_RIGHT, FTT_TOP, FTT_BACK}, {FTT_LEFT, FTT_TOP, FTT_BACK}, 
+    {FTT_LEFT, FTT_BOTTOM, FTT_BACK}, {FTT_RIGHT, FTT_BOTTOM, FTT_BACK},
+#endif /* 3D */
+  };
+  gdouble u[(FTT_DIMENSION - 1)*4];
+  guint i;
+
+  g_return_if_fail (cell != NULL);
+  g_return_if_fail (v != NULL);
+  g_return_if_fail (n != NULL);
+
+  for (i = 0; i < (FTT_DIMENSION - 1)*4; i++)
+    u[i] = gfs_cell_corner_value (cell, d[i], v, -1);
+
+#if FTT_2D
+  n->x = (u[0] + u[3] - u[1] - u[2])/2.;
+  n->y = (u[0] + u[1] - u[2] - u[3])/2.;
+#else  /* 3D */
+  n->x = (u[0] + u[3] + u[4] + u[7] - u[1] - u[2] - u[5] - u[6])/4.;
+  n->y = (u[0] + u[1] + u[4] + u[5] - u[2] - u[3] - u[6] - u[7])/4.;
+  n->z = (u[0] + u[1] + u[2] + u[3] - u[4] - u[5] - u[6] - u[7])/4.;
+#endif /* 3D */
+}
+
 static void gfs_cell_vof_advected_face_values (FttCell * cell,
 					       FttComponent c,
 					       GfsAdvectionParams * par)
@@ -284,7 +412,6 @@ static void gfs_cell_vof_advected_face_values (FttCell * cell,
 
   g_return_if_fail (cell != NULL);
   g_return_if_fail (par != NULL);
-  g_return_if_fail (par->cfl <= 0.5);
 
   f = GFS_VARIABLE (cell, par->v->i);
   THRESHOLD (f);
@@ -305,7 +432,7 @@ static void gfs_cell_vof_advected_face_values (FttCell * cell,
     GFS_STATE (cell)->f[left].v = f;
   }
   else {
-    FttVector m;
+    FttVector m1, m;
     gdouble alpha, n;
     static FttComponent d[FTT_DIMENSION][FTT_DIMENSION - 1] = {
 #if FTT_2D
@@ -316,9 +443,10 @@ static void gfs_cell_vof_advected_face_values (FttCell * cell,
     };
     guint i;
 
-    m.x = - gfs_center_gradient (cell, c, par->v->i);
+    gfs_youngs_normal (cell, par->v, &m1);
+    m.x = - (&m1.x)[c];
     for (i = 0; i < FTT_DIMENSION - 1; i++)
-      (&m.x)[i + 1] = - gfs_center_gradient (cell, d[c][i], par->v->i);
+      (&m.x)[i + 1] = - (&m1.x)[d[c][i]];
     
     if (m.x < 0.) {
       n = - u_left; u_left = - u_right; u_right = n;
@@ -330,7 +458,7 @@ static void gfs_cell_vof_advected_face_values (FttCell * cell,
     /* normalize */
     n = 0.;
     for (i = 0; i < FTT_DIMENSION; i++) {
-      (&m.x)[i] = fabs ((&m.x)[i]) + 1e-6;
+      (&m.x)[i] = fabs ((&m.x)[i]);
       n += (&m.x)[i];
     }
     for (i = 0; i < FTT_DIMENSION; i++)
@@ -459,6 +587,7 @@ void gfs_tracer_vof_advection (GfsDomain * domain,
     gfs_domain_cell_traverse (domain, FTT_PRE_ORDER, FTT_TRAVERSE_LEAFS, -1,
 			      (FttCellTraverseFunc) save_previous, data);
   }
+  par->fv = gfs_temporary_variable (domain);
   data[0] = par;
   data[1] = &c1;
   for (c = 0; c < FTT_DIMENSION; c++) {
@@ -475,6 +604,8 @@ void gfs_tracer_vof_advection (GfsDomain * domain,
     gfs_domain_bc (domain, FTT_TRAVERSE_LEAFS, -1, par->v);
   }
   cstart = (cstart + 1) % FTT_DIMENSION;
+  gts_object_destroy (GTS_OBJECT (par->fv));
+  par->fv = NULL;
 
   if (half) {
     data[0] = par->v;
@@ -486,3 +617,159 @@ void gfs_tracer_vof_advection (GfsDomain * domain,
 
   gfs_domain_timer_stop (domain, "tracer_vof_advection");
 }
+
+/**
+ * gfs_vof_coarse_fine:
+ * @parent: a #FttCell.
+ * @v: a #GfsVariable.
+ *
+ * Fills the @v variable of the children of @parent with
+ * VOF-interpolated values.
+ */
+void gfs_vof_coarse_fine (FttCell * parent, GfsVariable * v)
+{
+  gdouble f;
+  FttCellChildren child;
+  guint i;
+
+  g_return_if_fail (parent != NULL);
+  g_return_if_fail (!FTT_CELL_IS_LEAF (parent));
+  g_return_if_fail (v != NULL);
+
+  ftt_cell_children (parent, &child);
+
+  f = GFS_VARIABLE (parent, v->i);
+  THRESHOLD (f);
+
+  if (f < 1e-6 || 1. - f < 1.e-6)
+    for (i = 0; i < FTT_CELLS; i++)
+      GFS_VARIABLE (child.c[i], v->i) = f;
+  else {
+    FttVector m, m1;
+    FttComponent c;
+    gdouble n = 0., alpha;
+
+    gfs_youngs_normal (parent, v, &m);
+    for (c = 0; c < FTT_DIMENSION; c++) {
+      (&m1.x)[c] = fabs ((&m.x)[c]);
+      n += (&m1.x)[c];
+    }
+    for (c = 0; c < FTT_DIMENSION; c++)
+      (&m1.x)[c] /= n;
+    alpha = gfs_plane_alpha (&m1, f);
+
+    for (i = 0; i < FTT_CELLS; i++) {
+      gdouble alpha1 = alpha;
+      FttVector p;
+
+      ftt_cell_relative_pos (child.c[i], &p);
+      for (c = 0; c < FTT_DIMENSION; c++) {
+	(&p.x)[c] = (&m.x)[c] < 0. ? (&p.x)[c] + 0.25 : 0.25 - (&p.x)[c];
+	alpha1 -= (&m1.x)[c]*(&p.x)[c];
+      }
+      GFS_VARIABLE (child.c[i], v->i) = gfs_plane_volume (&m1, 2.*alpha1, 1.);
+    }
+  }
+}
+
+/**
+ * gfs_vof_plane:
+ * @cell: a #FttCell.
+ * @v: a #GfsVariable.
+ * @m: a #FttVector.
+ * @alpha: a double.
+ *
+ * If @cell is cut by the interface defined by @v, fills @m and @alpha
+ * with the equation of the VOF-reconstructed interface.
+ *
+ * Returns: %TRUE if @cell is cut by the interface, %FALSE otherwise.
+ */
+gboolean gfs_vof_plane (FttCell * cell, GfsVariable * v,
+			FttVector * m, gdouble * alpha)
+{
+  gdouble f;
+
+  g_return_val_if_fail (cell != NULL, FALSE);
+  g_return_val_if_fail (v != NULL, FALSE);
+  g_return_val_if_fail (m != NULL, FALSE);
+  g_return_val_if_fail (alpha != NULL, FALSE);
+
+  f = GFS_VARIABLE (cell, v->i);
+  THRESHOLD (f);
+
+  if (f < 1e-6 || 1. - f < 1.e-6)
+    return FALSE;
+  else {
+    FttVector m1;
+    FttComponent c;
+    gdouble n = 0.;
+
+    gfs_youngs_normal (cell, v, m);
+    for (c = 0; c < FTT_DIMENSION; c++) {
+      (&m1.x)[c] = fabs ((&m->x)[c]);
+      n += (&m1.x)[c];
+    }
+    for (c = 0; c < FTT_DIMENSION; c++) {
+      (&m1.x)[c] /= n;
+      (&m->x)[c] /= n;
+    }
+    *alpha = gfs_plane_alpha (&m1, f);
+
+    for (c = 0; c < FTT_DIMENSION; c++)
+      if ((&m->x)[c] < 0.)
+	*alpha += (&m->x)[c];
+    return TRUE;
+  }
+}
+
+/**
+ * gfs_vof_facet:
+ * @cell: a #FttCell.
+ * @v: a #GfsVariable.
+ *
+ * Returns: a list of newly allocated #GtsPoint defining the
+ * VOF-reconstructed interface facet defined by @v, or %NULL if @cell
+ * is not cut by the interface.
+ */
+GSList * gfs_vof_facet (FttCell * cell, GfsVariable * v)
+{
+  FttVector m;
+  gdouble alpha;
+
+  g_return_val_if_fail (cell != NULL, NULL);
+  g_return_val_if_fail (v != NULL, NULL);
+
+#if FTT_3D
+  g_assert_not_implemented ();
+#endif  
+
+  if (!gfs_vof_plane (cell, v, &m, &alpha))
+    return NULL;
+  else {
+    GSList * l = NULL;
+    gdouble x, y;
+
+    if (m.y != 0.) {
+      y = alpha/m.y;
+      if (y >= 0. && y <= 1.)
+	l = g_slist_prepend (l, gts_point_new (gts_point_class (), 0.5, 0.5 - y, 0.));
+    }
+    if (m.x != 0.) {
+      x = alpha/m.x;
+      if (x >= 0. && x <= 1.)
+	l = g_slist_prepend (l, gts_point_new (gts_point_class (), 0.5 - x, 0.5, 0.));
+    }
+    if (m.y != 0.) {
+      y = (alpha - m.x)/m.y;
+      if (y >= 0. && y <= 1.)
+	l = g_slist_prepend (l, gts_point_new (gts_point_class (), -0.5, 0.5 - y, 0.));
+    }
+    if (m.x != 0.) {
+      x = (alpha - m.y)/m.x;
+      if (x >= 0. && x <= 1.)
+	l = g_slist_prepend (l, gts_point_new (gts_point_class (), 0.5 - x, -0.5, 0.));
+    }
+    g_assert (g_slist_length (l) == 2);
+    return l;
+  }
+}
